# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""
This module defines classes used by storm to represent PostgreSQL
operators that are used for array searching.
"""

__metaclass__ = type

__all__ = [
    "ArrayConcat",
    "ArrayContains",
    "ArrayIsContainedBy",
    "ArrayOverlaps",
    ]

from storm.exceptions import ExprError
from storm.expr import (
    BinaryOper,
    Expr,
    compile,
    )
from storm.variables import Variable


class ArrayConcat(BinaryOper):
    """Concatenate an array with another array object."""
    __slots__ = ()
    oper = " || "


class ArrayContains(BinaryOper):
    """Check that an array contains another array object."""
    __slots__ = ()
    oper = " @> "


class ArrayIsContainedBy(BinaryOper):
    """Check that an array is contained by another array object."""
    __slots__ = ()
    oper = " <@ "


class ArrayOverlaps(BinaryOper):
    """Check that an array overlaps with another array object."""
    __slots__ = ()
    oper = " && "


@compile.when(ArrayConcat, ArrayContains, ArrayIsContainedBy, ArrayOverlaps)
def compile_array_oper(compile, expr, state):
    expr1 = expr.expr1
    expr2 = expr.expr2
    if expr1 is not None and isinstance(expr1, (Expr, Variable)):
        expr2 = getattr(expr1, "variable_factory", Variable)(value=expr2)
    elif expr2 is not None and isinstance(expr2, (Expr, Variable)):
        expr1 = getattr(expr2, "variable_factory", Variable)(value=expr1)
    else:
        raise ExprError(
            "Expected list variable for %r or %r" % (expr1, expr2))

    return "%s%s%s" % (
        compile(expr1, state), expr.oper, compile(expr2, state))
