# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "ResultSetDecorator",
    ]

from storm import Undef
from storm.expr import Select
from storm.zope.interfaces import IResultSet

from lazr.delegates import delegates

from lpresults.database.expr import CountDistinct


class ResultSetDecorator:
    """A decorated Storm ResultSet for 'Magic' (presenter) classes.

    Because `DistroSeriesBinaryPackage` doesn't actually exist in the
    database, the `DistroSeries`.searchPackages method uses the
    `DistroSeriesPackageCache` object to search for packages within a
    `DistroSeries`.

    Nonetheless, the users of the searchPackages method (such as the
    `DistroSeriesView`) expect a result set of DSBPs. Rather than executing
    the query prematurely and doing a list comprehension on the complete
    result set (which could be very large) to convert all the results (even
    though batching may mean only a window of 10 results is required), this
    adapted result set converts the results only when they are needed.

    This behaviour is required for other classes as well (Distribution,
    DistroArchSeries), hence a generalised solution.
    """
    delegates(IResultSet, context="result_set")

    def __init__(self, result_set, result_decorator):
        """
        :param result_set: The original result set to be decorated.
        :param result_decorator: The method with which individual results
                                 will be passed through before being
                                 returned.
        """
        self.result_set = result_set
        self.result_decorator = result_decorator

    def decorate_or_none(self, result):
        """Decorate a result or return None if the result is itself None"""
        if result is None:
            return None
        else:
            return self.result_decorator(result)

    def copy(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned result set.
        """
        new_result_set = self.result_set.copy(*args, **kwargs)
        return ResultSetDecorator(new_result_set, self.result_decorator)

    def config(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated result set.after updating the config.
        """
        self.result_set.config(*args, **kwargs)
        return self

    def __iter__(self, *args, **kwargs):
        """See `IResultSet`.

        Yield a decorated version of the returned value.
        """
        for value in self.result_set.__iter__(*args, **kwargs):
            yield self.decorate_or_none(value)

    def __getitem__(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned value.
        """
        # Can be a value or result set...
        value = self.result_set.__getitem__(*args, **kwargs)
        if isinstance(value, type(self.result_set)):
            return ResultSetDecorator(value, self.result_decorator)
        else:
            return self.decorate_or_none(value)

    def any(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned value.
        """
        value = self.result_set.any(*args, **kwargs)
        return self.decorate_or_none(value)

    def first(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned value.
        """
        value = self.result_set.first(*args, **kwargs)
        return self.decorate_or_none(value)

    def last(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned value.
        """
        value = self.result_set.last(*args, **kwargs)
        return self.decorate_or_none(value)

    def one(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned value.
        """
        value = self.result_set.one(*args, **kwargs)
        return self.decorate_or_none(value)

    def order_by(self, *args, **kwargs):
        """See `IResultSet`.

        :return: The decorated version of the returned result set.
        """
        new_result_set = self.result_set.order_by(*args, **kwargs)
        return ResultSetDecorator(new_result_set, self.result_decorator)

    def count(self, *args, **kwargs):
        """See `IResultSet`.

        Currently Storm.store.ResultSet has a limitation where aggregate
        methods do not respect the group_by config option. This decorated
        version ensures that the count method *does* respect the group_by
        config option when called without args/kwargs.
        """
        if self.result_set._group_by is not Undef:
            # distinct=False not supported for this result set
            select = Select(
                columns=CountDistinct(self.result_set._group_by),
                tables=self.result_set._tables,
                where=self.result_set._where,
                )
            result = self.result_set._store.execute(select)
            return result.get_one()[0]
        else:
            return self.result_set.count(*args, **kwargs)
