# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "ISchemaManager",
    "SchemaError",
    "SchemaManager",
    "global_schema_manager",
    ]

import sys

from zope.interface import implements

from lpresults.database.interfaces import (
    SchemaError,
    ISchemaManager,
    )


class SchemaManager:

    implements(ISchemaManager)

    _modules = {}

    def __init__(self):
        self._default_modules = {}
        self._default_packages = {}
        self._named = {}
        self._name_index = {}

    def _createModule(self, package):
        __import__(package, {}, {})
        return sys.modules[package]

    def _getModule(self, package):
        module = self._modules.get(package)
        if module is None:
            module = self._createModule(package)
            self._modules.setdefault(package, module)

        return module

    def getDefaultPackages(self):
        """Return a list of name, package tuples that are named as the default
        packages for those names.
        """
        return self._default_packages.copy()

    def setDefaultPackage(self, name, default_package):
        """Set the default package for schemas by name."""
        self._default_modules[name] = self._getModule(default_package)
        self._default_packages[name] = default_package

    def create(self, name, package=None):
        """Create a new schema.

        :param name: Name of the schema.
        :param package: Optionally, package name to use if no default.
        """
        if package is None:
            module = self._default_modules.get(name)
            if module is None:
                raise SchemaError("Schema named '%s' not found" % name)
        else:
            module = self._getModule(package)

        if name is not None and self._named.get(name) is not None:
            raise SchemaError("Schema named '%s' already exists" % name)

        schema = module.create()
        if name is not None:
            self._named[name] = name
        self._name_index[schema] = name

        return schema

    def get(self, name, package=None):
        """Get the schema by name, creating it first if necessary.

        :param name: Name of the store to get.
        :param default_uri: Optionally, URI to use if no default.
        """
        schema = self._named.get(name)
        if not schema:
            schema = self.create(name, package)

        return schema

    def remove(self, schema):
        """Remove the given schema from SchemaManager."""
        name = self._name_index[schema]
        del self._name_index[schema]
        if name in self._named:
            del self._named[name]

    def getName(self, schema):
        """Returns the name for the given schema or None if one isn't
        available."""
        return self._name_index.get(schema)


global_schema_manager = SchemaManager()
