# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "SearchAny",
    "SearchGreaterThan",
    "SearchGreaterThanOrEqualTo",
    "SearchLessThan",
    "SearchLessThanOrEqualTo",
    "SearchNotEqual",
    "SearchNull",
    "search_to_expression",
    ]

from storm.locals import Like


class SearchAny:
    """Search for any values."""
    __slots__ = ("values",)

    def __init__(self, values):
        self.values = values


class SearchGreaterThan:
    """Search for a greater than value."""
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value


class SearchGreaterThanOrEqualTo:
    """Search for a greater than or equal to value."""
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value


class SearchLessThan:
    """Search for a less than value."""
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value


class SearchLessThanOrEqualTo:
    """Search for a less than or equal to value."""
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value


class SearchNotEqual:
    """Search for a not equal value."""
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value


class SearchNullType:
    "Type for search parameters which might be unset."
    __slots__ = ()

    def __repr__(self):
        return "Null"

    def __reduce__(self):
        return "Null"

    def __getattr__(self, name):
        return None


SearchNull = SearchNullType()


def search_to_expression(column, search_value, case_sensitive=None,
        modifier=lambda x: x):
    """Convert a search value to an expression.

    :param column: Column to search on.
    :param search_value: The value to search for.
    :param case_sensitive: Whether to search case sensitive or not.
    """
    # Only apply modifier if the given value is not None
    def apply_modifier(x):
        if x is None:
            return x
        else:
            return modifier(x)

    if isinstance(search_value, SearchAny):
        return column.is_in(apply_modifier(v) for v in search_value.values)
    elif isinstance(search_value, SearchGreaterThan):
        return column > apply_modifier(search_value.value)
    elif isinstance(search_value, SearchGreaterThanOrEqualTo):
        return column >= apply_modifier(search_value.value)
    elif isinstance(search_value, SearchLessThan):
        return column < apply_modifier(search_value.value)
    elif isinstance(search_value, SearchLessThanOrEqualTo):
        return column <= apply_modifier(search_value.value)
    elif isinstance(search_value, SearchNotEqual):
        return column != apply_modifier(search_value.value)
    elif isinstance(search_value, SearchNullType):
        return column == None
    elif search_value is not None:
        if case_sensitive is not None:
            return Like(column, apply_modifier(search_value),
                case_sensitive=case_sensitive)
        else:
            return column == apply_modifier(search_value)
    else:
        return None
