# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "DatabaseTestCase",
    ]

import re
import unittest

from storm.zope.interfaces import IZStorm

from zope.component import (
    getUtility,
    queryUtility,
    provideUtility,
    )


class UtilityTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(UtilityTestCase, self).__init__(*args, **kwargs)

        self._utilities = []

    def provideUtility(self, component, interface, name=""):
        original = queryUtility(interface, name=name)
        self._utilities.append((original, interface, name))
        provideUtility(component, interface, name=name)

    def tearDown(self):
        super(UtilityTestCase, self).tearDown()

        # Restore previously utility state.
        for component, interface, name in self._utilities:
            provideUtility(component, interface, name=name)


class DatabaseTestCase(UtilityTestCase):

    def _assertSQL(self, func, expected_statement, all_statements):
        space_re = re.compile("\s")

        def crush(statement):
            statement = space_re.sub("", statement)
            return statement

        func(crush(expected_statement),
             [crush(statement) for statement in all_statements])

    def assertSQLIn(self, expected_statement, all_statements):
        """Ensure that C{expected_statement} is in C{all_statements}.

        Whitespace is completely collapsed before statements are compared.
        """
        self._assertSQL(self.assertIn, expected_statement, all_statements)

    def assertSQLNotIn(self, expected_statement, all_statements):
        """Ensure that C{expected_statement} is not in C{all_statements}.

        Whitespace is completely collapsed before statements are compared.
        """
        self._assertSQL(self.assertNotIn, expected_statement, all_statements)

    def tearDown(self):
        # Flush stores to try to catch improper statements unflushed.
        zstorm = getUtility(IZStorm)
        for name, store in zstorm.iterstores():
            store.flush()

        # Clear the alive cache *before* abort is called, to prevent
        # an useless loop over those objects in invalidate. Ideally, it
        # whould be in DatabaseLayer, but FunctionalTestCase calls abort
        # before that DatabaseLayer.testTearDown is called.
        for name, store in zstorm.iterstores():
            store._alive.clear()

        super(DatabaseTestCase, self).tearDown()
