# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "DatabaseLayer",
    ]

import os

import transaction

from zope.component import (
    queryUtility,
    provideUtility,
    )

from storm.schema.patch import UnknownPatchError
from storm.zope.interfaces import IZStorm
from storm.zope.zstorm import ZStorm

from lpresults.database.interfaces import ISchemaManager
from lpresults.database.schema import SchemaManager


class DatabaseManager:

    def __init__(self, get_store_uri, get_patch_uri, get_schema_package):
        self._databases = []
        self._get_store_uri = get_store_uri
        self._get_patch_uri = get_patch_uri
        self._get_schema_package = get_schema_package

    def addDatabase(self, store, database):
        self._databases.append({
            "store": store,
            "database": database,
            "store-uri": self._get_store_uri(store, database),
            "patch-uri": self._get_patch_uri(database),
            "schema-package": self._get_schema_package(store),
            })

    def setupDatabases(self, drop=False):
        for data in self._databases:
            name = data["database"]
            if drop:
                self._dropDatabase(name)
            self._createDatabase(name)

    def dropDatabases(self):
        for data in self._databases:
            name = data["database"]
            self._dropDatabase(name)

    def _dropDatabase(self, name):
        os.system("dropdb %s > /dev/null 2>&1" % name)

    def _createDatabase(self, name):
        os.system("createdb %s > /dev/null 2>&1" % name)

    def setupPatchStores(self, layer):
        layer._old_patch_zstorm = queryUtility(IZStorm)
        # Try to reuse the patch_zstorm from the previous run if possible,
        # since we don't want tons of open connections.
        patch_zstorm = getattr(layer, "_patch_zstorm", None)
        if not patch_zstorm:
            layer._patch_zstorm = patch_zstorm = ZStorm()
            for data in self._databases:
                name = data["store"]
                patch_uri = data["patch-uri"]
                patch_zstorm.set_default_uri(name, patch_uri)

            # IZStorm for patches.
            provideUtility(patch_zstorm)

        # Load stores into our internal data structure even if patch_zstorm
        # is already in place.
        for data in self._databases:
            name = data["store"]
            data["patch-store"] = patch_zstorm.get(name)

    def setupSchemaManager(self, layer):
        layer._old_schema_manager = queryUtility(ISchemaManager)
        # Same as for setupPatchStores
        schema_manager = getattr(layer, "_schema_manager", None)
        if not schema_manager:
            layer._old_schema_manager = schema_manager = SchemaManager()
            for data in self._databases:
                name = data["store"]
                schema_package = data["schema-package"]
                schema_manager.setDefaultPackage(name, schema_package)
            provideUtility(schema_manager)

        for data in self._databases:
            name = data["store"]
            data["schema"] = schema_manager.get(name)

    def setupStores(self, layer):
        layer._zstorm = ZStorm()
        layer._zstorm_called = set()

        real_get = layer._zstorm.get

        def get_proxy(name):
            store = real_get(name)
            if not getattr(store, "__commit_tracked__", False):
                # It hasn't been overriden yet
                real_function = store.commit

                def proxy_function():
                    layer._zstorm_called.add(name)
                    real_function()

                store.commit = proxy_function
                store.__commit_tracked__ = True

            return store

        layer._zstorm.get = get_proxy

        for data in self._databases:
            name = data["store"]
            store_uri = data["store-uri"]
            layer._zstorm.set_default_uri(name, store_uri)
            data["store"] = layer._zstorm.get(name)

        # Provide utility for other layers.
        provideUtility(layer._zstorm)

    def setupSchemas(self, layer):
        for data in self._databases:
            store = data["patch-store"]
            schema = data["schema"]
            try:
                schema.upgrade(store)
            except UnknownPatchError:
                transaction.abort()
                store.close()
                layer._patch_zstorm.remove(store)
                name = data["database"]
                self._dropDatabase(name)
                self._createDatabase(name)
                store = layer._patch_zstorm.create(data["store"])
                data["patch-store"] = store
                schema.upgrade(store)

    def deleteSchemas(self, layer, force=False):
        # Rollback any on-going transactions, so that we can delete data.
        transaction.abort()
        zstorm_called = getattr(layer, "_zstorm_called", ())
        for data in self._databases:
            if (force or data["store"] in zstorm_called or
                data["store"].endswith("-autocommit")):

                store = data["patch-store"]
                schema = data["schema"]
                schema.delete(store)


class DatabaseLayerFactory:
    """Sets up PostgreSQL databases for use in tests."""

    __bases__ = ()

    stores = []

    uri_template = "postgres://lpresults:lpresults@localhost/%s"
    patch_uri_template = "postgres:%s"
    schema_package_template = "lpresults.%s.schema"

    def __init__(self, name, database_id=None):
        self.__name__ = name
        if database_id is None:
            database_id = "database-layer"
        self.database_id = database_id

    @property
    def databases(self):
        for store in self.stores:
            tail = store
            if tail.endswith("-autocommit"):
                tail = tail[:-11]

            yield (store, "lpresults-%s-%s" % (self.database_id, tail))

    def getSchemaManager(self):

        def _get_store_uri(name, database):
            uri = self.uri_template % database
            if name.endswith("-autocommit"):
                uri += "?isolation=autocommit"
            return uri

        def _get_patch_uri(database):
            return self.patch_uri_template % database

        def _get_schema_package(name):
            name = name.split("-")[0]
            return self.schema_package_template % name

        manager = DatabaseManager(
            _get_store_uri, _get_patch_uri, _get_schema_package)
        for store, database in self.databases:
            manager.addDatabase(store, database)
        return manager

    def setUp(self):
        manager = self.getSchemaManager()
        manager.setupDatabases()
        manager.setupPatchStores(self)
        manager.setupSchemaManager(self)
        manager.setupSchemas(self)
        manager.deleteSchemas(self, force=True)
        manager.setupStores(self)

    def tearDown(self):
        if self._old_patch_zstorm:
            provideUtility(self._old_patch_zstorm)

        if self._old_schema_manager:
            provideUtility(self._old_schema_manager)

        self._zstorm = None
        self._old_patch_zstorm = None
        self._old_schema_manager = None

        # Signal Layer cannot be torn down fully
        raise NotImplementedError()

    def testSetUp(self):
        provideUtility(self._zstorm)

    def testTearDown(self):
        manager = self.getSchemaManager()
        manager.setupPatchStores(self)
        manager.setupSchemaManager(self)
        manager.deleteSchemas(self)
        # Empty the cache to track only those stores used by the test.
        self._zstorm_called = set()


DatabaseLayer = DatabaseLayerFactory("DatabaseLayer")
