# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import os

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from storm.database import create_database
from storm.locals import (
    Desc,
    Int,
    Store,
    Unicode,
    )

from lpresults.database.resultset import ResultSetDecorator


class Foo:
    __storm_table__ = "foo"

    id = Int(primary=True)
    title = Unicode()


def foo_decorator(foo):
    return "Title is: %s" % foo.title


class TestResultSet(TestCase):

    name = "lpresults-functional-layer-database"

    def setUp(self):
        self.createDatabase()
        self.connection = self.database.connect()
        self.dropTables()
        self.createTables()
        self.createSampleData()
        self.createStore()
        self.result = self.store.find(Foo).order_by(Foo.id)
        self.decorated_result = ResultSetDecorator(self.result, foo_decorator)

    def tearDown(self):
        self.dropStore()
        self.dropSampleData()
        self.dropTables()
        self.dropDatabase()
        self.connection.close()

    def createDatabase(self):
        os.system("createdb %s > /dev/null 2>&1" % self.name)
        self.database = create_database("postgres:%s" % self.name)

    def createSampleData(self):
        connection = self.connection
        connection.execute("INSERT INTO foo (id, title) VALUES (1, 'One')")
        connection.execute("INSERT INTO foo (id, title) VALUES (2, 'Two')")
        connection.commit()

    def createStore(self):
        self.store = Store(self.database)

    def createTables(self):
        connection = self.connection
        connection.execute(
            "CREATE TABLE foo ("
            " id SERIAL PRIMARY KEY,"
            " title VARCHAR DEFAULT 'Default Title')")
        connection.commit()

    def dropDatabase(self):
        pass

    def dropSampleData(self):
        pass

    def dropStore(self):
        self.store.rollback()
        # Closing the store is needed because testcase objects are all
        # instantiated at once, and thus connections are kept open.
        self.store.close()

    def dropTables(self):
        for table in ["foo"]:
            try:
                self.connection.execute("DROP TABLE %s" % table)
                self.connection.commit()
            except:
                self.connection.rollback()

    def test_copy(self):
        result = self.decorated_result
        result_copy = result.copy()
        self.assertNotEquals(result, result_copy)
        self.assertNotEquals(result.result_set, result_copy.result_set)

    def test_config(self):
        self.assertFalse(self.result._distinct)
        returned_result = self.decorated_result.config(distinct=True)
        self.assertTrue(self.result._distinct)
        self.assertEquals(self.decorated_result, returned_result)

    def test_getitem(self):
        self.assertEquals(self.decorated_result[0], "Title is: One")
        self.assertEquals(self.decorated_result[1], "Title is: Two")
        self.assertRaises(IndexError, self.decorated_result.__getitem__, 2)

    def test_any(self):
        orig_result = self.decorated_result.result_set.any()
        self.assertEquals(
            self.decorated_result.any(), foo_decorator(orig_result))

    def test_first(self):
        self.assertEquals(self.decorated_result.first(), "Title is: One")

    def test_last(self):
        self.assertEquals(self.decorated_result.last(), "Title is: Two")

    def test_order_by(self):
        ordered_result = self.decorated_result.order_by(Desc(Foo.id))
        self.assertEquals(ordered_result[0], "Title is: Two")
        self.assertEquals(ordered_result[1], "Title is: One")

    def test_one(self):
        self.assertEquals(
            self.decorated_result.config(limit=1).one(),
            "Title is: One")

    def test_slice(self):
        self.assertTrue(
            isinstance(self.decorated_result[0:1], ResultSetDecorator))


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
