# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""
This module defines classes used by storm to represent PostgreSQL
operators and functions that are used for full text searching.
"""

__metaclass__ = type

__all__ = [
    "TSVectorPart",
    "TSVectorValidator",
    "get_search_query",
    "get_grams_query",
    ]

from storm.expr import (
    Coalesce,
    CompoundOper,
    NamedFunc,
    )

from lpresults.database.textgrams import (
    WORD_SPLIT_PATTERN,
    get_word_grams,
    )


# This is the language configuration for full text search.
TEXT_SEARCH_CONFIG = u"public.lpresults_english"
GRAMS_SEARCH_CONFIG = u"simple"


class TSConcat(CompoundOper):
    """Concatenate multiple strings"""
    __slots__ = ()
    oper = " || "


class TSMatch(CompoundOper):
    """
    The full text match operator is used to compare a compiled search
    (tsquery) expression to a text search vector (tsvector). In
    PostgreSQL, the operator returns a "true" value if the tsvector
    matches the tsquery.
    """
    __slots__ = ()
    oper = " @@ "


class TSQueryAND(CompoundOper):
    """Combine two or more TSQuery objects with an AND operator"""
    __slots__ = ()
    oper = " && "


class TSQueryOR(CompoundOper):
    """Combine two or more TSQuery objects with an OR operator"""
    __slots__ = ()
    oper = " || "


class TSRankCD(NamedFunc):
    """
    This full text search rank is a PostgreSQL function that computes
    the cover density of the search terms in the given text search vector.
    """
    __slots__ = ()
    name = "TS_RANK_CD"


class TSHeadline(NamedFunc):
    """
    This full text search function shows fragments of the document with marked
    search terms.
    """
    __slots__ = ()
    name = "TS_HEADLINE"


class ToTSQuery(NamedFunc):
    """
    This function converts user-written search expressions into a proper
    PostgreSQL full text search query, doing such things as normalizing
    words in the search terms. The correct format produced by this
    function is needed in order to match a text search query against a
    text search vector.
    """
    __slots__ = ()
    name = "TO_TSQUERY"


class PlainToTSQuery(NamedFunc):
    """
    This function transforms unformatted text to a tsquery. The text is
    parsed and normalized much as for to_tsvector, then the AND operator "&"
    is inserted between surviving words.
    """
    __slots__ = ()
    name = "PLAINTO_TSQUERY"


class ToTSVector(NamedFunc):
    """
    Parses and normalizes a document string, converting text into a proper
    PostgreSQL tsquery data type.
    """
    __slots__ = ()
    name = "TO_TSVECTOR"


class SetWeight(NamedFunc):
    """
    Sets the weight of a TSVector.
    """
    __slots__ = ()
    name = "SETWEIGHT"


class Replace(NamedFunc):
    """String replacement."""
    __slots__ = ()
    name = "REPLACE"


class Text(NamedFunc):
    """Cast a given type as text."""
    __slots__ = ()
    name = "TEXT"


class Strip(NamedFunc):
    """Strip a string."""
    __slots__ = ()
    name = "STRIP"


class TSVectorPart:
    """Describes how a model attribute is converted to a document part.

    Each document part can have different weight, and optionally a
    list of callables that should be applied to prepare the attribute
    value for use in a L{TSVector}.

    Valid weight labels are "A", "B", "C", "D", and defaults to "D"
    if none provided.

    Default weights for each one are "1.0", "0.5", "0.2", "0.1", in
    that order.
    """
    def __init__(
        self, name, weight=u"D", transform=None, config=TEXT_SEARCH_CONFIG):

        self.name = name
        self._config = config
        self._weight = weight
        self._transform = transform

    def __call__(self, value=None):
        """
        Transform the value provided, returning a TSVector with the
        specified weight.
        """
        if self._transform is not None:
            for transform in self._transform:
                value = transform(value)
        return SetWeight(
            ToTSVector(self._config, Coalesce(value, u"")), self._weight)


class TSVectorValidator:
    """
    This validator object is a trick; we are not checking values nor
    changing the resultant value of the passed attribute. Instead, we
    are using the validation mechanism to set custom values in the
    database, in particular, values needed in order to perform full
    text search. This obviates the need for a database trigger, and
    ultimately, allows us to do more in Python than the database
    trigger could do easily.
    """
    def __init__(self, name, *parts):
        self._name = name
        self._parts = parts or None

    def __call__(self, model_object, attr=None, value=None):
        vector = self.get(model_object, attr=attr, value=value)
        setattr(model_object, self._name, vector)
        return value

    def get(self, model_object, attr=None, value=None):
        vectors = []

        if self._parts is not None:
            for part in self._parts:
                if part.name == attr:
                    part_value = value
                else:
                    part_value = getattr(model_object, part.name)
                if part_value is not None:
                    vectors.append(part(part_value))

        if vectors:
            return TSConcat(*vectors)
        return ToTSVector("")


def get_search_query(
    search_terms, search_mode="AND", search_config=TEXT_SEARCH_CONFIG):
    """
    This utility function separates the logic for parsing user input. For now,
    this is fairly simple, providing three basic uses, as defined below.

    :param search_terms: this can either be a list of individual words which
        need to be joined to form a search expression or can be a search
        expression itself, this being determined by the search_mode parameter.

    :param search_mode: search_mode can be one of three things:
        OR - all terms are joined with an OR operator "|" to form a search
        expression suitable for ts_query.

        AND - all terms are passed to the PostgreSQL plainto_tsquery function,
        which normalizes the search terms, ANDing the results.

        RAW - the search_terms variable is treated like a search expression,
        ready to be passed to ts_query. In such an expression, multiple search
        terms are combined with "&", "|", and "!" and can be grouped with
        parentheses to enforce precedence. In the absence of parentheses, "!"
        binds most tightly, and "&" binds more tightly than "|".
    """
    if search_mode == "RAW":
        query_function = ToTSQuery
        search_expression = search_terms
    elif search_mode == "OR":
        query_function = ToTSQuery
        # Convert the vector to text.
        text = Text(Strip(ToTSVector(search_terms)))
        # Remove the single quotes that the vector put around terms.
        text = Replace(text, u"'", u"")
        # Insert the OR operator between surviving search terms.
        search_expression = Replace(text, u" ", u" | ")
    elif search_mode == "AND":
        query_function = PlainToTSQuery
        search_expression = search_terms
    return query_function(search_config, search_expression)


def get_grams_query(
    search_terms, search_mode="AND", search_config=GRAMS_SEARCH_CONFIG):
    """Build a TSQuery expression from a plain query string.

    :param search_terms: this can either be a list of individual
        n-grams which need to be joined to form a search expression or
        can be a search expression itself, this being determined by
        the search_mode parameter.

    :param search_mode: search_mode can be one of three things:
        OR - all terms are joined with an OR operator "|" to form a search
        expression suitable for ts_query.

        AND - all terms are passed to the PostgreSQL plainto_tsquery function,
        which normalizes the search terms, ANDing the results.

        RAW - behaves just like AND, for compatibility with get_search_query.

    @return: A TSQuery as a Storm expression that can be used as part
        of a larger Storm query.
    """
    words = WORD_SPLIT_PATTERN.split(search_terms)
    queries = []

    if search_mode == "OR":
        query_function = TSQueryOR
    else:
        query_function = TSQueryAND

    for word in words:
        grams = " ".join(get_word_grams(word))
        queries.append(PlainToTSQuery(search_config, grams))

    return query_function(*queries)
