# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Formatter functions."""

__metaclass__ = type

__all__ = [
    "datetime_to_string",
    "string_to_datetime",
    "item_to_tuple",
    "tuple_to_item",
    "string_to_type",
    ]

import re
import pytz

from datetime import (
    datetime,
    timedelta,
    )


DATETIME_RE = re.compile(r"""
    ^(?P<year>\d\d\d\d)-?(?P<month>\d\d)-?(?P<day>\d\d)
    T(?P<hour>\d\d):?(?P<minute>\d\d):?(?P<second>\d\d)
    (?:\.(?P<second_fraction>\d{0,6}))?
    (?P<tz>
        (?:(?P<tz_sign>[-+])(?P<tz_hour>\d\d):(?P<tz_minute>\d\d))
        | Z)?$
    """, re.VERBOSE)

TYPE_FORMATS = (
    ("(yes|true)", lambda v: True),
    ("(no|false)", lambda v: False),
    ("-?\d+", lambda v: int(v.group(0))),
    ("-?\d+\.\d+", lambda v: float(v.group(0))),
    ("(-?\d+) ?([kmgt]?b?)", lambda v: int(v.group(1))),
    ("(-?\d+\.\d+) ?([kmgt]?b?)", lambda v: float(v.group(1))),
    ("(-?\d+) ?([kmgt]?hz)", lambda v: int(v.group(1))),
    ("(-?\d+\.\d+) ?([kmgt]?hz)", lambda v: float(v.group(1))))
TYPE_FORMATS = tuple(
    (re.compile("^%s$" % pattern, re.IGNORECASE), format)
    for pattern, format in TYPE_FORMATS)

TYPE_MULTIPLIERS = (
    ("b", 1),
    ("kb?", 1024),
    ("mb?", 1024 * 1024),
    ("gb?", 1024 * 1024 * 1024),
    ("tb?", 1024 * 1024 * 1024 * 1024),
    ("hz", 1),
    ("khz?", 1024),
    ("mhz?", 1024 * 1024),
    ("ghz?", 1024 * 1024 * 1024),
    ("thz?", 1024 * 1024 * 1024 * 1024))
TYPE_MULTIPLIERS = tuple(
    (re.compile("^%s$" % pattern, re.IGNORECASE), multiplier)
    for pattern, multiplier in TYPE_MULTIPLIERS)

UTC = pytz.timezone("UTC")


def datetime_to_string(dt):
    """Return a consistent string representation for a given datetime.

    :param dt: The datetime object.
    """
    return dt.isoformat()


def string_to_datetime(string):
    """Return a datetime object from a consistent string representation.

    :param string: The string representation.
    """
    # we cannot use time.strptime: this function accepts neither fractions
    # of a second nor a time zone given e.g. as '+02:30'.
    match = DATETIME_RE.match(string)

    # The Relax NG schema allows a leading minus sign and year numbers
    # with more than four digits, which are not "covered" by _time_regex.
    if not match:
        raise ValueError("Datetime with unreasonable value: %s" % string)

    time_parts = match.groupdict()

    year = int(time_parts['year'])
    month = int(time_parts['month'])
    day = int(time_parts['day'])
    hour = int(time_parts['hour'])
    minute = int(time_parts['minute'])
    second = int(time_parts['second'])
    second_fraction = time_parts['second_fraction']
    if second_fraction is not None:
        milliseconds = second_fraction + '0' * (6 - len(second_fraction))
        milliseconds = int(milliseconds)
    else:
        milliseconds = 0

    # The Relax NG validator accepts leap seconds, but the datetime
    # constructor rejects them. The time values submitted by the HWDB
    # client are not necessarily very precise, hence we can round down
    # to 59.999999 seconds without losing any real precision.
    if second > 59:
        second = 59
        milliseconds = 999999

    dt = datetime(
        year, month, day, hour, minute, second, milliseconds, tzinfo=UTC)

    tz_sign = time_parts['tz_sign']
    tz_hour = time_parts['tz_hour']
    tz_minute = time_parts['tz_minute']
    if tz_sign in ('-', '+'):
        delta = timedelta(hours=int(tz_hour), minutes=int(tz_minute))
        if tz_sign == '-':
            dt = dt + delta
        else:
            dt = dt - delta

    return dt


def item_to_tuple(item):
    """If item is not a tuple, return it as a tuple, otherwise the
    item itself.
    """
    if isinstance(item, tuple):
        return item

    elif isinstance(item, list):
        return tuple(item)

    elif item is None:
        return tuple()

    else:
        return (item,)


def tuple_to_item(item):
    """If item is a tuple, return first element, otherwise the item
    itself.
    """
    if isinstance(item, (tuple, list,)):
        return item[0]

    return item


def string_to_type(string, string_type=None):
    """Return a typed representation for the given string.

    The result might be a bool, int or float. The string might also be
    supplemented by a multiplier like KB which would return an int or
    float multiplied by 1024 for example.

    :param string: The string representation.
    :param string_type: Optional type to cast the result.
    """
    if not isinstance(string, basestring):
        raise TypeError(
            "string_to_type() argument 1 must be string, not %s"
            % type(string))

    for regex, formatter in TYPE_FORMATS:
        match = regex.match(string)
        if match:
            string = formatter(match)
            if len(match.groups()) > 1:
                unit = match.group(2)
                for regex, multiplier in TYPE_MULTIPLIERS:
                    match = regex.match(unit)
                    if match:
                        string *= multiplier
                        break
                else:
                    raise ValueError("Unknown multiplier: %s" % unit)

            break

    if string_type is not None:
        string = string_type(string)

    return string
