# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "OAUTH_REALM",
    "OAUTH_CHALLENGE",
    "OAuthPermission",
    "IOAuthConsumer",
    "IOAuthAccessToken",
    "IOAuthRequestToken",
    "IOAuthDataStore",
    ]

from zope.interface import Interface
from zope.schema import (
    Bool,
    Choice,
    Datetime,
    Object,
    TextLine,
    )

from lazr.enum import (
    DBEnumeratedType,
    DBItem,
    )

from lpresults.registry.interfaces.date import IHasDateCreated
from lpresults.registry.interfaces.person import IPerson


# The challenge included in responses with a 401 status.
OAUTH_REALM = "https://api.launchpad.net"
OAUTH_CHALLENGE = 'OAuth realm="%s"' % OAUTH_REALM


class OAuthPermission(DBEnumeratedType):
    """The permission granted by the user to the OAuth consumer."""

    UNAUTHORIZED = DBItem(10, """
        No Access

        The application will not be allowed to access Launchpad on your
        behalf.
        """)

    WRITE_PRIVATE = DBItem(50, """
        Change Anything

        The application will be able to access Launchpad on your behalf
        for reading and changing anything, including private data.
        """)


class IOAuthConsumer(IHasDateCreated):
    """An application which acts on behalf of a Launchpad user."""

    key = TextLine(
        title=u"Key",
        required=True,
        readonly=True,
        description=u"The unique key which identifies a consumer. It is "
                     "included by the consumer in each request made.")

    secret = TextLine(
        title=u"Secret",
        required=False,
        readonly=False,
        description=u"The secret which, if not empty, should be used by the "
                     "consumer to sign its requests.")

    def getAccessToken(key):
        """Return the `IOAuthAccessToken` with the given key.

        If the token with the given key does not exist or is associated with
        another consumer, return None.
        """

    def getRequestToken(key):
        """Return the `IOAuthRequestToken` with the given key.

        If the token with the given key does not exist or is associated with
        another consumer, return None.
        """

    def createRequestToken(key, secret):
        """Return a new `IOAuthRequestToken` with a random key and secret."""


class IOAuthToken(IHasDateCreated):
    """Base class for `IOAuthAccessToken` and `IOAuthRequestToken`.

    This class contains the commonalities of the two token classes we actually
    care about and shall not be used on its own.
    """
    consumer = Object(
        schema=IOAuthConsumer,
        title=u"Consumer",
        description=u"The consumer which will access Launchpad on the "
                      "user's behalf.")

    person = Object(
        schema=IPerson,
        title=u"Person",
        required=False,
        readonly=False,
        description=u"The user on whose behalf the consumer is accessing.")

    key = TextLine(
        title=u"Key",
        required=True,
        readonly=True,
        description=u"The key used to identify this token.  It is included "
                     "by the consumer in each request.")

    secret = TextLine(
        title=u"Secret",
        required=True,
        readonly=True,
        description=u"The secret associated with this token.  It is used "
                     "by the consumer to sign its requests.")


class IOAuthAccessToken(IOAuthToken):
    """A token used by a consumer to access protected resources in LP.

    It's created automatically once a user logs in and grants access to a
    consumer.  The consumer then exchanges an `IOAuthRequestToken` for it.
    """
    permission = Choice(
        title=u"Permission",
        required=True,
        readonly=False,
        vocabulary=OAuthPermission,
        description=u"The permission given to the application acting "
                     "on your behalf.")


class IOAuthRequestToken(IOAuthToken):
    """A token used by a consumer to ask the user to authenticate on LP.

    After the user has authenticated and granted access to that consumer, the
    request token is exchanged for an access token and is then destroyed.
    """
    permission = Choice(
        title=u"Permission",
        required=True,
        readonly=False,
        vocabulary=OAuthPermission,
        description=u"The permission you give to the application which may "
                     "act on your behalf.")

    date_reviewed = Datetime(
        title=u"Date reviewed",
        required=True,
        readonly=True,
        description=u"The date in which the user authorized (or not) the "
                     "consumer to access his protected resources on "
                     "Launchpad.")

    is_reviewed = Bool(
        title=u"Has this token been reviewed?",
        required=False,
        readonly=True,
        description=u"A reviewed request token can only be exchanged for an "
                     "access token (in case the user granted access).")

    def review(user, permission):
        """Grant `permission` as `user` to this token's consumer.

        Set this token's person, permission and date_reviewed.  This will also
        cause this token to be marked as used, meaning it can only be
        exchanged for an access token with the same permission, consumer and
        person.
        """

    def createAccessToken():
        """Create an `IOAuthAccessToken` identical to this request token.

        After the access token is created, this one is deleted as it can't be
        used anymore.

        You must not attempt to create an access token if the request token
        hasn't been reviewed or if its permission is UNAUTHORIZED.
        """


class IOAuthDataStore(Interface):
    """A database abstraction used to lookup consumers and tokens."""

    def createConsumer(consumer_key, secret=""):
        """Return the newly created consumer.

        You must make sure the given `key` is not already in use by another
        consumer before trying to create a new one.

        The `secret` defaults to an empty string because most consumers will
        be open source desktop applications for which it wouldn't be actually
        secret.

        :param consumer_key: The unique key which will be associated with the
            new consumer.
        :param secret: A secret which should be used by the consumer to sign
            its requests.
        """

    def lookupConsumer(consumer_key):
        """Return the consumer with the given key.

        If there's no consumer with the given key, return None.

        :param key: The unique key associated with a consumer.
        """

    def lookupRequest(request_key):
        """Return the request token with the given key.

        If there's no request token with the given key, return None.

        :param key: The unique key associated with a request token.
        """

    def lookupToken(self, token_type, token_field):
        """Return the access token of the given type and field.

        If there's no access token with the given key, return None.

        :param token_type: type of token to lookup
        :param token_field: token to look up

        :note: token_type should always be 'access' as only such tokens are
               stored in database
        """

    def lookupNonce(self, oauth_consumer, oauth_token, nonce):
        """Always return None."""
