# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "IPerson",
    "IPersonSet",
    "IHasPerson",
    "PersonParamsMixin",
    "PersonSearchParams",
    ]

from zope.interface import (
    Attribute,
    Interface,
    )
from zope.schema import (
    Object,
    TextLine,
    )

from lazr.restful.declarations import exported
from lazr.restful.frameworks.django import IDjangoLocation

from lpresults.registry.interfaces.search import (
    ISearchSet,
    SearchParams,
    )
from lpresults.registry.validators import name_validator


class IPerson(IDjangoLocation):
    """A Person.

    This maps to the concept of Person in the Launchpad Registry.
    """
    id = Attribute("A unique identifier for the person")

    oauth_access_tokens = Attribute("Non-expired access tokens")

    oauth_request_tokens = Attribute("Non-expired request tokens")

    name = exported(
        TextLine(
            title=u"Name",
            required=True,
            readonly=True,
            constraint=name_validator,
            description=u"A short name for the person"))


class IPersonSet(IDjangoLocation, ISearchSet):
    """The set of person."""

    def create(name, identity):
        """Return the newly created person.

        You must make sure the given `identity` is not already in use
        by another person before trying to create a new one.

        :param name: The person's name.
        :param identity: The identity URL which will be associated with
            the new person.
        """

    def get(id):
        """Return the person with the given id.

        If there's no person with the given id, return None.

        :param id: The primary id associated with a person.
        """

    def getByIdentity(identity):
        """Return the person with the given identity.

        If there's no person with the given identity, return None.

        :param identity: The unique identity associated with a person.
        """

    def getByName(name):
        """Return the person with the given name.

        If there's no person with the given name, return None.

        :param name: The unique name associated with a person.
        """

    def search():
        """Return the default set of people without filtering."""


class IHasPerson(Interface):

    person = Object(
        schema=IPerson,
        title=u"Person",
        required=True,
        readonly=True,
        description=u"The person associated with this object.")


class PersonParamsMixin:

    person = None

    def setPerson(self, person=None):
        if person:
            self.person = person


class PersonSearchParams(SearchParams):
    """Encapsulates search parameters for IPersonSet.search()"""

    def __init__(self, name=None, identity=None, **kwargs):
        super(PersonSearchParams, self).__init__(**kwargs)
        self.name = name
        self.identity = identity
