# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "IHasSeries",
    "ISeries",
    "ISeriesSet",
    "ISeriesTarget",
    ]

from zope.interface import Interface
from zope.schema import TextLine

from lazr.restful.declarations import (
    export_operation_as,
    export_read_operation,
    exported,
    operation_parameters,
    operation_returns_entry,
    )
from lazr.restful.fields import (
    CollectionField,
    Reference,
    )
from lazr.restful.frameworks.django import IDjangoLocation

from lpresults.registry.interfaces.search import (
    ISearchSet,
    SearchParams,
    )
from lpresults.registry.validators import name_validator


class ISeries(IDjangoLocation):
    """A series of releases. For example '2.0' or '1.3' or 'dev'."""

    name = exported(
        TextLine(
            title=u"Name",
            required=True,
            readonly=True,
            constraint=name_validator,
            description=u"A short name for the series"))


class ISeriesSet(ISearchSet):

    def new(name):
        """Create a new series.

        :param name: The unique name of the series.
        """

    def get(id):
        """Get a series by its id.

        :param id: The identifier of the series.
        """

    def getByName(name):
        """Get a series by name.

        :param name: The name of the series.
        """


class IHasSeries(Interface):
    """An entity which has a collection of series."""

    series = exported(
        CollectionField(
            title=u"List of series.",
            value_type=Reference(schema=ISeries),
            readonly=True))

    def getSeries(name):
        """Return the series for the given name, or None."""


class ISeriesTarget(IHasSeries):
    """An entity on which a series can be created."""

    def newSeries(name):
        """Return a new series.

        :param name: The unique name of this series.
        """

    @operation_parameters(
        name=TextLine(
            title=u"Name",
            required=True),
        )
    @operation_returns_entry(ISeries)
    @export_operation_as("getSeries")
    @export_read_operation()
    def retrieveSeries(name):
        """Return the series for the given name, or a new series
        retrieved from Launchpad.

        :param name: The name of the series to retrieve.
        """


class SeriesSearchParams(SearchParams):
    """Encapsulates search parameters for ISeriesSet.search()"""

    def __init__(self, name=None, **kwargs):
        super(SeriesSearchParams, self).__init__(**kwargs)
        self.name = name
