# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Distribution",
    "distribution_set",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Reference,
    Unicode,
    )

from zope.component import getUtility
from zope.event import notify
from zope.interface import implements

from lazr.lifecycle.event import ObjectCreatedEvent
from lazr.restful.interfaces import IServiceRootResource

from lpresults.database.search import search_to_expression

from lpresults.registry.interfaces.distribution import (
    IDistribution,
    IDistributionSet,
    DistributionSearchParams,
    )
from lpresults.registry.models.distroseries import DistroSeries
from lpresults.registry.models.pillar import Pillar
from lpresults.registry.models.search import SearchSet
from lpresults.registry.models.series import SeriesTargetMixin
from lpresults.registry.models.store import get_registry_store


class Distribution(Pillar, SeriesTargetMixin):

    implements(IDistribution)

    __storm_table__ = "distribution"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    # While the interface defines this field as required, we need to
    # allow it to be NULL so we can create new distribution records
    # before the corresponding series records.
    current_series_id = Int(name="current_series")

    current_series = Reference(current_series_id, DistroSeries.id)


class DistributionSet(SearchSet):
    """See `IDistributionSet`."""

    implements(IDistributionSet)

    search_params_factory = DistributionSearchParams

    __url_path__ = "distributions"

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    def create(self, name, series_name):
        """See `IDistributionSet`."""
        distribution = Distribution(name)

        # This will result in an exception being raised if the distribution
        # name is a duplicate.
        store = get_registry_store()
        store.add(distribution)
        store.flush()

        # Create a default series and set it as the current series
        series = distribution.newSeries(series_name)
        distribution.current_series = series

        notify(ObjectCreatedEvent(distribution))

        # This is for the benefit of derived classes
        return self.get(distribution.id)

    def get(self, id):
        """See `IDistributionSet`."""
        return get_registry_store().get(Distribution, id)

    def getByName(self, name):
        """See `IDistributionSet`."""
        search_params = self.search_params_factory(name=name)
        return self.search(search_params).one()

    def search(self, *args):
        if not args:
            search_params = self.search_params_factory()
            args = (search_params,)

        return super(DistributionSet, self).search(*args)

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(DistributionSet, self).buildSearchExpressions(
            search_params)

        if search_params.name is not None:
            expressions.append(
                search_to_expression(Distribution.name, search_params.name))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return Distribution

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return get_registry_store()


distribution_set = DistributionSet()

DistroSeries.distribution = Reference(
    DistroSeries.distribution_id, Distribution.id)
