# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "oauth_data_store",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Reference,
    ReferenceSet,
    Store,
    Unicode,
    )

from zope.interface import implements

from lpresults.database.constants import UTC_NOW
from lpresults.database.datetime import UTCDateTime
from lpresults.database.enum import DBEnum

from lpresults.registry.interfaces.oauth import (
    OAuthPermission,
    IOAuthConsumer,
    IOAuthAccessToken,
    IOAuthRequestToken,
    IOAuthDataStore,
    )
from lpresults.registry.models.person import Person
from lpresults.registry.models.store import get_registry_store


class OAuthConsumer:
    """See `IOAuthConsumer`."""
    implements(IOAuthConsumer)

    __storm_table__ = "oauthconsumer"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    key = Unicode(allow_none=False)
    secret = Unicode(allow_none=False)
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)

    # access_token reference defined after OAuthAccessToken

    def __init__(self, key, secret):
        self.key = key
        self.secret = secret

    def getAccessToken(self, key):
        """See `IOAuthConsumer`."""
        store = Store.of(self)
        return store.find(OAuthAccessToken, consumer=self, key=key).one()

    def getRequestToken(self, key):
        """See `IOAuthConsumer`."""
        store = Store.of(self)
        return store.find(OAuthRequestToken, consumer=self, key=key).one()

    def createRequestToken(self, key, secret):
        """See `IOAuthConsumer`."""
        store = Store.of(self)
        request_token = OAuthRequestToken(self, key, secret)
        store.add(request_token)
        return request_token


class OAuthToken:
    """Compatibility layer with the Python OAuth implementation."""

    callback = None
    verifier = None


class OAuthAccessToken(OAuthToken):
    """See `IOAuthAccessToken`."""
    implements(IOAuthAccessToken)

    __storm_table__ = "oauthaccesstoken"

    key = Unicode(primary=True, allow_none=False)
    secret = Unicode(allow_none=False)
    consumer_id = Int(name="consumer", allow_none=False)
    person_id = Int(name="person", allow_none=False)
    permission = DBEnum(allow_none=False, enum=OAuthPermission)
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)

    consumer = Reference(consumer_id, OAuthConsumer.id)
    person = Reference(person_id, Person.id)

    def __init__(self, consumer, person, permission, key, secret):
        self.consumer = consumer
        self.person = person
        self.permission = permission
        self.key = key
        self.secret = secret


OAuthConsumer.access_tokens = ReferenceSet(OAuthConsumer.id,
    OAuthAccessToken.consumer_id)


class OAuthRequestToken(OAuthToken):
    """See `IOAuthRequestToken`."""
    implements(IOAuthRequestToken)

    __storm_table__ = "oauthrequesttoken"

    key = Unicode(primary=True, allow_none=False)
    secret = Unicode(allow_none=False)
    consumer_id = Int(name="consumer", allow_none=False)
    person_id = Int(name="person")
    permission = DBEnum(enum=OAuthPermission)
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    date_reviewed = UTCDateTime()

    consumer = Reference(consumer_id, OAuthConsumer.id)
    person = Reference(person_id, Person.id)

    def __init__(self, consumer, key, secret):
        self.consumer = consumer
        self.key = key
        self.secret = secret

    @property
    def is_reviewed(self):
        """See `IOAuthRequestToken`."""
        return self.date_reviewed is not None

    def review(self, user, permission):
        """See `IOAuthRequestToken`."""
        self.date_reviewed = UTC_NOW
        self.person = user
        self.permission = permission

    def createAccessToken(self, key, secret):
        """See `IOAuthRequestToken`."""
        assert self.is_reviewed, (
            "Cannot create an access token from an unreviewed request token.")
        assert self.permission != OAuthPermission.UNAUTHORIZED, (
            "The user did not grant access to this consumer.")

        access_token = OAuthAccessToken(
            self.consumer, self.person, self.permission, key, secret)
        store = Store.of(self)
        store.add(access_token)
        store.remove(self)

        return access_token


class OAuthDataStore:
    """See `IOAuthDataStore`."""
    implements(IOAuthDataStore)

    def createConsumer(self, consumer_key, secret=u""):
        """See `IOAuthDataStore`."""
        assert self.lookupConsumer(consumer_key) is None, (
            "The key '%s' is already in use by another consumer."
            % consumer_key)

        key = unicode(consumer_key)
        consumer = OAuthConsumer(key, secret)
        get_registry_store().add(consumer)

        return consumer

    def lookupConsumer(self, consumer_key):
        """See `IOAuthDataStore`."""
        key = unicode(consumer_key)
        return get_registry_store().find(OAuthConsumer, key=key).one()

    def lookupRequest(self, request_key):
        """See `IOAuthDataStore`."""
        key = unicode(request_key)
        return get_registry_store().get(OAuthRequestToken, key)

    def lookupToken(self, token_type, token_field="access"):
        """See `IOAuthDataStore`."""
        assert token_type == "access"

        key = unicode(token_field)
        return get_registry_store().get(OAuthAccessToken, key)

    def lookupNonce(self, oauth_consumer, oauth_token, nonce):
        """See `IOAuthDataStore`."""
        return None

    # Compatibility layer with oauth.oauth
    lookup_consumer = lookupConsumer
    lookup_token = lookupToken
    lookup_nonce = lookupNonce


oauth_data_store = OAuthDataStore()
