# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Person",
    "person_set",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Store,
    Unicode,
    )

from zope.component import getUtility
from zope.event import notify
from zope.interface import implements

from lazr.lifecycle.event import ObjectCreatedEvent
from lazr.restful.interfaces import IServiceRootResource

from lpresults.database.search import search_to_expression

from lpresults.registry.interfaces.person import (
    IPerson,
    IPersonSet,
    PersonSearchParams,
    )
from lpresults.registry.models.search import SearchSet
from lpresults.registry.models.store import get_registry_store


class Person:
    """Users within the Django authentication system are represented by
    this model.
    """
    implements(IPerson)

    __storm_table__ = "person"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    identity = Unicode()

    def __init__(self, name, identity=None):
        self.name = name
        self.identity = identity

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    @property
    def __url_path__(self):
        return "~%s" % self.name

    @property
    def oauth_access_tokens(self):
        """See `IPerson`."""
        from lpresults.registry.models.oauth import OAuthAccessToken

        return Store.of(self).find(
            OAuthAccessToken,
            OAuthAccessToken.person == self)

    @property
    def oauth_request_tokens(self):
        """See `IPerson`."""
        from lpresults.registry.models.oauth import OAuthRequestToken

        return Store.of(self).find(
            OAuthRequestToken,
            OAuthRequestToken.person == self)

    def isAnonymous(self):
        """Always returns False. This is a way of comparing Person
        objects to anonymous users.
        """
        return False

    def isAuthenticated(self):
        """Always return True. This is a way to tell if the user has
        been authenticated in templates.
        """
        return True

    def save(self):
        # XXX This should be handled by the ORM.
        pass

    # Compatibility layer with Django
    is_anonymous = isAnonymous
    is_authenticated = isAuthenticated


class PersonSet(SearchSet):
    """See `IPersonSet`."""

    implements(IPersonSet)

    search_params_factory = PersonSearchParams

    __url_path__ = "people"

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    def create(self, name, identity=None):
        """See `IPersonSet`."""
        person = Person(name, identity)

        # This will result in an exception being raised if the person
        # name is a duplicate.
        store = get_registry_store()
        store.add(person)
        store.flush()

        notify(ObjectCreatedEvent(person))

        # This is for the benefit of derived classes
        return self.get(person.id)

    def get(self, id):
        """See `IPersonSet`."""
        return get_registry_store().get(Person, id)

    def getByIdentity(self, identity):
        """See `IPersonSet`."""
        search_params = self.search_params_factory(identity=identity)
        return self.search(search_params).one()

    def getByName(self, name):
        """See `IPersonSet`."""
        search_params = self.search_params_factory(name=name)
        return self.search(search_params).one()

    def search(self, *args):
        if not args:
            search_params = self.search_params_factory()
            args = (search_params,)

        return super(PersonSet, self).search(*args)

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(PersonSet, self).buildSearchExpressions(
            search_params)

        if search_params.name is not None:
            expressions.append(
                search_to_expression(Person.name, search_params.name))

        if search_params.identity is not None:
            expressions.append(
                search_to_expression(Person.identity, search_params.identity))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return Person

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return get_registry_store()


person_set = PersonSet()
