# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Launchpad Pillars share a namespace.

Pillars are currently Project and Distribution.
"""

__metaclass__ = type

__all__ = [
    "Pillar",
    "PillarName",
    "PillarSet",
    "pillar_set",
    ]

from os.path import basename
from urlparse import urlparse

from storm.expr import LeftJoin
from storm.locals import (
    AutoReload,
    Int,
    Unicode,
    )

from zope.component import getUtility
from zope.interface import implements

from lazr.restful.interfaces import IServiceRootResource

from lpresults.database.search import search_to_expression

from lpresults.registry.interfaces.distribution import IDistributionSet
from lpresults.registry.interfaces.launchpad import ILaunchpad
from lpresults.registry.interfaces.pillar import (
    IPillar,
    IPillarName,
    IPillarSet,
    PillarSearchParams,
    )
from lpresults.registry.interfaces.project import IProjectSet
from lpresults.registry.models.search import SearchSet
from lpresults.registry.models.store import get_registry_store


class Pillar:

    implements(IPillar)

    def __init__(self, name):
        self.name = name

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    @property
    def __url_path__(self):
        return self.name


class PillarName:

    implements(IPillarName)

    __storm_table__ = "pillarname"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    distribution_id = Int(name="distribution")
    project_id = Int(name="project")

    @property
    def pillar(self):
        if self.distribution_id is not None:
            return getUtility(IDistributionSet).get(self.distribution_id)
        elif self.project_id is not None:
            return getUtility(IProjectSet).get(self.project_id)
        else:
            raise AssertionError("Unknown pillar type: %s" % self.name)


class PillarSet(SearchSet):
    """See `IPillarSet`."""

    implements(IPillarSet)

    search_params_factory = PillarSearchParams

    # See ITopLevelEntryLink
    link_name = "pillars"
    entry_type = IPillarSet

    __url_path__ = "pillars"

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    def createDistribution(self, *args, **kwargs):
        """See `IPillarSet`."""
        return getUtility(IDistributionSet).create(*args, **kwargs)

    def createProject(self, *args, **kwargs):
        """See `IPillarSet`."""
        return getUtility(IProjectSet).create(*args, **kwargs)

    def getByName(self, name):
        """See `IPillarSet`."""
        search_params = self.search_params_factory(name=name)
        return self.search(search_params).one()

    def retrieveByName(self, name):
        """See `IPillarSet`."""
        pillar = self.getByName(name)
        if pillar is None:
            launchpad_pillar = getUtility(ILaunchpad).getPillar(name)
            if launchpad_pillar is not None:
                resource_type_link = launchpad_pillar["resource_type_link"]
                resource_type = urlparse(resource_type_link)[5]
                if resource_type == "distribution":
                    series_link = launchpad_pillar["current_series_link"]
                    series_name = basename(urlparse(series_link)[2])
                    pillar = self.createDistribution(name, series_name)
                elif resource_type == "project":
                    series_link = launchpad_pillar["development_focus_link"]
                    series_name = basename(urlparse(series_link)[2])
                    pillar = self.createProject(name, series_name)
                else:
                    raise AssertionError("Unknown pillar type: %s" % name)

        return pillar

    def search(self, *args):
        if not args:
            search_params = self.search_params_factory()
            args = (search_params,)

        return super(PillarSet, self).search(*args)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        from lpresults.registry.models.distribution import Distribution
        from lpresults.registry.models.project import Project

        tables = super(PillarSet, self).buildSearchTables(
            search_params)

        tables.extend([
            LeftJoin(
                Distribution,
                Distribution.id == PillarName.distribution_id),
            LeftJoin(
                Project,
                Project.id == PillarName.project_id),
            ])

        return tables

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(PillarSet, self).buildSearchExpressions(
            search_params)

        if search_params.name is not None:
            expressions.append(search_to_expression(
                PillarName.name, search_params.name, lambda x: x.lower()))

        return expressions

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return result[0].pillar

    def getSearchResultTables(self):
        """See SearchSet."""
        from lpresults.registry.models.distribution import Distribution
        from lpresults.registry.models.project import Project

        return (PillarName, Distribution, Project,)

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return get_registry_store()


pillar_set = PillarSet()
