# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Project",
    "project_set",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Reference,
    Unicode,
    )

from zope.component import getUtility
from zope.event import notify
from zope.interface import implements

from lazr.lifecycle.event import ObjectCreatedEvent
from lazr.restful.interfaces import IServiceRootResource

from lpresults.database.search import search_to_expression

from lpresults.registry.interfaces.project import (
    IProject,
    IProjectSet,
    ProjectSearchParams,
    )
from lpresults.registry.models.pillar import Pillar
from lpresults.registry.models.projectseries import ProjectSeries
from lpresults.registry.models.search import SearchSet
from lpresults.registry.models.series import SeriesTargetMixin
from lpresults.registry.models.store import get_registry_store


class Project(Pillar, SeriesTargetMixin):

    implements(IProject)

    __storm_table__ = "project"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    # While the interface defines this field as required, we need to
    # allow it to be NULL so we can create new project records before
    # the corresponding series records.
    development_focus_id = Int(name="development_focus")

    development_focus = Reference(development_focus_id, ProjectSeries.id)


class ProjectSet(SearchSet):
    """See `IProjectSet`."""

    implements(IProjectSet)

    search_params_factory = ProjectSearchParams

    __url_path__ = "projects"

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    def create(self, name, series_name):
        """See `IProjectSet`."""
        project = Project(name)

        # This will result in an exception being raised if the project
        # name is a duplicate.
        store = get_registry_store()
        store.add(project)
        store.flush()

        # Create a default series and set it as the development focus
        series = project.newSeries(series_name)
        project.development_focus = series

        notify(ObjectCreatedEvent(project))

        # This is for the benefit of derived classes
        return self.get(project.id)

    def get(self, id):
        """See `IProjectSet`."""
        return get_registry_store().get(Project, id)

    def getByName(self, name):
        """See `IProjectSet`."""
        search_params = self.search_params_factory(name=name)
        return self.search(search_params).one()

    def search(self, *args):
        if not args:
            search_params = self.search_params_factory()
            args = (search_params,)

        return super(ProjectSet, self).search(*args)

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(ProjectSet, self).buildSearchExpressions(
            search_params)

        if search_params.name is not None:
            expressions.append(
                search_to_expression(Project.name, search_params.name))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return Project

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return get_registry_store()


project_set = ProjectSet()

ProjectSeries.project = Reference(ProjectSeries.project_id, Project.id)
