# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "SearchSet",
    ]

from storm.expr import Expr
from storm.info import get_cls_info
from storm.properties import PropertyColumn
from storm.locals import (
    Asc,
    Desc,
    )
from storm.store import EmptyResultSet

from lpresults.registry.formatters import (
    item_to_tuple,
    tuple_to_item,
    )

from lpresults.database.resultset import ResultSetDecorator


class SearchSet:

    def search(self, *args):
        """See `ISearchSet`."""
        results = EmptyResultSet()
        for arg in args:
            results = results.union(self.buildSearchResult(arg))

        return self.buildSearchResultDecorator(results)

    def buildSearchResult(self, search_params):
        """Build and return a search result with the given parameters.

        :param search_params: A SearchParams object.
        """
        tables, expressions, order_by = self.buildSearchQuery(search_params)

        store = self.getSearchStore(search_params)
        if tables:
            store = store.using(*tables)

        query_tables = self.getSearchResultTables()
        result = store.find(query_tables, *expressions)

        if order_by:
            result = result.order_by(order_by)

        return result

    def buildSearchQuery(self, search_params):
        """Build and return a search query consisting of tables,
        expressions and an order by clause.

        :param search_params: A SearchParams object.
        """
        tables = self.buildSearchTables(search_params)
        expressions = self.buildSearchExpressions(search_params)

        if search_params.search_text:
            expressions.extend(
                self.buildSearchText(search_params.search_text))

        order_by = self.buildSearchOrderBy(search_params.order_by)

        return tables, expressions, order_by

    def buildSearchExpressions(self, search_params):
        """Build and return the search expressions for the given
        parameters.

        By default, return an empty list of expressions.

        :param search_params: A SearchParams object.
        """
        return []

    def buildSearchTables(self, search_params):
        """Build and return the search tables for the given
        parameters.

        By default, return the first item from `getSearchResultTables`.

        :param search_params: A SearchParams object.
        """
        return [tuple_to_item(self.getSearchResultTables())]

    def buildSearchText(self, search_text):
        """Build and return the expressions for the given search_text.

        By default, return an empty list of expressions.
        """
        return []

    def buildSearchOrderBy(self, order_by):
        """Build and return the columns for the given order_by."""
        columns = []
        for name in item_to_tuple(order_by):
            if name.startswith("-"):
                column = Desc(self.getSearchOrderByColumn(name[1:]))

            else:
                column = Asc(self.getSearchOrderByColumn(name))

            if column:
                columns.append(column)

        return columns

    def buildSearchResultDecorator(self, results):
        """Build and return the decorator for a search result."""
        if not isinstance(results, EmptyResultSet):
            decorator = self.getSearchResultDecorator
            results = ResultSetDecorator(results, decorator)

        return results

    def getSearchOrderByColumn(self, name):
        """Get the table column for the given name.

        If the name is not recognized, a KeyError is raised.
        """
        for table in item_to_tuple(self.getSearchResultTables()):
            if isinstance(table, type):
                attributes = get_cls_info(table).attributes
                if name in attributes:
                    return attributes[name]

            elif isinstance(table, Expr):
                if name == table.name:
                    return table

            elif isinstance(table, PropertyColumn):
                return table.name

        raise KeyError(name)

    def getSearchResultDecorator(self, result):
        """Get the decorator for a search result."""
        return tuple_to_item(result)

    def getSearchResultTables(self):
        """Get the tables to return in a search result."""
        raise NotImplementedError()

    def getSearchStore(self, search_aprams):
        """Get the store to search."""
        raise NotImplementedError()
