# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import pytz

from datetime import datetime

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from lpresults.registry.formatters import (
    datetime_to_string,
    string_to_datetime,
    item_to_tuple,
    tuple_to_item,
    string_to_type,
    )


UTC = pytz.timezone("UTC")


class TestDatetimeToString(TestCase):

    def test_none(self):
        self.assertRaises(
            AttributeError,
            datetime_to_string,
            None)

    def test_datetime(self):
        dt = datetime(2010, 1, 1, tzinfo=UTC)
        string = datetime_to_string(dt)
        self.assertNotEquals(string, None)
        self.assertEquals(string, "2010-01-01T00:00:00+00:00")


class TestStringToDatetime(TestCase):

    def test_none(self):
        self.assertRaises(
            TypeError,
            string_to_datetime,
            None)

    def test_short(self):
        string = "20100101T000000"
        dt = string_to_datetime(string)
        self.assertNotEquals(dt, None)
        self.assertEquals(dt, datetime(2010, 1, 1, tzinfo=UTC))

    def test_long(self):
        string = "2010-01-01T00:00:00.0+00:00"
        dt = string_to_datetime(string)
        self.assertNotEquals(dt, None)
        self.assertEquals(dt, datetime(2010, 1, 1, tzinfo=UTC))


class TestItemToTuple(TestCase):

    def test_item(self):
        self.assertEquals(item_to_tuple(None), tuple())
        self.assertEquals(item_to_tuple(0), (0,))
        self.assertEquals(item_to_tuple("foo"), ("foo",))

    def test_list(self):
        self.assertEquals(item_to_tuple([None]), (None,))
        self.assertEquals(item_to_tuple([0]), (0,))
        self.assertEquals(item_to_tuple(["foo"]), ("foo",))

    def test_tuple(self):
        self.assertEquals(item_to_tuple((None,)), (None,))
        self.assertEquals(item_to_tuple((0,)), (0,))
        self.assertEquals(item_to_tuple(("foo",)), ("foo",))


class TestTupleToItem(TestCase):

    def test_item(self):
        self.assertEquals(tuple_to_item(None), None)
        self.assertEquals(tuple_to_item(0), 0)
        self.assertEquals(tuple_to_item("foo"), "foo")

    def test_list(self):
        self.assertEquals(tuple_to_item([None]), None)
        self.assertEquals(tuple_to_item([0]), 0)
        self.assertEquals(tuple_to_item(["foo"]), "foo")

    def test_tuple(self):
        self.assertEquals(tuple_to_item((None,)), None)
        self.assertEquals(tuple_to_item((0,)), 0)
        self.assertEquals(tuple_to_item(("foo",)), "foo")


class TestStringToType(TestCase):

    def test_boolTrue(self):
        self.assertEquals(string_to_type("true"), True)
        self.assertEquals(string_to_type("yes"), True)

    def test_boolFalse(self):
        self.assertEquals(string_to_type("false"), False)
        self.assertEquals(string_to_type("no"), False)

    def test_int(self):
        self.assertEquals(string_to_type("0"), 0)
        self.assertEquals(string_to_type("1"), 1)
        self.assertEquals(string_to_type("-1"), -1)

    def test_intBytes(self):
        self.assertEquals(string_to_type("0b"), 0)
        self.assertEquals(string_to_type("1b"), 1)
        self.assertEquals(string_to_type("-1 b"), -1)
        self.assertEquals(string_to_type("1 kb"), 1024)

    def test_intHertz(self):
        self.assertEquals(string_to_type("0hz"), 0)
        self.assertEquals(string_to_type("1hz"), 1)
        self.assertEquals(string_to_type("-1 hz"), -1)
        self.assertEquals(string_to_type("1 khz"), 1024)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
