# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "ProberFactory",
    "ProberFactoryRedirect",
    "ProberManager",
    ]

import base64
import httplib
import logging

from os import getenv
from email.Utils import formatdate
from time import mktime
from urlparse import urlparse

from bz2 import decompress as bz2_decompress
from zlib import decompress as zlib_decompress

from twisted.internet import (
    defer,
    reactor,
    )
from twisted.internet.defer import DeferredSemaphore
from twisted.internet.protocol import ClientFactory
from twisted.internet.ssl import ClientContextFactory
from twisted.python.failure import Failure
from twisted.web.http import HTTPClient


class ProberError(Exception):
    """A generic prober error.

    This class should be used as a base for more specific prober errors.
    """


class ProberTimeout(ProberError):
    """The initialized URL did not return in time."""

    def __init__(self, url, timeout, *args):
        self.url = url
        self.timeout = timeout
        ProberError.__init__(self, *args)

    def __str__(self):
        return (
            "GET request on %s took longer than %s seconds"
            % (self.url, self.timeout))


class ProberResponseError(ProberError):

    def __init__(self, status, *args):
        ProberError.__init__(self, *args)
        self.status = status

    def __str__(self):
        return "Bad response code: %s" % self.status


class ProberLoopError(ProberError):

    def __str__(self):
        return "Infinite loop detected"


class ProberSkipError(ProberError):

    def __str__(self):
        return (
            "Connection skipped because of too many timeouts on this "
            "host. It will be retried on the next probing run.")


class ProberSchemeError(ProberError):

    def __init__(self, url, *args):
        ProberError.__init__(self, *args)
        self.url = url

    def __str__(self):
        return ("The prober doesn't know how to check this kind of URL: %s"
            % self.url)


class ProberProtocol(HTTPClient):
    """Simple HTTP client to probe a path via GET."""

    def connectionMade(self):
        """Simply requests path presence."""
        self.makeRequest()
        self.headers = {}

    def makeRequest(self):
        """Request a path via HTTP/1.1 using GET.

        Uses factory.connect_path and factory.connect_host
        """
        self.sendCommand("GET", self.factory.connect_path)
        self.sendHeader("Host", self.factory.connect_host)

        # If the http_proxy variable is set, we don't want the proxy to
        # cache information being probed.
        if getenv("http_proxy"):
            self.sendHeader("Cache-Control", "no-cache")
            self.sendHeader("Pragma", "no-cache")

        if self.factory.connect_username:
            auth = base64.encodestring("%s:%s" % (
                self.factory.connect_username, self.factory.connect_password))
            self.sendHeader("Authorization", "Basic %s" % auth)

        if self.factory.connect_date_modified:
            timeval = mktime(self.factory.connect_date_modified.timetuple())
            modified_since = formatdate(
                timeval=timeval, localtime=False, usegmt=True)
            self.sendHeader("If-Modified-Since", modified_since)

        self.endHeaders()

    def handleStatus(self, version, status, message):
        if int(status) not in (httplib.OK, httplib.NOT_MODIFIED):
            self.factory.failed(Failure(ProberResponseError(status)))

    def handleHeader(self, key, value):
        key = key.lower()
        l = self.headers.setdefault(key, [])
        l.append(value)

    def handleResponse(self, response):
        if response:
            # Uncompress the response based on content type.
            content_type = self.headers["content-type"][0]
            if content_type == "application/x-bzip2":
                response = bz2_decompress(response)
            elif content_type == "application/x-gzip":
                response = zlib_decompress(response)

        self.factory.succeeded(response)
        self.transport.loseConnection()


class ProberProtocolRedirect(ProberProtocol):
    """A specialized version of ProberProtocol that follows HTTP redirects."""

    redirect_to_location = False

    # The different redirect statuses supported.
    redirect_statuses = (
        httplib.MOVED_PERMANENTLY, httplib.FOUND, httplib.SEE_OTHER)

    def handleStatus(self, version, status, message):
        if int(status) in self.redirect_statuses:
            # We need to redirect to the location specified in the headers.
            self.redirect_to_location = True
        else:
            # We have the result immediately.
            ProberProtocol.handleStatus(self, version, status, message)

    def handleEndHeaders(self):
        # Server responded redirecting us to another location.
        if self.redirect_to_location:
            url = self.headers["location"][0]
            self.factory.redirect(url)
            self.transport.loseConnection()
        else:
            ProberProtocol.handleEndHeaders(self)


class ProberFactory(ClientFactory):
    """Factory using ProberProtocol to probe single URL existence."""

    protocol = ProberProtocol

    min_requests_to_consider_ratio = 30
    min_request_timeout_ratio = 3

    # Details of the URL of the host in which we actually want to request the
    # confirmation from.
    request_scheme = None
    request_host = None
    request_port = None
    request_path = None
    request_timeout = 180

    # Details of the URL of the host to which we'll connect.
    connect_scheme = None
    connect_username = None
    connect_password = None
    connect_host = None
    connect_port = None
    connect_path = None
    connect_date_modified = None

    host_requests = {}
    host_timeouts = {}

    def __init__(self, url, date_modified=None, timeout=None):
        # We want the deferred to be a private attribute (_deferred) to make
        # sure our clients will only use the deferred returned by the defer()
        # method; this is to ensure self._cancelTimeout is always the first
        # callback in the chain.
        self._deferred = defer.Deferred()
        self.timeout = timeout or self.request_timeout
        self.timeoutCall = None
        self.setURL(url.encode("ascii"), date_modified=date_modified)

    def shouldSkipHost(self, host):
        """Return True if the request/timeouts ratio is too low."""
        requests = self.host_requests[host]
        timeouts = self.host_timeouts[host]
        if timeouts == 0 or requests < self.min_requests_to_consider_ratio:
            return False
        else:
            ratio = float(requests) / timeouts
            return ratio < self.min_request_timeout_ratio

    def defer(self):
        if self.shouldSkipHost(self.request_host):
            reactor.callLater(0, self.failed, ProberSkipError(self.url))
            logging.debug(
                "Skipping %s as we've had too many timeouts on this "
                "host already." % self.url)
            return self._deferred

        self.connect()
        logging.debug("Probing %s" % self.url)
        return self._deferred

    def connect(self):
        self.host_requests[self.request_host] += 1
        if self.connect_scheme.endswith("s"):
            reactor.connectSSL(
                self.connect_host, self.connect_port, self,
                ClientContextFactory())
        else:
            reactor.connectTCP(self.connect_host, self.connect_port, self)

        if self.timeoutCall is not None and self.timeoutCall.active():
            self._cancelTimeout(None)

        self.timeoutCall = reactor.callLater(
            self.timeout, self.failWithTimeoutError)
        self._deferred.addBoth(self._cancelTimeout)

    connector = None

    def failWithTimeoutError(self):
        self.host_timeouts[self.request_host] += 1
        self.failed(ProberTimeout(self.url, self.timeout))
        if self.connector is not None:
            self.connector.disconnect()

    def startedConnecting(self, connector):
        self.connector = connector

    def succeeded(self, response):
        if not self._deferred.called:
            self._deferred.callback(response)

    def failed(self, reason):
        self._deferred.errback(reason)

    def _cancelTimeout(self, result):
        if self.timeoutCall.active():
            self.timeoutCall.cancel()
        return result

    def _parse(self, url):
        """Parse the given URL returning the scheme, host, port and path."""
        scheme, host, path, dummy, dummy, dummy = urlparse(url)

        if "@" in host:
            username, host = host.rsplit("@", 1)
            if ":" in username:
                username, password = username.split(":", 1)
            else:
                password = None
        else:
            username = password = None

        if ":" in host:
            host, port = host.split(":")
            assert port.isdigit()
            port = int(port)
        elif scheme == "http":
            port = 80
        elif scheme == "https":
            port = 443
        else:
            raise ProberSchemeError(url)

        return scheme, username, password, host, port, path

    def setURL(self, url, date_modified=None, redirect=False):
        self.url = url
        scheme, username, password, host, port, path = self._parse(url)
        if scheme not in ("http", "ftp"):
            raise ProberSchemeError(url)

        if scheme and host:
            self.request_scheme = scheme
            self.request_host = host
            self.request_port = port
            self.request_path = path

        if self.request_host not in self.host_requests:
            self.host_requests[self.request_host] = 0
        if self.request_host not in self.host_timeouts:
            self.host_timeouts[self.request_host] = 0

        # If the http_proxy variable is set, we want to use it as the host
        # we're going to connect to.
        proxy = getenv("http_proxy")
        if proxy:
            scheme, dummy, dummy, host, port, dummy = self._parse(proxy)
            path = url

        if not redirect:
            self.connect_username = username
            self.connect_password = password

        self.connect_scheme = scheme
        self.connect_host = host
        self.connect_port = port
        self.connect_path = path
        self.connect_date_modified = date_modified


class ProberFactoryRedirect(ProberFactory):

    redirection_max = 3

    protocol = ProberProtocolRedirect
    redirection_count = 0

    def redirect(self, url):
        self.timeoutCall.reset(self.timeout)

        try:
            if self.redirection_count >= self.redirection_max:
                raise ProberLoopError()
            self.redirection_count += 1

            logging.debug("Got redirected from %s to %s" % (self.url, url))
            self.setURL(url, redirect=True)
        except (ProberLoopError, ProberSchemeError), e:
            self.failed(e)
        else:
            self.connect()


class ProberLock(defer._ConcurrencyPrimitive):
    """Lock that acquires multiple underlying locks before it is acquired."""

    def __init__(self, overall_lock, host_lock):
        defer._ConcurrencyPrimitive.__init__(self)
        self.overall_lock = overall_lock
        self.host_lock = host_lock
        # host_lock will always be the scarcer resource, so it should be the
        # first to be acquired.
        self._locks = [host_lock, overall_lock]

    def acquire(self):
        return defer.gatherResults([lock.acquire() for lock in self._locks])

    def release(self):
        for lock in self._locks:
            lock.release()


class ProberManager:

    factory = ProberFactory

    overall_requests = 100
    overall_semaphore = DeferredSemaphore(overall_requests)

    per_host_requests = 2

    # Yes, I want a mutable class attribute because I want changes done in an
    # instance to be visible in other instances as well.
    host_locks = {}

    # Yes, I also want pending keys to be shared across all prober instances.
    pending_keys = []

    def __init__(self, factory=None):
        if factory is not None:
            self.factory = factory

    def addKey(self, deferred, key):
        self.pending_keys.append(key)
        deferred.addBoth(self.removeKey, key)

    def removeKey(self, response, key):
        self.pending_keys.remove(key)
        if not len(self.pending_keys):
            reactor.callLater(0, reactor.stop)
        return response

    def run(self, url, *args, **kwargs):
        prober = self.factory(url, *args, **kwargs)
        host = prober.request_host

        # Use a ProberLock with one semaphore limiting the overall
        # connections and another limiting the per-host connections.
        if host in self.host_locks:
            multi_lock = self.host_locks[host]
        else:
            multi_lock = ProberLock(
                self.overall_semaphore,
                DeferredSemaphore(self.per_host_requests))
            self.host_locks[host] = multi_lock
        deferred = multi_lock.run(prober.defer)
        self.addKey(deferred, url)

        return deferred
