# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "State",
    ]


_state_types = {}


def register_state(state):
    """Register a class to be an implementation for L{State}s.

    The class's C{__state_name__} must be defined as a string and
    C{__state_parent__} as a class.
    """
    if "__state_name__" not in state.__dict__:
        state.__state_name__ = ""
    if "__state_parent__" not in state.__dict__:
        state.__state_parent__ = None

    _state_types[state] = {}
    if state.__state_parent__ is not None:
        if state.__state_parent__ not in _state_types:
            raise RuntimeError("%r state ancestor does not exist already"
                % state.__state_parent__)

        if state.__state_name__ in _state_types[state.__state_parent__]:
            raise RuntimeError("%s state name already in ancestor %r"
                % (state.__state_name__, state.__state_parent__))

        _state_types[state.__state_parent__][state.__state_name__] = state


class StateMeta(type):

    def __init__(cls, name, bases, dict):
        super(StateMeta, cls).__init__(name, bases, dict)
        register_state(cls)


class State:

    __metaclass__ = StateMeta

    __slots__ = ("parent",)

    def __init__(self, parent=None):
        self.parent = parent

    @property
    def depth(self):
        depth = 0
        state = self.parent
        while state is not None:
            depth += 1
            state = state.parent

        return depth

    def getSibling(self):
        cls = self.__class__
        return cls(self.parent)

    def getChild(self, name=""):
        cls = self.__class__
        children = _state_types[cls]
        if not children:
            raise Exception("State has no children: %s" % cls)

        if name not in children:
            raise Exception("State does not have child: %s" % name)

        return children[name](self)

    def handleEntry(self, result, *args):
        pass
