# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import os

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from twisted.internet import (
    defer,
    reactor,
    )

from lpresults.repository.prober import (
    ProberFactory,
    ProberFactoryRedirect,
    ProberLock,
    ProberProtocolRedirect,
    )
from lpresults.repository.testing.fakes import (
    FakeFactory,
    FakeTimeoutCall,
    FakeTransport,
    )


class TestProberFactory(TestCase):
    """Tests to ensure we stop issuing requests on a given host if the
    requests/timeouts ratio on that host is too low.
    """
    host = 'foo.bar'

    def setUp(self):
        self.orig_host_requests = dict(
            ProberFactory.host_requests)
        self.orig_host_timeouts = dict(
            ProberFactory.host_timeouts)

    def tearDown(self):
        # Restore the globals that our tests fiddle with.
        ProberFactory.host_requests = self.orig_host_requests
        ProberFactory.host_timeouts = self.orig_host_timeouts

    def _createProber(self, requests, timeouts):
        def connect():
            prober.connectCalled = True

        ProberFactory.host_requests = {self.host: requests}
        ProberFactory.host_timeouts = {self.host: timeouts}
        prober = ProberFactory('http://%s/baz' % self.host)
        prober.connectCalled = False
        prober.failed = lambda error: None
        prober.connect = connect
        prober.defer()
        return prober

    def test_shouldSkipHostIfNotEnoughRequests(self):
        requests = ProberFactory.min_requests_to_consider_ratio - 1
        timeouts = requests
        prober = self._createProber(requests, timeouts)
        self.failUnless(prober.connectCalled)
        # Ensure the number of requests and timeouts we're using should
        # _NOT_ cause a given host to be skipped.
        self.failIf(prober.shouldSkipHost(self.host))

    def test_shouldSkipHostIfNotManyTimeouts(self):
        # If the ratio is not too small we consider it's safe to keep
        # issuing connections on that host.
        requests = ProberFactory.min_requests_to_consider_ratio
        timeouts = (ProberFactory.min_requests_to_consider_ratio
            / ProberFactory.min_request_timeout_ratio) - 2
        prober = self._createProber(requests, timeouts)
        self.failUnless(prober.connectCalled)
        # Ensure the number of requests and timeouts we're using should
        # _NOT_ cause a given host to be skipped.
        self.failIf(prober.shouldSkipHost(self.host))

    def test_shouldSkipHostAfterTooManyTimeouts(self):
        # If the ratio is small enough and we have a considerable number of
        # requests, we won't issue more connections on that host.
        requests = ProberFactory.min_requests_to_consider_ratio
        timeouts = (ProberFactory.min_requests_to_consider_ratio
            / ProberFactory.min_request_timeout_ratio) + 2
        prober = self._createProber(requests, timeouts)
        self.failIf(prober.connectCalled)
        # Ensure the number of requests and timeouts we're using should
        # actually cause a given host to be skipped.
        self.failUnless(prober.shouldSkipHost(self.host))


class TestProberProtocolAndFactory(TestCase):

    def assertConnectToHost(self, url, host):
        """Check that a ProberFactory created with the given url will actually
        connect to the given host.
        """
        prober = ProberFactory(url)

        def fakeConnect(host, port, factory):
            factory.connecting_to = host
            factory.succeeded("200")

        prober.connecting_to = None
        orig_connect = reactor.connectTCP
        reactor.connectTCP = fakeConnect

        def restore_connect(result, orig_connect):
            self.failUnlessEqual(prober.connecting_to, host)
            reactor.connectTCP = orig_connect
            return None

        deferred = prober.defer()
        return deferred.addCallback(restore_connect, orig_connect)

    def test_withProxy(self):
        os.environ["http_proxy"] = "http://squid.internal:3128"
        url = "http://localhost"
        self.assertConnectToHost(url, "squid.internal")

    def test_withoutProxy(self):
        try:
            del os.environ["http_proxy"]
        except KeyError:
            pass
        url = "http://localhost"
        self.assertConnectToHost(url, "localhost")


class TestProberProtocolAndFactoryRedirect(TestCase):

    def _createProberRedirect(self, url=None):
        if url is None:
            url = "http://foo.bar"
        prober = ProberFactoryRedirect(url)
        prober.timeoutCall = FakeTimeoutCall()
        prober.connectCalled = False

        def connect():
            prober.connectCalled = True

        prober.connect = connect
        return prober

    def test_redirectResetsTimeout(self):
        prober = ProberFactoryRedirect("http://foo.bar")
        prober.timeoutCall = FakeTimeoutCall()
        prober.connect = lambda: None
        self.assertFalse(prober.timeoutCall.resetCalled)
        prober.redirect("http://bar.foo")
        self.assertTrue(prober.timeoutCall.resetCalled)

    def test_noConnectionOnInfiniteLoop(self):
        prober = self._createProberRedirect()
        prober.failed = lambda error: None
        prober.redirection_count = prober.redirection_max
        prober.redirect("http://bar.foo")
        self.assertFalse(prober.connectCalled)

    def test_noConnectionOnUknownScheme(self):
        prober = self._createProberRedirect()
        prober.failed = lambda error: None
        prober.redirect("ssh://bar.foo")
        self.assertFalse(prober.connectCalled)

    def test_connectionOnRedirect(self):
        prober = self._createProberRedirect()
        prober.redirect("http://bar.foo")
        self.assertTrue(prober.connectCalled)

    def test_connectionClosedOnRedirect(self):
        protocol = ProberProtocolRedirect()
        protocol.factory = FakeFactory("http://foo.bar/")
        protocol.makeConnection(FakeTransport())
        protocol.dataReceived(
            "HTTP/1.1 301 Moved Permanently\r\n"
            "Location: http://foo.baz/\r\n"
            "Length: 0\r\n"
            "\r\n")
        self.assertEqual("http://foo.baz/", protocol.factory.redirectedTo)
        self.assertTrue(protocol.transport.disconnecting)


class TestProberLock(TestCase):

    def setUp(self):
        super(TestProberLock, self).setUp()
        self.lock_one = defer.DeferredLock()
        self.lock_two = defer.DeferredLock()
        self.multi_lock = ProberLock(self.lock_one, self.lock_two)
        self.count = 0

    def callback(self):
        self.count += 1

    def test_runDoesNotWait(self):
        """Multilock.run will run any given task if it's not locked and
        there's no task currently running.
        """
        self.multi_lock.run(self.callback)
        self.assertEquals(self.count, 1, "self.callback should have run.")

        self.multi_lock.run(self.callback)
        self.assertEquals(
            self.count, 2, "self.callback should have run twice.")

    def test_runWaitsFirstLock(self):
        """MultiLock.run acquires the first lock before running a function."""
        # Keep lock_one busy.
        deferred = defer.Deferred()
        self.lock_one.run(lambda: deferred)

        # Run self.callback when self.multi_lock is acquired.
        self.multi_lock.run(self.callback)
        self.assertEquals(
            self.count, 0, "self.callback should not have run yet.")

        # Release lock_one.
        deferred.callback(None)

        # multi_lock will now have been able to acquire both semaphores, and
        # so it will have run its task.
        self.assertEquals(self.count, 1, "self.callback should have run.")

    def test_runWaitsSecondLock(self):
        """MultiLock.run acquires the second lock before running functions."""
        # Keep lock_two busy.
        deferred = defer.Deferred()
        self.lock_two.run(lambda: deferred)

        # Run self.callback when self.multi_lock is acquired.
        self.multi_lock.run(self.callback)
        self.assertEquals(
            self.count, 0, "self.callback should not have run yet.")

        # Release lock_two.
        deferred.callback(None)

        # multi_lock will now have been able to acquire both semaphores, and
        # so it will have run its task.
        self.assertEquals(self.count, 1, "self.callback should have run.")

    def test_runWaitsCurrentTask(self):
        """MultiLock.run waits the end of the current task before running the
        next.
        """
        # Keep multi_lock busy.
        deferred = defer.Deferred()
        self.multi_lock.run(lambda: deferred)

        # Run self.callback when self.multi_lock is acquired.
        self.multi_lock.run(self.callback)
        self.assertEquals(
            self.count, 0, "self.callback should not have run yet.")

        # Release lock_one.
        deferred.callback(None)

        # multi_lock will now have been able to acquire both semaphores, and
        # so it will have run its task.
        self.assertEquals(self.count, 1, "self.callback should have run.")


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
