# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from zope.component import (
    getUtility,
    provideUtility,
    )

from lpresults.repository.source import Source
from lpresults.repository.exceptions import RepositoryError

from lpresults.repository.zope.interfaces import IZRepository
from lpresults.repository.zope.zrepository import ZRepository


class TestZRepository(TestCase):

    def setUp(self):
        self.zrepository = ZRepository()

    def tearDown(self):
        self.zrepository._reset()

    def test_noName(self):
        self.assertRaises(RepositoryError, self.zrepository.get, "pci")

    def test_createNamed(self):
        source = self.zrepository.create("pci", "file:///pci")
        self.assertTrue(self.zrepository.get("pci") is source)

    def test_createNamedTwice(self):
        self.zrepository.create("pci", "file:///pci")
        self.assertRaises(
            RepositoryError, self.zrepository.create, "pci", "file:///pci")

    def test_createDefault(self):
        self.zrepository.setDefaultURL("pci", "file:///pci")
        source = self.zrepository.create("pci")
        self.assertTrue(isinstance(source, Source))

    def test_createDefaultTwice(self):
        self.zrepository.setDefaultURL("pci", "file:///pci")
        self.zrepository.create("pci")
        self.assertRaises(RepositoryError, self.zrepository.create, "pci")

    def test_getName(self):
        source = self.zrepository.create("pci", "file:///pci")
        self.assertEquals(self.zrepository.getName(source), "pci")

    def test_getWithURL(self):
        source = self.zrepository.get("pci", "file:///pci")
        self.assertTrue(isinstance(source, Source))
        self.assertTrue(self.zrepository.get("pci") is source)
        self.assertTrue(self.zrepository.get("pci", "file:///pci") is source)

    def test_itersources(self):
        source1 = self.zrepository.create("pci", "file:///pci")
        source2 = self.zrepository.create("pnp", "file:///pnp")
        source3 = self.zrepository.create("usb", "file:///usb")
        sources = []
        for name, source in self.zrepository.itersources():
            sources.append((name, source))
        self.assertEquals(len(sources), 3)
        self.assertEquals(set(sources),
                          set([("pci", source1), ("pnp", source2),
                               ("usb", source3)]))

    def test_remove(self):
        source = self.zrepository.create("pci", "file:///pci")
        self.assertEquals(self.zrepository.getName(source), "pci")
        self.zrepository.remove(source)
        self.assertEquals(self.zrepository.getName(source), None)

    def test_setDefaultURL(self):
        self.zrepository.setDefaultURL("pci", "file:///pci")
        source = self.zrepository.get("pci")
        self.assertTrue(isinstance(source, Source))

    def test_getDefaultURLs(self):
        self.zrepository.setDefaultURL("pci", "file:///pci")
        self.zrepository.setDefaultURL("pnp", "file:///pnp")
        self.zrepository.setDefaultURL("usb", "file:///usb")
        default_uris = self.zrepository.getDefaultURLs()
        self.assertEquals(default_uris, {
            "pci": "file:///pci",
            "pnp": "file:///pnp",
            "usb": "file:///usb",
            })


class TestZRepositoryUtility(TestCase):

    def test_utility(self):
        provideUtility(ZRepository())
        self.assertTrue(isinstance(getUtility(IZRepository), ZRepository))


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
