# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "ZRepository",
    "global_zrepository",
    ]

from zope.interface import implements

from lpresults.repository.exceptions import RepositoryError
from lpresults.repository.source import create_source
from lpresults.repository.zope.interfaces import IZRepository


class ZRepository:

    implements(IZRepository)

    _sources = {}

    def __init__(self):
        self._named = {}
        self._name_index = {}
        self._default_sources = {}
        self._default_urls = {}

    def _reset(self):
        for name, source in list(self.itersources()):
            self.remove(source)

        self._sources.clear()
        self._default_sources.clear()
        self._default_urls.clear()

    def _getSource(self, name, url):
        source = self._sources.get(name)
        if source is None:
            return self._sources.setdefault(name, create_source(name, url))

        return source

    def getDefaultURLs(self):
        """Return a list of name, url tuples that are named as the default
        sources for those names.
        """
        return self._default_urls.copy()

    def setDefaultURL(self, name, default_url):
        """Set the default URL for sources by name."""
        self._default_sources[name] = self._getSource(name, default_url)
        self._default_urls[name] = default_url

    def create(self, name, url=None):
        """Create a new source.

        :param name: Name of the source.
        :param URL: Optionally, URL to use if no default.
        """
        if url is None:
            source = self._default_sources.get(name)
            if source is None:
                raise RepositoryError("Source named '%s' not found" % name)
        else:
            source = self._getSource(name, url)

        if name is not None and self._named.get(name) is not None:
            raise RepositoryError("Source named '%s' already exists" % name)

        self._named[name] = source
        self._name_index[source] = name

        return source

    def get(self, name, default_url=None):
        """Get the source by name, creating it first if necessary.

        :param name: Name of the source to get.
        :param default_url: Optionally, URL to use if no default.
        """
        source = self._named.get(name)
        if not source:
            return self.create(name, default_url)

        return source

    def getName(self, source):
        """Returns the name for the given source or None if one isn't
        available."""
        return self._name_index.get(source)

    def remove(self, source):
        """Remove the given source from ZRepository."""
        name = self._name_index[source]
        del self._name_index[source]
        if name in self._named:
            del self._named[name]

    def itersources(self):
        """Iterate over name, source tuples."""
        for source, name in self._name_index.items():
            yield name, source


global_zrepository = ZRepository()
