# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "SessionStore",
    ]

import hmac
import random
import string
import time

from UserDict import DictMixin
from datetime import timedelta
from hashlib import sha1

from storm.locals import (
    Pickle,
    RawStr,
    Int,
    )

from django.conf import settings
from django.contrib.sessions.backends.base import SessionBase

from lpresults.session.store import get_session_store


COOKIE_SAFE_TRANS = string.maketrans("+/", "-.")


def digest_encode(s):
    """Encode SHA digest for cookie."""
    return s.encode("base64")[:-2].translate(COOKIE_SAFE_TRANS)


class SessionStore(SessionBase):
    """
    Implements database session store.
    """
    resolution = timedelta(seconds=10 * 60)
    timeout = timedelta(seconds=60 * 60 * 60)

    def _get_new_session_key(self):
        """See `SessionBase`."""
        data = "%.20f%.20f%.20f" % (
            random.random(),
            time.time(),
            time.clock())
        digest = sha1(data).digest()
        s = digest_encode(digest)
        # we store a HMAC of the random value together with it, which makes
        # our session ids unforgeable.
        mac = hmac.new(settings.SECRET_KEY, s, digestmod=sha1).digest()
        return s + digest_encode(mac)

    def load(self):
        """
        Return a SessionPkgData for the given C{session_key}. If it
        already exists, it returns it from the database, if not, it
        creates a new one.
        """
        # XXX In the future, we'll most probably want to cache the
        #     generation locally for all threads (protected by a lock),
        #     but this function is very fast and this is far from being
        #     an issue for a long time, so we can push the first version
        #     to production without any caching.
        result = get_session_store(autocommit=True).execute(
            "SELECT get_generation(?, INTERVAL ?, INTERVAL ?)", (
                self.session_key,
                unicode(self.resolution),
                unicode(self.timeout)))
        return SessionPkgDataProxy(self.session_key, result.get_one()[0])

    def flush(self):
        # XXX At some point, sessions should be purged from the
        #     session_data table which is currently only accessed
        #     by the get_generation function.
        get_session_store().find(
            SessionPkgData,
            SessionPkgData.session_key == self.session_key).remove()

    def save(self, must_create=False):
        # Session data is persisted automatically
        pass

    def cycle_key(self):
        # XXX This still needs to be implemented
        pass


class SessionPkgDataProxy(DictMixin):

    def __init__(self, session_key, generation):
        self.session_key = session_key
        self.generation = generation
        self.store = get_session_store()

    def __getitem__(self, key):
        obj = self.store.get(SessionPkgData,
            (self.session_key, self.generation, key,))
        if obj is not None:
            return obj.data
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        obj = self.store.get(SessionPkgData,
            (self.session_key, self.generation, key,))
        if obj is not None:
            obj.data = value
        else:
            obj = SessionPkgData(
                self.session_key, self.generation, key, value)
            self.store.add(obj)

    def __delitem__(self, key):
        obj = self.store.get(SessionPkgData,
            (self.session_key, self.generation, key,))
        if obj is not None:
            self.store.remove(obj)
        else:
            raise KeyError(key)

    def keys(self):
        result = self.store.find(
            SessionPkgData,
            session_key=self.session_key,
            generation=self.generation)
        return result.values(SessionPkgData.key)


class SessionPkgData:

    __storm_table__ = "sessionpkgdata"
    __storm_primary__ = ("session_key", "generation", "key",)

    session_key = RawStr(allow_none=False)
    generation = Int(allow_none=False)
    key = RawStr(allow_none=False)
    data = Pickle(allow_none=False)

    def __init__(self, session_key, generation, key, value):
        self.session_key = session_key
        self.generation = generation
        self.key = key
        self.data = value
