# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import

__metaclass__ = type

__all__ = [
    "get_openid_store",
    ]

import time

from openid.association import Association
from openid.store import (
    interface,
    nonce,
    )

from storm.locals import Int
from storm.properties import SimpleProperty
from storm.variables import RawStrVariable

from lpresults.session.store import get_session_store


class EncodedStrVariable(RawStrVariable):

    def __init__(self, *args, **kwargs):
        self._encoding = kwargs.pop("encoding")
        super(EncodedStrVariable, self).__init__(*args, **kwargs)

    def parse_set(self, value, from_db):
        if isinstance(value, unicode):
            value = value.encode(self._encoding)
        return super(EncodedStrVariable, self).parse_set(value, from_db)


class EncodedStr(SimpleProperty):
    """
    python-openid is not consistent about whether it passes unicode or
    str to the OpenIDStore.  The easiest way to deal with this is just
    to automatically encode all unicode objects to str in a property.
    Hence, EncodedStr.

    Pass the C{encoding} parameter to the constructor to specify which
    encoding should be used.
    """
    variable_class = EncodedStrVariable


class OpenIDAssociation:

    __storm_table__ = "openidassociation"
    __storm_primary__ = ("server_url", "handle",)

    server_url = EncodedStr(encoding="ascii", allow_none=False)
    handle = EncodedStr(encoding="ascii", allow_none=False)
    secret = EncodedStr(encoding="ascii", allow_none=False)
    issued = Int(allow_none=False)
    lifetime = Int(allow_none=False)
    association_type = EncodedStr(encoding="ascii", allow_none=False)

    def __init__(
        self, server_url, handle, secret, issued, lifetime, association_type):

        self.server_url = server_url
        self.handle = handle
        self.secret = secret
        self.issued = issued
        self.lifetime = lifetime
        self.association_type = association_type

    def getAssociation(self):
        return Association(
            self.handle, self.secret, self.issued,
            self.lifetime, self.association_type)


class OpenIDNonce:

    __storm_table__ = "openidnonce"
    __storm_primary__ = ("server_url", "timestamp", "salt",)

    server_url = EncodedStr(encoding="ascii", allow_none=False)
    timestamp = Int(allow_none=False)
    salt = EncodedStr(encoding="ascii", allow_none=False)

    def __init__(self, server_url, timestamp, salt):
        self.server_url = server_url
        self.timestamp = timestamp
        self.salt = salt


class OpenIDStore(interface.OpenIDStore):

    def __init__(self, storm_store):
        self._store = storm_store

    def storeAssociation(self, server_url, association):
        existing = self._store.find(
            OpenIDAssociation, server_url=server_url,
            handle=association.handle).one()
        if existing is not None:
            existing.secret = association.secret
            existing.issued = association.issued
            existing.lifetime = association.lifetime
            existing.association_type = association.assoc_type
        else:
            self._store.add(
                OpenIDAssociation(
                    server_url,
                    association.handle,
                    association.secret,
                    association.issued,
                    association.lifetime,
                    association.assoc_type))

    def getAssociation(self, server_url, handle=None):
        if handle is not None:
            association = self._store.find(
                OpenIDAssociation, server_url=server_url,
                handle=handle).one()
            if association is not None:
                return association.getAssociation()
            else:
                return None

        results = self._store.find(OpenIDAssociation, server_url=server_url)

        results = results.order_by(OpenIDAssociation.issued)
        results = list(results)
        if len(results) == 0:
            return None
        else:
            result = None
            for storm_association in results:
                association = storm_association.getAssociation()
                if association.getExpiresIn() == 0:
                    self._store.remove(storm_association)
                else:
                    result = association
            return result

    def removeAssociation(self, server_url, handle):
        results = self._store.find(
            OpenIDAssociation, server_url=server_url, handle=handle)
        # if remove() returned the number of deleted objects this would be
        # more efficient.
        count = results.count()
        if count > 0:
            results.remove()
        return count > 0

    def useNonce(self, server_url, timestamp, salt):
        if abs(timestamp - time.time()) > nonce.SKEW:
            return False

        old_nonces = self._store.find(
            OpenIDNonce, server_url=server_url,
            timestamp=timestamp, salt=salt)[:1]
        if old_nonces.count() > 0:
            return False
        else:
            self._store.add(OpenIDNonce(server_url, timestamp, salt))
            return True

    def cleanupNonces(self):
        expiry_time = time.time() - nonce.SKEW
        nonces = self._store.find(
            OpenIDNonce, OpenIDNonce.timestamp < expiry_time)
        # Again with the inefficiency
        count = nonces.count()
        if count > 0:
            nonces.remove()
        return count

    def cleanupAssociations(self):
        assocs = self._store.find(
            OpenIDAssociation,
            (OpenIDAssociation.issued + OpenIDAssociation.lifetime)
                < time.time())
        # Again with the inefficiency
        count = assocs.count()
        if count > 0:
            assocs.remove()
        return count


def get_openid_store():
    store = get_session_store()
    return OpenIDStore(store)
