# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Backend",
    "create_backend",
    "register_scheme",
    ]

from lpresults import storage
from lpresults.storage.connection import Connection
from lpresults.storage.exceptions import StorageError
from lpresults.storage.uri import URI


class Backend:
    """
    Base class for a backend.
    """
    connection_factory = Connection

    def connect(self):
        """Create a connection to the backend."""
        return self.connection_factory(self)


_backend_schemes = {}


def create_backend(uri):
    """Create a backend instance.

    :param uri: URI instance, or a string describing the URI.
    """
    if isinstance(uri, basestring):
        uri = URI(uri)

    if uri.scheme in _backend_schemes:
        factory = _backend_schemes[uri.scheme]

    else:
        module_name = "%s.backends.%s" % (
            storage.__name__,
            uri.scheme.replace("-", "_"))
        try:
            module = __import__(module_name, None, None, [""])
        except ImportError:
            raise StorageError("Backend named '%s' not found" % uri.scheme)
        factory = module.create_from_uri

    return factory(uri)


def register_scheme(scheme, factory):
    """Register a handler for a new backend URI scheme.

    :param scheme: the backend URI scheme
    :param factory: a function taking a URI instance and returning a backend.
    """
    _backend_schemes[scheme] = factory
