# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "create_from_uri",
    ]

import os
import errno
import random
import shutil
import stat

from datetime import datetime
from tempfile import mkstemp
from zlib import adler32

from lpresults.storage.backend import Backend
from lpresults.storage.connection import Connection
from lpresults.storage.exceptions import (
    ClientError,
    ServerError,
    )
from lpresults.storage.info import Info


class FileConnection(Connection):

    read_size = 1024 * 64

    def __init__(self, path):
        self._path = path
        self._tmp = os.path.join(path, "tmp")

    def getDirectoryPath(self, directory_name):
        checksum = adler32(directory_name)
        return os.path.join(
            self._path,
            "%02X" % (checksum & 0xFF),
            "%02X" % ((checksum >> 16) & 0xFF),
            directory_name)

    def getFilePath(self, directory_name, file_name):
        directory_path = self.getDirectoryPath(directory_name)
        return os.path.join(directory_path, file_name)

    def getTmp(self, file_name):
        assert "/" not in file_name
        if not os.path.exists(self._tmp):
            try:
                os.makedirs(self._tmp, mode=0750)
            except OSError:
                pass

        fd, path = mkstemp(prefix=file_name, dir=self._tmp)
        file = os.fdopen(fd, "w")

        # Change the mode to be readable by group
        mode = stat.S_IMODE(os.stat(path).st_mode)
        os.chmod(path, mode | stat.S_IRGRP)

        return file, path

    def deleteDirectory(self, directory_name):
        """See Connection."""
        # First move everything out of the way, then delete it
        # to avoid concurrency issues.
        directory_path = self.getDirectoryPath(directory_name)
        deleted_path = "%s.deleted.%d" % (
            directory_path,
            int(random.random() * 10 ** 16))
        os.rename(directory_path, deleted_path)
        shutil.rmtree(deleted_path)

        # Second, remove the adler directory if empty
        parent_path = os.path.dirname(directory_path)
        try:
            os.rmdir(parent_path)
        except OSError, error:
            if error.errno != errno.ENOTEMPTY:
                raise

    def listDirectory(self, directory_name):
        """See Connection."""
        directory_path = self.getDirectoryPath(directory_name)
        try:
            for file_name in os.listdir(directory_path):
                yield file_name
        except OSError:
            pass

    def getInfo(self, directory_name, file_name):
        file_path = self.getFilePath(directory_name, file_name)
        try:
            file_stat = os.stat(file_path)
        except OSError, error:
            raise ClientError(str(error))

        # Strip microseconds from timestamp
        size = file_stat.st_size
        date_modified_stamp = int(file_stat.st_mtime)
        date_modified = datetime.fromtimestamp(date_modified_stamp)

        return Info(file_name, size, date_modified)

    def getContent(self, directory_name, file_name):
        """See Connection."""
        file_path = self.getFilePath(directory_name, file_name)
        try:
            file = open(file_path, "r")
        except IOError, error:
            raise ClientError(str(error))

        try:
            for chunk in iter(lambda: file.read(self.read_size), ""):
                yield chunk
        finally:
            file.close()

    def putContent(self, directory_name, file_name, file):
        """See Connection."""
        file_path = self.getFilePath(directory_name, file_name)
        (tmp_file, tmp_path) = self.getTmp(file_name)

        # Read in and write out files chunk_size at a time
        try:
            for chunk in iter(lambda: file.read(self.read_size), ""):
                tmp_file.write(chunk)
        finally:
            tmp_file.close()

        # Create parent directories and set mode accordingly
        try:
            os.makedirs(os.path.dirname(file_path), mode=0750)
        except OSError, error:
            if error.errno != errno.EEXIST:
                raise

        # Move temporary path to file path
        try:
            if hasattr(os, "link"):
                os.link(tmp_path, file_path)
                os.remove(tmp_path)
            else:
                os.rename(tmp_path, file_path)
        except OSError, error:
            os.remove(tmp_path)
            if error.errno == errno.EEXIST:
                raise ServerError(
                    "Invalid destination path: '%s'" % file_path)
            else:
                raise

    def deleteFile(self, directory_name, file_name):
        """See Connection."""
        file_path = self.getFilePath(directory_name, file_name)
        os.unlink(file_path)


class FileBackend(Backend):

    connection_factory = FileConnection

    def __init__(self, uri):
        if not uri.path:
            raise ValueError("Empty path for file backend: %s" % uri)

        # Safe to cache connection instance
        self._connection = self.connection_factory(uri.path)

    def connect(self):
        return self._connection


create_from_uri = FileBackend
