# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""S3 storage backend."""

__metaclass__ = type

__all__ = [
    "create_from_uri",
    ]

import os
import re
import time

import hmac
import httplib
import mimetypes
import rfc822
import socket

from base64 import b64encode
from datetime import datetime
from hashlib import (
    md5,
    sha1,
    sha256,
    )

from lpresults.storage.backend import Backend
from lpresults.storage.cache import LRUCache
from lpresults.storage.connection import Connection
from lpresults.storage.exceptions import (
    ClientError,
    ServerError,
    )
from lpresults.storage.info import Info
from lpresults.storage.retry import retry_except
from lpresults.storage.uri import URI

try:
    import xml.etree.cElementTree as etree
except ImportError:
    import cElementTree as etree


class S3Credentials:

    def __init__(self, access_key, secret_key):
        if not access_key:
            raise ValueError("Empty access key for S3 credentials")
        if not secret_key:
            raise ValueError("Empty secret key for S3 credentials")

        self.access_key = access_key
        self.secret_key = secret_key

    def sign(self, data, hash_type="sha1"):
        """Sign data."""
        if hash_type == "sha1":
            digest = hmac.new(self.secret_key, data, sha1).digest()

        elif hash_type == "sha256":
            digest = hmac.new(self.secret_key, data, sha256).digest()

        else:
            raise RuntimeError("Unsupported hash type: '%s'" % hash_type)

        return b64encode(digest)


class S3Request:

    def __init__(self, action, uri, credentials=None, headers=None):
        if headers is None:
            headers = {}

        self.action = action
        self.uri = uri
        self._credentials = credentials
        self._headers = headers or {}

        self.attributes = {}

    def _sign(self):
        """Sign the request using the built in credentials."""
        headers = self._headers

        parts = [
            self.action,
            headers.get("content-md5", ""),
            headers.get("content-type", ""),
            headers.get("date", ""),
            ]

        for key, value in headers.iteritems():
            key = key.lower()
            if key.startswith("x-amz-"):
                parts.append("%s:%s" % (key, value))

        resource = os.path.join("/", self.uri.path or "")
        parts.append(resource)

        signature = self._credentials.sign("\n".join(parts))
        headers["authorization"] = "AWS %s:%s" % (
            self._credentials.access_key, signature)

    def getHeaders(self):
        """Return the headers needed to perform S3 operations."""
        headers = self._headers

        if "date" in headers:
            del headers["date"]

        headers["x-amz-date"] = time.strftime(
            "%a, %d %b %Y %H:%M:%S +0000",
            time.gmtime())

        if self._credentials:
            self._sign()

        return headers

    def getURL(self):
        """Return the URL where to perform S3 operations."""
        uri = self.uri
        parts = []
        for key, value in uri.options.iteritems():
            if value in (None, ""):
                parts.append(key)
            else:
                parts.append("=".join([key, value]))

        return "?".join([uri.path, "&".join(parts)])


class S3ServerError(ServerError):

    def __init__(self, data, status, reason=None):
        if reason is None:
            reason = httplib.responses.get(status, "")

        self.status = status
        self.reason = reason
        self.info = {}

        if data:
            tree = etree.fromstring(data)
            if tree.tag != "Error":
                tree = tree.find(".//Error")

            if tree:
                for child in tree.getchildren():
                    if child.text != "":
                        self.info[child.tag] = child.text

    def __str__(self):
        string = "%d (%s)" % (self.status, self.info.get("Code", self.reason))
        if "Message" in self.info:
            string += ": %s" % self.info["Message"]

        return string


class S3RawConnection:

    def __init__(self, cache=None):
        if cache is None:
            self._cache = LRUCache()
        else:
            self._cache = cache

    def __getitem__(self, uri):
        host = uri.host
        if host not in self._cache:
            if uri.scheme == "https":
                connection_factory = httplib.HTTPSConnection

            elif uri.scheme == "http":
                connection_factory = httplib.HTTPConnection

            else:
                raise RuntimeError("Unsupported scheme: '%s'" % uri.scheme)

            connection = connection_factory(host, uri.port)
            self._cache[host] = connection

        else:
            connection = self._cache[host]

        return connection

    def __delitem__(self, uri):
        host = uri.host
        if host in self._cache:
            del self._cache[host]


class S3Connection(Connection):

    default_content_type = "binary/octet-stream"

    http_exceptions = (
        httplib.HTTPException,
        socket.error,
        socket.gaierror,
        )

    receive_size = 4096
    send_size = 4096

    def __init__(self, endpoint, credentials=None):
        self._endpoint = endpoint
        self._credentials = credentials
        self._raw_connection = S3RawConnection()

    def makeRequest(
        self, action, directory_name="", file_name="",
        headers={}, options={}):

        endpoint = self._endpoint.copy()
        endpoint.path = os.path.join(endpoint.path, directory_name, file_name)
        endpoint.options = options

        return S3Request(action, endpoint, self._credentials, headers)

    @retry_except(ClientError)
    def receiveFile(self, request):
        """Receive a file from the S3 backend.

        Returns an iterator of chunks received from the backend. If a
        ClientError is encountered, this method will attempt to receive
        the remaining chunks on a few retries.

        :param request: An `S3Request` about the file to receive.
        """
        headers = request.getHeaders()
        start_position = request.attributes.get("position", 0)
        md5_hash = request.attributes.get("md5", md5())

        # Send headers
        try:
            connection = self._raw_connection[request.uri]
            connection.putrequest(request.action, request.getURL())
            for key, value in headers.iteritems():
                connection.putheader(key, value)
            if start_position > 0:
                connection.putheader("Range", "bytes=%d-" % start_position)
            connection.endheaders()
            response = connection.getresponse()
            response_headers = dict(response.getheaders())
        except self.http_exceptions:
            del self._raw_connection[request.uri]
            raise ClientError("Download failed for: %s" % request.uri)

        if response.status >= 400:
            # Unrecoverable error
            raise S3ServerError(
                response.read(), response.status, response.reason)

        if response.status >= 300:
            # Redirect
            location = response_headers["location"]
            request.uri = URI(location)
            for chunk in self.receiveFile(request):
                yield chunk

        size_left = long(response_headers["content-length"])
        size_total = start_position + size_left
        current_position = start_position

        try:
            while (current_position < size_total):
                size = min(self.receive_size, size_left)
                chunk = response.read(size)
                md5_hash.update(chunk)
                current_position += len(chunk)
                yield chunk
        except Exception:
            request.attributes["position"] = current_position
            request.attributes["md5"] = md5_hash
            raise ClientError("Download failed for: %s" % request.uri)

        # Check content-length
        content_length = long(response_headers["content-length"])
        if current_position != start_position + content_length:
            raise ClientError("Sizes don't match")

        # Check etag
        md5_digest = md5_hash.hexdigest()
        if response_headers["etag"].strip("\"'") != md5_digest:
            raise ClientError("MD5 sums don't match")

    @retry_except(ClientError)
    def sendFile(self, request, file):
        """Send a file to the S3 backend.

        Iterates over the given file in chunks to send to the backend. If
        a ClientError is encountered, this method will attempt to continue
        sending the remaining chunks on a few retries.

        :param request: An `S3Request` about the file to send.
        :param file: A file-like object to read content.
        """
        headers = request.getHeaders()

        # Send headers
        try:
            connection = self._raw_connection[request.uri]
            connection.putrequest(request.action, request.getURL())

            for key, value in headers.iteritems():
                connection.putheader(key, value)
            connection.endheaders()
        except self.http_exceptions:
            del self._raw_connection[request.uri]
            raise ClientError("Upload failed for: %s" % request.uri)

        # Send data
        md5_hash = md5()
        size_left = long(headers.get("content-length"))
        try:
            while size_left > 0:
                data = file.read(self.send_size)
                md5_hash.update(data)
                connection.send(data)
                size_left -= len(data)

            response = connection.getresponse()
            response_headers = dict(response.getheaders())
        except Exception:
            connection.close()
            raise ClientError("Upload failed for: %s" % request.uri)

        # Check response
        if response.status >= 400:
            if response.status < 500:
                error = S3ServerError(
                    response.read(), response.status, response.reason)
                if error.info["Code"] not in [
                   "BadDigest", "OperationAborted",
                   "TokenRefreshRequired", "RequestTimeout"]:
                    # Unrecoverable error
                    raise error

            raise ClientError("Upload failed for: %s (%s)" % (
                request.uri, response.status))

        elif response.status >= 300:
            # Redirect
            location = response_headers["location"]
            request.uri = URI(location)
            self.sendFile(request, file)

        else:
            # S3 from time to time doesn't send etag back in a response
            md5_digest = md5_hash.hexdigest()
            if "etag" not in response_headers \
               or response_headers["etag"].strip("\"'") != md5_digest:
                raise ClientError("MD5 sums don't match")

    @retry_except(ClientError)
    def sendRequest(self, request):
        headers = request.getHeaders()

        try:
            connection = self._raw_connection[request.uri]
            connection.request(
                request.action, request.getURL(), headers=headers)
            response = connection.getresponse()
            response_headers = dict(response.getheaders())
        except self.http_exceptions:
            del self._raw_connection[request.uri]
            raise ClientError("Request failed for: %s" % request.uri)

        if response.status >= 500:
            # This is an unrecoverable error but worth retrying
            raise ClientError("Request failed for: %s (%s)" % (
                request.uri, response.status))

        elif response.status >= 400:
            # Unrecoverable error and response.read will block
            raise S3ServerError("", response.status, response.reason)

        elif response.status >= 300:
            # Redirect
            location = response_headers["location"]
            request.uri = URI(location)
            return self.sendRequest(request)

        return response

    def deleteFile(self, directory_name, file_name):
        """See Connection."""
        request = self.makeRequest("DELETE", directory_name, file_name)
        self.sendRequest(request)

    def listDirectory(self, directory_name):
        """See Connection."""
        namespace_re = re.compile("""^(<?[^>]+?>\s?)(<\w+)\s+"""
            """xmlns=['"](http://[^'\"]+)['"](.*)""",
            re.MULTILINE)

        options = {
            "prefix": "%s/" % directory_name,
            }

        while True:
            request = self.makeRequest("GET", options=options)
            response = self.sendRequest(request)
            data = namespace_re.sub("\\1\\2\\4", response.read())

            tree = etree.fromstring(data)
            for node in tree.findall(".//Contents"):
                options["marker"] = node.findtext(".//Key")
                file_name = options["marker"][len(options["prefix"]):]
                yield file_name

            truncated = tree.findtext(".//IsTruncated") or "false"
            if truncated == "false":
                break

    def getInfo(self, directory_name, file_name):
        """See Connection."""
        request = self.makeRequest("HEAD", directory_name, file_name)
        response = self.sendRequest(request)
        headers = dict(response.getheaders())

        size = int(headers["content-length"])
        content_type = headers["content-type"]
        date_modified_tuple = rfc822.parsedate_tz(headers["last-modified"])
        date_modified_stamp = rfc822.mktime_tz(date_modified_tuple)
        date_modified = datetime.fromtimestamp(date_modified_stamp)

        return Info(file_name, size, date_modified, content_type)

    def getContent(self, directory_name, file_name):
        """See Connection."""
        request = self.makeRequest("GET", directory_name, file_name)
        try:
            return self.receiveFile(request)
        except S3ServerError, error:
            if error.status == 404:
                raise ClientError(error.info["Message"])

            else:
                raise error

    def putContent(self, directory_name, file_name, file):
        """See Connection."""
        file.seek(0, os.SEEK_END)
        content_length = str(file.tell())
        content_type = mimetypes.guess_type(file_name)[0] \
            or self.default_content_type

        headers = {
            "content-length": content_length,
            "content-type": content_type,
            }
        request = self.makeRequest("PUT", directory_name, file_name, headers)

        file.seek(0)
        self.sendFile(request, file)


class S3Backend(Backend):

    connection_factory = S3Connection

    default_host = "s3.amazonaws.com"
    default_path = "/"

    def __init__(self, uri):
        if bool(uri.options.get("secure", 1)):
            scheme = "https"
        else:
            scheme = "http"

        self._endpoint = URI()
        self._endpoint.scheme = scheme
        self._endpoint.host = uri.host or self.default_host
        self._endpoint.path = uri.path or self.default_path

        if uri.username or uri.password:
            self._credentials = S3Credentials(uri.username, uri.password)
        else:
            self._credentials = None

    def connect(self):
        return self.connection_factory(self._endpoint, self._credentials)


create_from_uri = S3Backend
