# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import time

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from lpresults.storage.bucket import Bucket
from lpresults.storage.cache import LRUCache
from lpresults.storage.exceptions import ClientError
from lpresults.storage.uri import URI

from lpresults.storage.backends.s3 import (
    S3Backend,
    S3Connection,
    S3Credentials,
    S3RawConnection,
    S3Request,
    S3ServerError,
    )

from lpresults.storage.testing.cases import TestCaseWithFactory

from lpresults.storage.tests.test_backend import TestBackendMixin
from lpresults.storage.tests.test_connection import TestConnectionMixin
from lpresults.storage.tests.test_directory import TestDirectoryMixin


class FakeResponse:

    def __init__(self, data):
        self.data = data

    def read(self):
        return self.data


class FakeS3Connection(S3Connection):

    def __init__(self, *args, **kwargs):
        super(FakeS3Connection, self).__init__(*args, **kwargs)
        self._files = {}

    def receiveFile(self, request):
        if request.uri.path not in self._files:
            raise ClientError()

        if request.uri.path in self._files:
            yield self._files[request.uri.path]

    def sendFile(self, request, file):
        self._files[request.uri.path] = file.read()

    def sendRequest(self, request):
        if request.action == "GET":
            name = self._endpoint.path[1:]
            prefix = request.uri.options["prefix"]
            data = """<?xml version="1.0" encoding="UTF-8"?><ListBucketResult
                xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
                <Name>%s</Name>
                <Prefix>%s</Prefix>
                <Marker></Marker>
                <MaxKeys>1000</MaxKeys>
                <IsTruncated>false</IsTruncated>""" % (name, prefix)
            if self._files:
                data += """<Contents>"""
                for filename in self._files.iterkeys():
                    start = filename.index(name) + len(name) + 1
                    key = filename[start:]
                    data += """<Key>%s</Key>""" % key
                data += """</Contents>"""
            data += """</ListBucketResult>"""

            return FakeResponse(data)

        elif request.action == "DELETE":
            del self._files[request.uri.path]


class FakeS3Backend(S3Backend):

    connection_factory = FakeS3Connection


class TestS3Credentials(TestCase):

    def test_init(self):
        credentials = S3Credentials("foo", "bar")
        self.assertTrue(isinstance(credentials, S3Credentials))
        self.assertEquals(credentials.access_key, "foo")
        self.assertEquals(credentials.secret_key, "bar")

    def test_initNoAccessKey(self):
        self.assertRaises(
            ValueError,
            S3Credentials,
            access_key=None,
            secret_key="bar")
        self.assertRaises(
            ValueError,
            S3Credentials,
            access_key="",
            secret_key="bar")

    def test_initNoSecretKey(self):
        self.assertRaises(
            ValueError,
            S3Credentials,
            access_key="foo",
            secret_key=None)
        self.assertRaises(
            ValueError,
            S3Credentials,
            access_key="foo",
            secret_key="")

    def test_sign(self):
        credentials = S3Credentials("foo", "bar")
        signature = credentials.sign("baz")
        self.assertTrue(isinstance(signature, str))

    def test_signSha1(self):
        credentials = S3Credentials("foo", "bar")
        signature = credentials.sign("baz", hash_type="sha1")
        self.assertEquals(len(signature), 28)

    def test_signSha256(self):
        credentials = S3Credentials("foo", "bar")
        signature = credentials.sign("baz", hash_type="sha256")
        self.assertEquals(len(signature), 44)

    def test_signInvalid(self):
        credentials = S3Credentials("foo", "bar")
        self.assertRaises(
            RuntimeError,
            credentials.sign,
            "baz",
            hash_type="invalid")


class TestS3Request(TestCase):

    def test_action(self):
        request = S3Request("GET", URI())
        self.assertTrue(request.action, "GET")

    def test_credentials(self):
        uri = URI()
        credentials = S3Credentials("foo", "bar")
        request = S3Request("GET", uri, credentials)
        headers = request.getHeaders()
        self.assertTrue("authorization" in headers)

    def test_noCredentials(self):
        uri = URI()
        request = S3Request("GET", uri)
        headers = request.getHeaders()
        self.assertTrue("authorization" not in headers)

    def test_dateDelete(self):
        uri = URI()
        headers = {"date": "test-date"}
        request = S3Request("GET", uri, headers=headers)
        headers = request.getHeaders()
        self.assertTrue("date" not in headers)
        self.assertTrue("x-amz-date" in headers)

    def test_dateUpdate(self):
        uri = URI()
        request = S3Request("GET", uri)
        old_date = request.getHeaders()["x-amz-date"]

        time.sleep(1)
        new_date = request.getHeaders()["x-amz-date"]
        self.assertNotEquals(old_date, new_date)


class TestS3ServerError(TestCase):

    def test_status(self):
        error = S3ServerError("<dummy />", 200)
        self.assertEquals(error.status, 200)

    def test_unknownStatus(self):
        error = S3ServerError("<dummy />", 0)
        self.assertEquals(error.reason, "")

    def test_reason(self):
        error = S3ServerError("<dummy />", 100, "Dummy Error")
        self.assertEquals(error.status, 100)
        self.assertEquals(error.reason, "Dummy Error")

    def test_noReason(self):
        error = S3ServerError("<dummy />", 200)
        self.assertEquals(error.reason, "OK")

    def test_emptyError(self):
        error = S3ServerError("<a><Error /></a>", 500)
        self.assertEquals(error.info, {})

        error = S3ServerError("<Error />", 500)
        self.assertEquals(error.info, {})

    def test_errorMessage(self):
        error = S3ServerError("<Error><Message>Dummy</Message></Error>", 500)
        self.assertEquals(error.info, {"Message": "Dummy"})

    def test_strWithCode(self):
        error = S3ServerError("<dummy />", 200)
        self.assertEquals(str(error), "200 (OK)")

    def test_strWithReason(self):
        error = S3ServerError("<dummy />", 200, "Dummy")
        self.assertEquals(str(error), "200 (Dummy)")

    def test_strWithMessage(self):
        error = S3ServerError("<Error><Message>Dummy</Message></Error>", 500)
        self.assertEquals(str(error), "500 (Internal Server Error): Dummy")


class TestS3RawConnection(TestCase):

    def test_cache(self):
        cache = LRUCache(1)
        raw_connection = S3RawConnection(cache)
        uri1 = URI("http://foo.com/baz")
        uri2 = URI("http://bar.com/baz")
        item1 = raw_connection[uri1]
        item2 = raw_connection[uri2]
        self.assertNotEquals(item1, item2)

        item3 = raw_connection[uri1]
        self.assertNotEquals(item1, item3)

    def test_getitem(self):
        raw_connection = S3RawConnection()
        uri = URI("http://foo.com/")
        item1 = raw_connection[uri]
        item2 = raw_connection[uri]
        self.assertEquals(item1, item2)

    def test_getitemPath(self):
        raw_connection = S3RawConnection()
        uri1 = URI("http://foo.com/bar")
        uri2 = URI("http://foo.com/baz")
        item1 = raw_connection[uri1]
        item2 = raw_connection[uri2]
        self.assertEquals(item1, item2)

    def test_getitemHost(self):
        raw_connection = S3RawConnection()
        uri1 = URI("http://foo.com/baz")
        uri2 = URI("http://bar.com/baz")
        item1 = raw_connection[uri1]
        item2 = raw_connection[uri2]
        self.assertNotEquals(item1, item2)

    def test_delitem(self):
        raw_connection = S3RawConnection()
        uri = URI("http://foo.com/")
        item1 = raw_connection[uri]

        del raw_connection[uri]
        item2 = raw_connection[uri]
        self.assertNotEquals(item1, item2)


class TestS3Backend(TestCaseWithFactory, TestBackendMixin):

    def setUp(self):
        super(TestS3Backend, self).setUp()

        path = self.makeFile()
        self._uri = URI("s3:///%s" % path)

    def getBackend(self):
        return FakeS3Backend(self._uri)


class TestS3Connection(TestCaseWithFactory, TestConnectionMixin):

    def setUp(self):
        super(TestS3Connection, self).setUp()

        self._endpoint = URI("http://localhost/")

    def getConnection(self):
        return FakeS3Connection(self._endpoint)


class TestS3Directory(TestCaseWithFactory, TestDirectoryMixin):

    def setUp(self):
        super(TestS3Directory, self).setUp()

        endpoint = URI("http://localhost/")
        backend = FakeS3Backend(endpoint)
        self._bucket = Bucket(backend)

    def getDirectory(self, name):
        return self._bucket[name]


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
