# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Directory",
    ]

from lpresults.storage.exceptions import ClientError
from lpresults.storage.file import File
from lpresults.storage.validators import valid_name


class Directory:
    """A Storage Directory.

    This provides an interfaces to manage files contained in a directory.
    """

    file_factory = File

    def __init__(self, bucket, name):
        """Initialize a directory.

        :param bucket: Bucket instance to use.
        :param name: Name of this directory.
        """
        self.bucket = bucket
        self.name = name

    def __contains__(self, name):
        """Check if a file name is in this directory."""
        names = self.bucket._connection.listDirectory(self.name)
        return name in names

    def __iter__(self):
        """Iterate over the file names in this Directory."""
        names = self.bucket._connection.listDirectory(self.name)
        for name in names:
            yield self.getFile(name)

    def getFile(self, name):
        """Get a file by name from this directory and iterate over the
        content by chunks.

        :param name: Name of the file to get.
        """
        if not valid_name(name):
            raise ClientError("Invalid file name: '%s'" % name)

        return self.file_factory(self, name)

    def putFile(self, name, file):
        """Put a file into this directory and read the content
        from the given file.

        :param name: Name of the file to put.
        :param file: File-like object to read content.
        """
        if not valid_name(name):
            raise ClientError("Invalid file name: '%s'" % name)

        self.bucket._connection.putContent(self.name, name, file)
        return self.file_factory(self, name)

    def deleteFile(self, name):
        """Delete a file by name from this Directory."""
        if not valid_name(name):
            raise ClientError("Invalid file name: '%s'" % name)

        return self.bucket._connection.deleteFile(self.name, name)
