# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "File",
    ]


class LazyInfo:

    def __get__(self, obj, cls=None):
        if obj is not None:
            connection = obj.directory.bucket._connection
            obj.info = connection.getInfo(obj.directory.name, obj.name)
            return obj.info

        return self


class File:
    """A Storage File.

    This provides an interfaces to manage a file.
    """
    info = LazyInfo()

    def __init__(self, directory, name):
        """Initialize a file.

        :param directory: Directory instance to use.
        :param name: Name of this file.
        """
        self.directory = directory
        self.name = name

    @property
    def size(self):
        return self.info.size

    @property
    def content_type(self):
        return self.info.content_type

    @property
    def date_modified(self):
        return self.info.date_modified

    _content = None

    def open(self):
        """Open this file for reading."""
        connection = self.directory.bucket._connection
        self._content = connection.getContent(self.directory.name, self.name)

    def read(self, chunked=False):
        """Read the whole file or chunked.

        :param chunked: Read in chunks. Defaults to False.
        """
        if not self._content:
            if chunked:
                raise RuntimeError("Can't combine autoopen with chunked")
            self.open()
            autoopen = True
        else:
            autoopen = False

        if chunked:
            try:
                rv = self._content.next()
            except StopIteration:
                rv = ""
        else:
            rv = "".join(self._content)
            if autoopen:
                self.close()

        return rv

    def close(self):
        """Close this file."""
        # Don't die with an AttributeError if the '_content' property
        # is not set.
        if self._content is not None:
            self._content = None
