# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "FunctionalLayer",
    "FunctionalStorageLayer",
    "StorageLayer",
    ]

import os
import shutil
import tempfile

from zope.component import (
    queryUtility,
    provideUtility,
    )

from lpresults.testing import layers

from lpresults.storage.zope.interfaces import IZStorage
from lpresults.storage.zope.zstorage import ZStorage


class StorageLayerFactory:
    """Sets up file storage for use in tests."""

    __bases__ = ()

    buckets = [
        "attachments",
        ]

    uri_template = "file:%s"

    def __init__(self, name, storage_id=None):
        self.__name__ = name
        if storage_id is None:
            storage_id = "storage-layer"
        self.storage_id = storage_id

    def setUp(self):
        self._old_zstorage = queryUtility(IZStorage)
        self._zstorage = ZStorage()
        self._zstorage_called = set()

        real_get = self._zstorage.get

        def get_proxy(name):
            bucket = real_get(name)
            connection = bucket._connection
            if not getattr(connection, "__put_content_tracked__", False):
                # It hasn't been overriden yet
                real_function = connection.putContent

                def proxy_function(*args, **kwargs):
                    self._zstorage_called.add(name)
                    return real_function(*args, **kwargs)

                connection.putContent = proxy_function
                connection.__put_content_tracked__ = True

            return bucket

        self._zstorage.get = get_proxy

        self.storage_path = tempfile.mkdtemp(prefix=self.storage_id)
        for bucket in self.buckets:
            path = os.path.join(self.storage_path, bucket)
            uri = self.uri_template % path
            self._zstorage.setDefaultURI(bucket, uri)

        # Provide utility for other layers.
        provideUtility(self._zstorage)

    def tearDown(self):
        if self._old_zstorage:
            provideUtility(self._old_zstorage)
        self._zstorage = self._old_zstorage = None

        if self.storage_path is not None:
            os.rmdir(self.storage_path)

    def testSetUp(self):
        provideUtility(self._zstorage)

    def testTearDown(self):
        zstorage_called = getattr(self, "_zstorage_called", ())
        for bucket in self.buckets:
            if bucket in zstorage_called:
                path = os.path.join(self.storage_path, bucket)
                shutil.rmtree(path)

        # Empty the cache to track only those buckets used by the test.
        self._zstorage_called = set()


StorageLayer = StorageLayerFactory("StorageLayer")


class FunctionalLayerFactory(layers.FunctionalLayerFactory):
    """Loads the Zope3 component architecture in appserver mode."""

    zcml = "zope/ftesting.zcml"


FunctionalLayer = FunctionalLayerFactory("FunctionalLayer")


class FunctionalStorageLayerFactory:
    """Loads functional and storage layers for use in tests."""

    __bases__ = ()

    def __init__(self, name, *args, **kwargs):
        self.__name__ = name
        self.functional = FunctionalLayerFactory(name)
        self.storage = StorageLayerFactory(name, *args, **kwargs)

    def setUp(self):
        self.functional.setUp()
        self.storage.setUp()

    def tearDown(self):
        self.functional.tearDown()
        self.storage.tearDown()

    def testSetUp(self):
        self.functional.testSetUp()
        self.storage.testSetUp()

    def testTearDown(self):
        self.functional.testTearDown()
        self.storage.testTearDown()


FunctionalStorageLayer = FunctionalStorageLayerFactory(
    "FunctionalStorageLayer",
    storage_id="functional-layer")
