# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "request_class",
    "publication_class",
    ]

import martian

from zope.component import getSiteManager
from zope.location.interfaces import ILocation
from zope.traversing.browser import AbsoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

from lpresults.storage.zope.interfaces import (
    IStorageConfiguration,
    IStorageLayer,
    IStorageRootPage,
    )
from lpresults.storage.zope.pages import RootPage
from lpresults.storage.zope.simple import (
    StorageConfiguration,
    StoragePublication,
    StorageRequest,
    StorageRootAbsoluteURL,
    )


class request_class(martian.Directive):
    """Directive used to specify a service's request class.

    The default is lpresults.storage.zope.wsgi.StorageRequest.
    """
    scope = martian.CLASS
    store = martian.ONCE
    default = StorageRequest


class publication_class(martian.Directive):
    """Directive used to specify a service's publication class.

    The default is lpresults.storage.zope.wsgi.StoragePublication.
    """
    scope = martian.CLASS
    store = martian.ONCE
    default = StoragePublication


class ConfigurationGrokker(martian.ClassGrokker):

    martian.component(StorageConfiguration)
    martian.directive(request_class)
    martian.directive(publication_class)

    def execute(self, cls, config, request_class, publication_class, *kw):
        # If implementations of the IStorageConfiguration methods are
        # missing, create them from the declarations.
        if not getattr(cls, "createRequest", None):

            def createRequest(self, body_instream, environ):
                """See `IStorageConfiguration`."""
                request = request_class(body_instream, environ)
                publication = publication_class(None)
                request.setPublication(publication)
                return request

            cls.createRequest = createRequest

        # Register as utility.
        utility = cls()
        sm = getSiteManager()
        sm.registerUtility(utility, IStorageConfiguration)

        return True


class StorageRootGrokker(martian.ClassGrokker):
    """Registers your service root as a singleton object."""
    martian.component(RootPage)

    def execute(self, cls, config, *kw):
        utility = cls()
        sm = getSiteManager()
        sm.registerUtility(utility, IStorageRootPage)

        return True


class StorageRootAbsoluteURLGrokker(martian.ClassGrokker):
    """Registers a strategy for generating your service's base URL.

    In most cases you can simply create an empty subclass of
    StorageRootAbsoluteURL.
    """
    martian.component(StorageRootAbsoluteURL)

    def execute(self, cls, config, *kw):
        getSiteManager().registerAdapter(cls)

        return True


class location_interface(martian.Directive):
    """Directive to specify the location interface.

    The default is zope.location.interfaces.ILocation
    """
    scope = martian.CLASS
    store = martian.ONCE
    default = ILocation


class AbsoluteURLGrokker(martian.ClassGrokker):
    """Registers a strategy for generating an object's URL.

    In most cases you can simply create an empty subclass of
    AbsoluteURL.
    """
    martian.component(AbsoluteURL)
    martian.directive(location_interface)

    def execute(self, cls, config, location_interface, *kw):
        getSiteManager().registerAdapter(
            cls, (location_interface, IStorageLayer), IAbsoluteURL)

        return True
