# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "IDirectory",
    "IFile",
    "IStorageConfiguration",
    "IStorageLayer",
    "IStoragePage",
    "IStorageRequest",
    "IStorageResult",
    "IStorageRootPage",
    "IStorageTraverser",
    "IZStorage",
    ]

from zope.interface import (
    Attribute,
    Interface,
    )
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces.browser import (
    IBrowserRequest,
    IDefaultBrowserLayer,
    )
from zope.publisher.interfaces.http import IResult
from zope.schema import (
    Bool,
    Int,
    TextLine,
    )


class IFile(Interface):

    name = Attribute("Name of the file.")

    size = Attribute("Size of the file.")

    content_type = Attribute("Content type of the file.")

    date_modified = Attribute("Date modified of the file.")

    def open():
        """Open this file for reading."""

    def read(chunked=False):
        """Read the whole file or chunked.

        :param chunked: Read in chunks. Defaults to False.
        """

    def close():
        """Close this file."""


class IDirectory(Interface):

    name = Attribute("Name of the directory.")

    def getFile(name):
        """Get a file by name from this directory and iterate over the
        content by chunks.

        :param name: Name of the file to get.
        """

    def putFile(name, file):
        """Put a file into this directory and read the content
        from the given file.

        :param name: Name of the file to put.
        :param file: File-like object to read content.
        """


class IStorageConfiguration(Interface):

    path_override = TextLine(
        title=u"Storage path override.",
        default=u"storage",
        description=u"The path component for web clients to use when making "
            "HTTP requests to the storage service from their current virtual "
            "host.")

    use_https = Bool(
        title=u"Storage service is secured.",
        default=True,
        description=u"Whether or not requests to the storage service are "
            "secured through SSL.")

    hostname = TextLine(
        title=u"The hostname to be used in generated URLs.",
        default=u"localhost",
        description=u"You only need to specify this if you're using the "
            "RootAbsoluteURL class from lpresults.storage.zope.simple. This "
            "is the hostname of the lpresults.storage application.")

    port = Int(
        title=u"The TCP port on which the storage service is running.",
        description=u"Used in generated URLs.",
        default=0)

    root_uri_prefix = TextLine(
        title=u"Any URL prefix necessary for the service root.",
        default=u"",
        description=u"If your storage service is not located at the root of "
            "its domain (for instance, rooted at http://foo.com/storage/ "
            "instead of http://storage.foo.com/, put the URL prefix here. "
            "(In the example case, the URL prefix is 'storage/').")


class IStorageTraverser(IPublishTraverse):
    """Marker interface for for traversal of storage objects."""

    def get(request, name):
        """Traverse to a sub-object."""


class IStoragePage(Interface):
    """An page published through HTTP."""

    def __call__():
        """Publish the object."""


class IStorageRootPage(IStoragePage):
    """A root object that also acts as a resource."""


class IStorageRequest(IBrowserRequest):
    """Interface for requests to the storage."""


class IStorageLayer(IStorageRequest, IDefaultBrowserLayer):
    """Marker interface for registering views on the storage."""


class IStorageResult(IResult):
    """A result published through HTTP."""


class IZStorage(Interface):
    """A flag interface used to lookup the ZStorage utility."""
