# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "FilePage",
    "FileResult",
    "RootPage",
    ]

from zope.interface import implements
from zope.publisher.interfaces import NotFound

from lpresults.storage.zope.interfaces import (
    IStoragePage,
    IStorageResult,
    IStorageRootPage,
    )
from lpresults.storage.zope.utils import get_current_browser_request


ROOT_PAGE = """\
<html>
<body>
<h1>Launchpad Storage Extension</h1>
<p>
Launchpad extension to access file storage.
</p>
</body></html>"""


class PageMixin:
    """See `IStoragePage`."""
    implements(IStoragePage)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def traverseName(self, request, name):
        raise NotFound(self, name)

    def browserDefault(self, request):
        return self, ()

    def __call__(self, request=None):
        """Handle a GET request."""
        method = self.request.method
        if method == "GET":
            result = self.do_GET()
        else:
            request.response.setStatus(405)
            request.response.setHeader("Allow", "GET")
            result = ""

        return result

    def do_GET(self):
        """Retrieve the value of the page.

        :return: A string representation.
        """
        raise NotImplementedError()


class RootPage(PageMixin):

    implements(IStorageRootPage)

    def __init__(self):
        """Initialize the page.

        The root constructor is different from other PageMixin
        constructors because Zope initializes the object with no request
        or context, and then passes the request in when it calls the
        service root object.
        """

    @property
    def request(self):
        """Get the current browser request."""
        return get_current_browser_request()

    def do_GET(self):
        """See PageMixin."""
        self.request.response.setHeader("Content-Type", "text/html")
        return ROOT_PAGE


class FilePage(PageMixin):

    def do_GET(self):
        """See PageMixin."""
        file = self.context
        self.request.response.setHeader("Content-Type", file.content_type)
        self.request.response.setHeader("Length", file.size)

        return file


class FileResult:
    """See `IStorageResult`."""
    implements(IStorageResult)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __iter__(self):
        """See `IResult`."""
        file = self.context
        file.open()
        try:
            while True:
                chunk = file.read(chunked=True)
                if not chunk:
                    break

                yield chunk
        finally:
            file.close()
