# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "StoragePublicationMixin",
    "StorageRequestMixin",
    ]

from zope.component import (
    getUtility,
    queryMultiAdapter,
    )
from zope.interface import implements
from zope.publisher.interfaces import NotFound

from lpresults.storage.zope.interfaces import (
    IStorageConfiguration,
    IStoragePage,
    IStorageRequest,
    )


class StoragePublicationMixin:
    """Mixin for storage publication."""

    def getDefaultTraversal(self, request, ob):
        """See `zope.publisher.interfaces.browser.IBrowserPublication`.

        The Storage publication doesn't use the getDefaultTraversal()
        extension mechanism, because it only applies to GET, HEAD, and
        POST methods.

        See getPage() for the alternate mechanism.
        """
        # Don't traverse to anything else.
        return ob, None

    def getPage(self, request, ob):
        """Return the page that can publish the object ob.

        This is done at the end of traversal.  If the published object
        supports the IStoragePage, the corresponding adapter is
        returned.
        """
        if IStoragePage.providedBy(ob):
            return ob
        else:
            # self is marker
            page = queryMultiAdapter(
                (ob, request), IStoragePage, default=self)
            if page is not self:
                return page
            else:
                raise NotFound(ob, "")


class StorageRequestMixin:
    """Mixin providing storage page wrapping in traversal.

    This should be mixed in the request using to the base publication used.
    """
    implements(IStorageRequest)

    def traverse(self, ob):
        self._removeVirtualHostTraversals()

        ob = self.publication.getApplication(self)
        result = super(StorageRequestMixin, self).traverse(ob)
        return self.publication.getPage(self, result)

    def _removeVirtualHostTraversals(self):
        names = []
        config = getUtility(IStorageConfiguration)
        if config.path_override is not None:
            name = self._popTraversal(config.path_override)
            if name is not None:
                names.append(name)

        self.setVirtualHostRoot(names=names)

    def _popTraversal(self, name=None):
        """Remove a name from the traversal stack, if it is present.

        :name: The string to look for in the stack, or None to accept
        any string.

        :return: The name of the element removed, or None if the stack
            wasn't changed.
        """
        stack = self.getTraversalStack()
        if len(stack) > 0 and (name is None or stack[-1] == name):
            item = stack.pop()
            self.setTraversalStack(stack)
        else:
            item = None

        return item
