# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "StorageConfiguration",
    "StoragePublication",
    "StorageRequest",
    "StorageRoot",
    "StorageRootAbsoluteURL",
    ]

import traceback

from zope.component import (
    adapts,
    getUtility,
    queryMultiAdapter,
    )
from zope.interface import implements
from zope.publisher.browser import BrowserRequest
from zope.publisher.interfaces import (
    IPublication,
    IPublishTraverse,
    NotFound,
    )
from zope.publisher.interfaces.http import IHTTPRequest
from zope.publisher.publish import mapply
from zope.security.management import (
    endInteraction,
    newInteraction,
    )
from zope.traversing.browser.interfaces import IAbsoluteURL

import grokcore.component

from lpresults.storage.uri import URI
from lpresults.storage.zope.interfaces import (
    IStorageConfiguration,
    IStorageLayer,
    IStorageRootPage,
    )
from lpresults.storage.zope.pages import RootPage
from lpresults.storage.zope.publisher import (
    StoragePublicationMixin,
    StorageRequestMixin,
    )
from lpresults.storage.zope.traversers import RootTraverser
from lpresults.storage.zope.utils import implement_from_dict


class PublicationMixin:
    """A very simple implementation of `IPublication`.

    The object passed to the constructor is returned by getApplication().
    """
    implements(IPublication)

    def __init__(self, application):
        """Create the test publication.

        The object at which traversal should start is passed as parameter.
        """
        self.application = application

    def getApplication(self, request):
        """Returns the application passed to the constructor."""
        return self.application

    def beforeTraversal(self, request):
        """Sets the request as the current interaction."""
        endInteraction()
        newInteraction(request)

    def afterTraversal(self, request, ob):
        """Does nothing."""

    def callTraversalHooks(self, request, ob):
        """Does nothing."""

    def traverseName(self, request, ob, name):
        """Traverse by looking for an `IPublishTraverse` adapter."""
        # This is copied from PublicationTraverse. This should really
        # live in zope.publisher, we are copying because we don't want to
        # depend on zope.app stuff. Namespace support was dropped.
        if name == ".":
            return ob

        if IPublishTraverse.providedBy(ob):
            return ob.publishTraverse(request, name)
        else:
            # self is marker.
            adapter = queryMultiAdapter(
                (ob, request), IPublishTraverse, default=self)
            if adapter is not self:
                return adapter.publishTraverse(request, name)
            else:
                raise NotFound(ob, name, request)

    def callObject(self, request, ob):
        """Call the object, returning the result."""
        return mapply(ob, request.getPositionalArguments(), request)

    def afterCall(self, request, ob):
        """Does nothing."""

    def handleException(self, object, request, exc_info, retry_allowed=1):
        """Prints the exception."""
        # Reproduce the behavior of ZopePublication by looking up a view
        # for this exception.
        exception = exc_info[1]
        view = queryMultiAdapter((exception, request), name='index.html')
        if view is not None:
            request.response.reset()
            request.response.setResult(view())
        else:
            traceback.print_exception(*exc_info)

    def endRequest(self, request, ob):
        """Ends the interaction."""
        endInteraction()


class StoragePublication(StoragePublicationMixin, PublicationMixin):
    """A simple publication.

    Combines the IPublication implementation of PublicationMixin
    with the storage implementation of StoragePublicationMixin,
    """


class StorageRequest(StorageRequestMixin, BrowserRequest):
    """A simple storage request with no special features."""

    implements(IStorageLayer)


class StorageRoot(RootPage, RootTraverser):
    """A simple root page and traverser."""

    grokcore.component.provides(IStorageRootPage)


class StorageRootAbsoluteURL(URI):
    """A basic implementation of `IAbsoluteURL` for a storage root."""

    implements(IAbsoluteURL)
    adapts(StorageRoot, IHTTPRequest)

    def __init__(self, context, request):
        """Initialize with respect to a context and request."""
        super(StorageRootAbsoluteURL, self).__init__()

        config = getUtility(IStorageConfiguration)
        if config.use_https:
            self.scheme = "https"
        else:
            self.scheme = "http"

        self.host = config.hostname

        if (config.port
            and (self.scheme != "https" or config.port != 443)
            and (self.scheme != "http" or config.port != 80)):
            self.port = config.port

        if config.root_uri_prefix:
            self.path = config.root_uri_prefix
            if self.path.endswith("/"):
                self.path = self.path[:-1]

    def __call__(self):
        """Return the URL to the storage root.

        This value is called when finding the URL to the storage root,
        and it needs to end with a slash so that clients will be able
        to do relative URL resolution correctly with respect to the root.
        """
        return str(self) + "/"


StorageConfiguration = implement_from_dict(
    "StorageConfiguration", IStorageConfiguration)
