# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import threading

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from zope.component import (
    getUtility,
    provideUtility,
    )

from lpresults.storage.bucket import Bucket
from lpresults.storage.exceptions import StorageError

from lpresults.storage.zope.interfaces import IZStorage
from lpresults.storage.zope.zstorage import ZStorage


class TestZStorage(TestCase):

    def setUp(self):
        self.zstorage = ZStorage()

    def tearDown(self):
        self.zstorage._reset()

    def test_noName(self):
        self.assertRaises(StorageError, self.zstorage.get, "name")

    def test_createUnnamed(self):
        bucket = self.zstorage.create(None, "file:///foo")
        self.assertTrue(isinstance(bucket, Bucket))

    def test_createNamed(self):
        bucket = self.zstorage.create("name", "file:///foo")
        self.assertTrue(self.zstorage.get("name") is bucket)

    def test_createNamedTwice(self):
        self.zstorage.create("name", "file:///foo")
        self.assertRaises(
            StorageError, self.zstorage.create, "name", "file:///foo")

    def test_createNamedInThread(self):
        self.zstorage.create("name", "file:///foo")

        raised = []

        def f():
            try:
                self.zstorage.get("name")
            except StorageError:
                raised.append(True)

        thread = threading.Thread(target=f)
        thread.start()
        thread.join()

        self.assertTrue(raised)

    def test_createDefault(self):
        self.zstorage.setDefaultURI("name", "file:///foo")
        bucket = self.zstorage.create("name")
        self.assertTrue(isinstance(bucket, Bucket))

    def test_createDefaultTwice(self):
        self.zstorage.setDefaultURI("name", "file:///foo")
        self.zstorage.create("name")
        self.assertRaises(StorageError, self.zstorage.create, "name")

    def test_getName(self):
        bucket = self.zstorage.create("name", "file:///foo")
        self.assertEquals(self.zstorage.getName(bucket), "name")

    def test_getWithURI(self):
        bucket = self.zstorage.get("name", "file:///foo")
        self.assertTrue(isinstance(bucket, Bucket))
        self.assertTrue(self.zstorage.get("name") is bucket)
        self.assertTrue(self.zstorage.get("name", "file:///foo") is bucket)

    def test_iterbuckets(self):
        bucket1 = self.zstorage.create(None, "file:///foo")
        bucket2 = self.zstorage.create(None, "file:///foo")
        bucket3 = self.zstorage.create("name", "file:///foo")
        buckets = []
        for name, bucket in self.zstorage.iterbuckets():
            buckets.append((name, bucket))
        self.assertEquals(len(buckets), 3)
        self.assertEquals(set(buckets),
                          set([(None, bucket1), (None, bucket2),
                               ("name", bucket3)]))

    def test_remove(self):
        bucket = self.zstorage.create("name", "file:///foo")
        self.assertEquals(self.zstorage.getName(bucket), "name")
        self.zstorage.remove(bucket)
        self.assertEquals(self.zstorage.getName(bucket), None)

    def test_setDefaultURI(self):
        self.zstorage.setDefaultURI("name", "file:///foo")
        bucket = self.zstorage.get("name")
        self.assertTrue(isinstance(bucket, Bucket))

    def test_getDefaultURIs(self):
        self.zstorage.setDefaultURI("name1", "file:///foo1")
        self.zstorage.setDefaultURI("name2", "file:///foo2")
        self.zstorage.setDefaultURI("name3", "file:///foo3")
        default_uris = self.zstorage.getDefaultURIs()
        self.assertEquals(default_uris, {
            "name1": "file:///foo1",
            "name2": "file:///foo2",
            "name3": "file:///foo3",
            })


class TestZStorageUtility(TestCase):

    def test_utility(self):
        provideUtility(ZStorage())
        self.assertTrue(isinstance(getUtility(IZStorage), ZStorage))


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
