# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "get_current_browser_request",
    "get_current_storage_request",
    "implement_from_dict",
    ]

from zope.interface import classImplements
from zope.schema import getFieldsInOrder
from zope.security.management import queryInteraction
from zope.publisher.interfaces.http import IHTTPApplicationRequest

from lpresults.storage.zope.interfaces import IStorageRequest


def get_current_browser_request():
    """Return the current browser request, looked up from the interaction.

    :return: A request that provides IHTTPApplicationRequest.
    """
    interaction = queryInteraction()
    if interaction is None:
        return None
    requests = [
        participation
        for participation in interaction.participations
        if IHTTPApplicationRequest.providedBy(participation)]
    if not requests:
        return None

    assert len(requests) == 1, (
        "We expect only one IHTTPApplicationRequest in the interaction."
        " Got %s." % len(requests))
    return requests[0]


def get_current_storage_request():
    """Return the current storage request.

    :return: An object providing IStorageRequest.
    """
    request = get_current_browser_request()
    if request is not None:
        request = IStorageRequest(request)

    return request


def implement_from_dict(class_name, interface, values={}, superclass=object):
    """Return a class that implements an interface's attributes.

    :param interface: The interface to implement.
    :param superclass: The superclass of the class to be generated.
    :param values: A dict of values to use when generating the class.
    :return: A class that implements 'interface'. Any attributes given
             values in 'values' will have those values in the generated
             interface. Other attributes will have their default
             values as defined in the interface. Attributes with no
             default values, will not be present.
    """
    class_dict = {}
    for name, field in getFieldsInOrder(interface):
        if field.default is not None:
            class_dict[name] = field.default
    class_dict.update(values)

    if not isinstance(superclass, tuple):
        superclass = (superclass,)
    new_class = type(class_name, superclass, class_dict)
    classImplements(new_class, interface)
    return new_class
