# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "ZStorage",
    "global_zstorage",
    ]

import weakref

from threading import local

from zope.interface import implements

from lpresults.storage.backend import create_backend
from lpresults.storage.bucket import Bucket
from lpresults.storage.exceptions import StorageError

from lpresults.storage.zope.interfaces import IZStorage


class ZStorage:
    """A utility which integrates Storage with Zope."""

    implements(IZStorage)

    _backends = {}

    def __init__(self):
        self._local = local()
        self._default_backends = {}
        self._default_uris = {}

    def _reset(self):
        for name, bucket in list(self.iterbuckets()):
            self.remove(bucket)

        self._local = local()
        self._backends.clear()
        self._default_backends.clear()
        self._default_uris.clear()

    @property
    def _buckets(self):
        try:
            return self._local.buckets
        except AttributeError:
            buckets = weakref.WeakValueDictionary()
            return self._local.__dict__.setdefault("buckets", buckets)

    @property
    def _named(self):
        try:
            return self._local.named
        except AttributeError:
            return self._local.__dict__.setdefault("named", {})

    @property
    def _name_index(self):
        try:
            return self._local.name_index
        except AttributeError:
            return self._local.__dict__.setdefault(
                "name_index", weakref.WeakKeyDictionary())

    def _getBackend(self, uri):
        backend = self._backends.get(uri)
        if backend is None:
            return self._backends.setdefault(uri, create_backend(uri))

        return backend

    def getDefaultURIs(self):
        """Return a list of name, uri tuples that are named as the default
        backends for those names.
        """
        return self._default_uris.copy()

    def setDefaultURI(self, name, default_uri):
        """Set the default URI for buckets by name."""
        self._default_backends[name] = self._getBackend(default_uri)
        self._default_uris[name] = default_uri

    def create(self, name, uri=None):
        """Create a new bucket.

        :param name: Name of the bucket.
        :param URI: Optionally, URI to use if no default.
        """
        if uri is None:
            backend = self._default_backends.get(name)
            if backend is None:
                raise StorageError("Bucket named '%s' not found" % name)
        else:
            backend = self._getBackend(uri)

        if name is not None and self._named.get(name) is not None:
            raise StorageError("Bucket named '%s' already exists" % name)

        bucket = Bucket(backend)
        self._buckets[id(bucket)] = bucket
        if name is not None:
            self._named[name] = bucket
        self._name_index[bucket] = name

        return bucket

    def get(self, name, default_uri=None):
        """Get the bucket by name, creating it first if necessary.

        :param name: Name of the bucket to get.
        :param default_uri: Optionally, URI to use if no default.
        """
        bucket = self._named.get(name)
        if not bucket:
            return self.create(name, default_uri)

        return bucket

    def getName(self, bucket):
        """Returns the name for the given bucket or None if one isn't
        available."""
        return self._name_index.get(bucket)

    def remove(self, bucket):
        """Remove the given bucket from ZStorage."""
        del self._buckets[id(bucket)]
        name = self._name_index[bucket]
        del self._name_index[bucket]
        if name in self._named:
            del self._named[name]

    def iterbuckets(self):
        """Iterate over name, bucket tuples."""
        for bucket, name in self._name_index.items():
            yield name, bucket


global_zstorage = ZStorage()
