# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "YUITestCase",
    "build_yui_unittest_suite",
    ]

import os
import simplejson

from unittest import (
    TestCase,
    TestSuite,
    )


class YUITestCase(TestCase):

    js_timeout = 30000

    TIMEOUT = object()
    MISSING_REPORT = object()

    _yui_results = None

    def __init__(self):
        super(YUITestCase, self).__init__("checkResults")

    def initialize(self, test_path):
        self.test_path = test_path

    def setUp(self):
        super(YUITestCase, self).setUp()
        # html5browser imports from the gir/pygtk stack which causes
        # twisted tests to break because of gtk's initialize.
        from html5browser import Browser

        browser = Browser()
        html_uri = "file://%s" % os.path.abspath(self.test_path)
        page = browser.load_page(html_uri, timeout=self.js_timeout)
        if page.return_code == page.CODE_FAIL:
            self._yui_results = self.TIMEOUT
            return
        # Data["type"] is complete (an event).
        # Data["results"] is a dict (type=report)
        # with 1 or more dicts (type=testcase)
        # with 1 for more dicts (type=test).
        report = simplejson.loads(page.content)
        if report.get("type", None) != "complete":
            # Did not get a report back.
            self._yui_results = self.MISSING_REPORT
            return
        self._yui_results = {}
        for key, value in report["results"].items():
            if isinstance(value, dict) and value["type"] == "testcase":
                testcase_name = key
                test_case = value
                for key, value in test_case.items():
                    if isinstance(value, dict) and value["type"] == "test":
                        test_name = "%s.%s" % (testcase_name, key)
                        test = value
                        self._yui_results[test_name] = dict(
                            result=test["result"], message=test["message"])

    def checkResults(self):
        """Check the results.

        The tests are run during `setUp()`, but failures need to be reported
        from here.
        """
        if self._yui_results == self.TIMEOUT:
            self.fail("js timed out.")
        elif self._yui_results == self.MISSING_REPORT:
            self.fail("The data returned by js is not a test report.")
        elif self._yui_results is None or len(self._yui_results) == 0:
            self.fail("Test harness or js report format changed.")
        failures = []
        for test_name in self._yui_results:
            result = self._yui_results[test_name]
            if result["result"] != "pass":
                failures.append(
                    "Failure in %s.%s: %s" % (
                    self.test_path, test_name, result["message"]))
        self.assertEqual([], failures, "\n".join(failures))


def build_yui_unittest_suite(app_testing_path, yui_test_class):
    suite = TestSuite()
    testing_path = os.path.abspath(app_testing_path)
    unit_test_names = [
        file_name for file_name in os.listdir(testing_path)
        if file_name.startswith("test_") and file_name.endswith(".html")]
    for unit_test_name in unit_test_names:
        test_path = os.path.join(app_testing_path, unit_test_name)
        test_case = yui_test_class()
        test_case.initialize(test_path)
        suite.addTest(test_case)
    return suite
