# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "OAuthBackend",
    ]

from oauth.oauth import OAuthRequest

from zope.component import getUtility

from lazr.restful.simple import Request as LazrRequest

from lpresults.registry.models.anonymous import Anonymous

from lpresults.registry.interfaces.oauth import (
    IOAuthDataStore,
    OAuthPermission,
    OAUTH_CHALLENGE,
    )


class OAuthBackend:

    def authenticate(self, **credentials):
        """Authenticate the user based on an OAuth request."""

        request = credentials.get("request")
        if request is None:
            return None

        if not isinstance(request, LazrRequest):
            return None

        form = get_oauth_authorization(request)
        consumer_key = form.get("oauth_consumer_key")
        consumers = getUtility(IOAuthDataStore)
        consumer = consumers.lookupConsumer(consumer_key)
        token_key = form.get("oauth_token")
        anonymous_request = (token_key == "")

        if consumer_key is None:
            # Either the client's OAuth implementation is broken, or
            # the user is trying to make an unauthenticated request
            # using wget or another OAuth-ignorant application.
            # Try to retrieve a consumer based on the User-Agent
            # header.
            anonymous_request = True
            consumer_key = request.getHeader("User-Agent", "")
            if consumer_key == "":
                return None

            consumer = consumers.lookupConsumer(consumer_key)

        if consumer is None:
            if anonymous_request:
                # This is the first time anyone has tried to make an
                # anonymous request using this consumer name (or user
                # agent). Dynamically create the consumer.
                #
                # In the normal website this wouldn't be possible
                # because GET requests have their transactions rolled
                # back. But webservice requests always have their
                # transactions committed so that we can keep track of
                # the OAuth nonces and prevent replay attacks.
                if consumer_key == "" or consumer_key is None:
                    return None

                consumer = consumers.createConsumer(consumer_key)
            else:
                # An unknown consumer can never make a non-anonymous
                # request, because access tokens are registered with a
                # specific, known consumer.
                return None

        if anonymous_request:
            # Skip the OAuth verification step and let the user access the
            # web service as an unauthenticated user.
            return Anonymous(consumer)

        token = consumer.getAccessToken(token_key)
        if token is None:
            return None

        if token.permission == OAuthPermission.UNAUTHORIZED:
            return None
        elif not check_oauth_signature(request, consumer, token):
            return None

        return token.person


def get_oauth_authorization(request):
    """Retrieve OAuth authorization information from a request.

    The authorization information may be in the Authorization header,
    or it might be in the query string or entity-body.

    :return: a dictionary of authorization information.
    """
    header = request._auth
    if header is not None and header.startswith("OAuth "):
        return OAuthRequest._split_header(unicode(header))
    else:
        return request.form


def check_oauth_signature(request, consumer, token):
    """Check that the given OAuth request is correctly signed.

    If the signature is incorrect or its method is not supported, set the
    appropriate status in the request's response and return False.
    """
    authorization = get_oauth_authorization(request)

    if authorization.get("oauth_signature_method") != "PLAINTEXT":
        # XXX: Only the PLAINTEXT method is supported now.
        request.response.setStatus(400)
        return False

    if token is not None:
        token_secret = token.secret
    else:
        token_secret = ""

    expected_signature = "&".join([consumer.secret, token_secret])
    if expected_signature != authorization.get("oauth_signature"):
        request.unauthorized(OAUTH_CHALLENGE)
        return False

    return True
