# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "IAttachment",
    "IHasAttachments",
    "IAttachmentTarget",
    "IAttachmentSet",
    ]

from zope.interface import Interface
from zope.schema import (
    Bytes,
    TextLine,
    )

from lazr.restful.declarations import (
    call_with,
    collection_default_content,
    export_as_webservice_collection,
    export_as_webservice_entry,
    exported,
    export_read_operation,
    operation_parameters,
    operation_returns_collection_of,
    REQUEST_USER,
    )
from lazr.restful.fields import (
    CollectionField,
    Reference,
    )
from lazr.restful.frameworks.django import IDjangoLocation
from lazr.restful.simple import IMultiplePathPartLocation

from lpresults.tracker.interfaces.file import IFile, IFileSet
from lpresults.tracker.validators import attachment_size_validator


class IAttachment(IMultiplePathPartLocation, IFile):
    """A file attachment."""

    export_as_webservice_entry()

    data = exported(
        Bytes(
            title=u"The attachment content",
            required=True,
            readonly=True))

    target = exported(
        Reference(
            schema=Interface,  # IAttachmentTarget
            title=u"Target",
            required=True,
            readonly=True,
            description=u"Target which contains attachments."))


class IHasAttachments(Interface):
    """An entity which has a collection of attachments."""

    attachments = exported(
        CollectionField(
            title=u"List of attachments.",
            value_type=Reference(schema=IAttachment),
            readonly=True))


class IAttachmentTarget(IHasAttachments):
    """An entity on which an attachment can be created.

    Examples include just an ITestRun for now.
    """
    export_as_webservice_entry()

    @call_with(person=REQUEST_USER)
    @operation_parameters(
        filename=TextLine(),
        data=Bytes(constraint=attachment_size_validator),
        )
    @export_read_operation()
    def addAttachment(person, filename, data):
        """Attach a file to this target.

        :param person: An IPerson.
        :param filename: A string.
        :param data: A file-like object, or a `str`.
        """


class IAttachmentSet(IDjangoLocation, IFileSet):

    export_as_webservice_collection(IAttachment)

    @collection_default_content()
    @operation_returns_collection_of(IAttachment)
    @export_read_operation()
    def getAll():
        """Return the default set of attachments without filtering."""


IAttachment["target"].schema = IAttachmentTarget
