# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestResultStatus",
    "ITestResult",
    "IHasTestResults",
    "ITestResultTarget",
    "ITestResultSet",
    "TestResultAddParams",
    "TestResultRemoveParams",
    "TestResultSearchParams",
    "TEST_RESULT_PASSED_STATUSES",
    "TEST_RESULT_FAILED_STATUSES",
    "TEST_RESULT_SKIPPED_STATUSES",
    ]

from zope.interface import Interface
from zope.schema import (
    Choice,
    Datetime,
    Int,
    Text,
    TextLine,
    )

from lazr.enum import (
    DBEnumeratedType,
    DBItem,
    )
from lazr.restful.declarations import (
    call_with,
    export_as_webservice_entry,
    export_read_operation,
    exported,
    operation_parameters,
    operation_returns_collection_of,
    REQUEST_USER,
    )
from lazr.restful.frameworks.django import IDjangoLocation
from lazr.restful.interface import copy_field

from lpresults.registry.interfaces.date import (
    IHasDateCreated,
    IHasDateDeleted,
    )
from lpresults.registry.interfaces.search import ISearchSet

from lpresults.tracker.interfaces.testcase import (
    ITestCase,
    TestCaseGetParams,
    TestCaseSearchParams,
    )


class TestResultStatus(DBEnumeratedType):
    """Result Status

    The various possible states for a result.
    """
    PASS = DBItem(0, u"""
        Pass

        A test has succeeded. That is, it demonstrated that the assertion
        is true.
        """)

    FAIL = DBItem(1, u"""
        Fail

        A test has produced the bug it was intended to capture. That is,
        it has demonstrated that the assertion is false.
        """)

    UNINITIATED = DBItem(2, u"""
        Uninitiated

        The particular test in question did not start to execute.
        """)

    UNRESOLVED = DBItem(3, u"""
        Unresolved

        A test produced indeterminate results; this outcome requires
        that a human being go over results, to determine if the test
        should have passed or failed.
        """)

    UNSUPPORTED = DBItem(4, u"""
        Unsupported

        An optional feature is not available or not supported in the
        implementation under test.
        """)

    UNTESTED = DBItem(5, u"""
        Untested

        There is no test for this assertion. This is a placeholder,
        used when there is no real test case yet.
        """)

    XFAIL = DBItem(6, u"""
        Expected fail

        A bug in the environment running the test is causing a known
        false assertion. This is outside the control of the test.
        """)

    @classmethod
    def get_enum(cls, title):
        """Find the matching enum value for the status title specified."""
        for value in cls.items:
            if value.title == title:
                return value
        else:
            return cls.UNINITIATED


class ITestResult(
    IDjangoLocation, ITestCase, IHasDateCreated, IHasDateDeleted):
    """A test result consists of the output from executing a test which
    makes a specific assertion."""

    export_as_webservice_entry()

    name = exported(
        TextLine(
            title=u"Name",
            description=u"The name of the test case.",
            required=True,
            readonly=True))

    status = exported(
        Choice(
            title=u"Status",
            readonly=True,
            vocabulary=TestResultStatus,
            default=TestResultStatus.UNINITIATED))

    output = exported(
        Text(
            title=u"Output",
            readonly=True))

    date_created = exported(
        Datetime(
            title=u"Date created",
            readonly=True,
            description=u"The date on which this test result was created."))

    date_deleted = exported(
        Datetime(
            title=u"Date deleted",
            readonly=True,
            description=u"The date on which this test result was deleted."))


class IHasTestResults(Interface):
    """An entity which has a collection of test results."""

    @call_with(person=REQUEST_USER)
    @operation_parameters(
        name=TextLine(
            title=u"Search test results by name",
            required=False),
        status=Choice(
            title=u"Search test results for a given status",
            vocabulary=TestResultStatus,
            required=False),
        date_created=Datetime(
            title=u"Search test results created on the given date",
            required=False),
        search_text=TextLine(
            title=u"Search test results for the given text",
            required=False),
        )
    @operation_returns_collection_of(ITestResult)
    @export_read_operation()
    def searchTestResults(
        person, name=None, status=None, date_created=None,
        search_text=None):
        """Search the ITestResults reported on this entity.

        Return an iterable of matching results.
        """


class ITestResultTarget(IHasTestResults):
    """An entity on which a test result can be added.

    Examples include just an ITestRun for now.
    """
    pass_count = exported(
        Int(
            title=TestResultStatus.PASS.title,
            description=TestResultStatus.PASS.description,
            readonly=True))
    fail_count = exported(
        Int(
            title=TestResultStatus.FAIL.title,
            description=TestResultStatus.FAIL.description,
            readonly=True))
    uninitiated_count = exported(
        Int(
            title=TestResultStatus.UNINITIATED.title,
            description=TestResultStatus.UNINITIATED.description,
            readonly=True))
    unresolved_count = exported(
        Int(
            title=TestResultStatus.UNRESOLVED.title,
            description=TestResultStatus.UNRESOLVED.description,
            readonly=True))
    unsupported_count = exported(
        Int(
            title=TestResultStatus.UNSUPPORTED.title,
            description=TestResultStatus.UNSUPPORTED.description,
            readonly=True))
    untested_count = exported(
        Int(
            title=TestResultStatus.UNTESTED.title,
            description=TestResultStatus.UNTESTED.description,
            readonly=True))
    xfail_count = exported(
        Int(
            title=TestResultStatus.XFAIL.title,
            description=TestResultStatus.XFAIL.description,
            readonly=True))

    # XXX: Use read operation instead of write or factory to prevent
    #      lazr.restfulclient from making two calls per operation.
    @call_with(person=REQUEST_USER)
    @operation_parameters(
        name=copy_field(ITestResult["name"]),
        status=copy_field(ITestResult["status"]),
        output=copy_field(ITestResult["output"]),
        date_created=copy_field(ITestResult["date_created"]),
        )
    @export_read_operation()
    def addTestResult(
        person, name, status=None, output=None,
        date_created=None):
        """Add a new test result on this target.

        :param name: Test name.
        :param status: New status from `TestResultStatus`.
        :param output: Output generated from the result.
        :param date_created: Timestamp when the result was created.
        """

    @call_with(person=REQUEST_USER)
    @operation_parameters(
        name=copy_field(ITestResult["name"]),
        date_created=copy_field(ITestResult["date_created"]),
        )
    @export_read_operation()
    def removeTestResult(person, name, date_created=None):
        """Remove a test result from this target.

        :param name: Test name.
        :param date_created: Timestamp when the result was created.
        """


class ITestResultSet(ISearchSet):

    def add(add_params):
        """Add a test result to the `ITarget` defined in the params and
        return nothing.

        :param add_params: A TestResultAddParams object.
        """

    def remove(remove_params):
        """Remove a test result from the `ITarget` defined in the params
        and return nothing. If no test result is found, nothing happens.

        :param remove_params: A TestResultRemoveParams object.
        """

    def search(search_params):
        """Search for test results.

        :param search_params: A TestResultSearchParams object.
        """


class TestResultAddParams(TestCaseGetParams):

    def __init__(
        self, person, name, status=None, output=None, date_created=None,
        **kwargs):

        super(TestResultAddParams, self).__init__(name, **kwargs)
        self.person = person
        self.status = status
        self.output = output
        self.date_created = date_created


class TestResultRemoveParams(TestCaseGetParams):

    def __init__(self, person, name, date_created=None, **kwargs):

        super(TestResultRemoveParams, self).__init__(name, **kwargs)
        self.person = person
        self.date_created = date_created


class TestResultSearchParams(TestCaseSearchParams):
    """Encapsulates search parameters for ITestResultSet.search()"""

    def __init__(self, person=None, status=None, date_created=None, **kwargs):
        super(TestResultSearchParams, self).__init__(**kwargs)
        self.person = person
        self.status = status
        self.date_created = date_created


TEST_RESULT_PASSED_STATUSES = (
    TestResultStatus.PASS,
    )
TEST_RESULT_FAILED_STATUSES = (
    TestResultStatus.FAIL,
    TestResultStatus.UNRESOLVED,
    )
TEST_RESULT_SKIPPED_STATUSES = (
    TestResultStatus.UNINITIATED,
    TestResultStatus.UNSUPPORTED,
    TestResultStatus.UNTESTED,
    TestResultStatus.XFAIL,
    )
