# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestRunDateCreatedExists",
    "ITestRun",
    "IHasTestRuns",
    "ITestRunTarget",
    "ITestRunSet",
    "TestRunCreateParams",
    "TestRunSearchParams",
    ]

from zope.interface import Interface
from zope.schema import (
    Datetime,
    Int,
    TextLine,
    )

from lazr.restful.declarations import (
    call_with,
    export_as_webservice_entry,
    exported,
    export_factory_operation,
    export_read_operation,
    operation_parameters,
    operation_returns_collection_of,
    webservice_error,
    REQUEST_USER,
    )
from lazr.restful.fields import Reference
from lazr.restful.frameworks.django import IDjangoLocation
from lazr.restful.simple import IMultiplePathPartLocation

from lpresults.registry.interfaces.date import IHasDateCreated
from lpresults.registry.interfaces.search import (
    ISearchSet,
    SearchParams,
    )

from lpresults.tracker.interfaces.attachment import IAttachmentTarget
from lpresults.tracker.interfaces.testperson import ITestPerson
from lpresults.tracker.interfaces.testresult import (
    IHasTestResults,
    ITestResultTarget,
    )
from lpresults.tracker.interfaces.testsystemunit import ITestSystemUnitTarget
from lpresults.tracker.interfaces.testtarget import (
    IHasTestTarget,
    ITestTarget,
    TestTargetParamsMixin,
    )


class TestRunDateCreatedExists(Exception):
    """Raised when creating a test run with a date that already exists."""

    webservice_error(400)


class ITestRun(
    IMultiplePathPartLocation, ITestResultTarget, ITestSystemUnitTarget,
    IHasTestTarget, IHasDateCreated, IAttachmentTarget):
    """A test run for a particular target."""

    export_as_webservice_entry()

    sequence = exported(
        Int(
            title=u"Sequence",
            required=True,
            readonly=True,
            description=u"Sequence number identifying the test run."))

    date_created = exported(
        Datetime(
            title=u"Date created",
            required=True,
            readonly=True,
            description=u"Date on which this test run was created."))

    person = exported(
        Reference(
            schema=ITestPerson,
            title=u"Person",
            required=True,
            readonly=True,
            description=u"Person who ran the tests."))

    target = exported(
        Reference(
            schema=Interface,  # ITestRunTarget
            title=u"Target",
            required=True,
            readonly=True,
            description=u"Target on which tests were run."))


class IHasTestRuns(Interface):
    """An entity which has a collection of test runs."""

    @call_with(person=REQUEST_USER, limit=5)
    @operation_returns_collection_of(ITestRun)
    @export_read_operation()
    def getLatestTestRuns(person, limit=None):
        """Get the latest test runs."""

    @call_with(user=REQUEST_USER)
    @operation_parameters(
        sequence=Int(
            title=u"Search test runs for a sequence",
            required=False),
        date_created=Datetime(
            title=u"Search test runs created on the given date",
            required=False),
        created_before=Datetime(
            title=u"Search test runs created before the given date",
            required=False),
        created_after=Datetime(
            title=u"Search test runs created after the given date",
            required=False),
        search_text=TextLine(
            title=u"Search test runs for the given text",
            required=False),
        )
    @operation_returns_collection_of(ITestRun)
    @export_read_operation()
    def searchTestRuns(
        user, person=None, sequence=None, date_created=None,
        created_before=None, created_after=None, search_text=None):
        """Search the ITestRuns reported on this entity.

        Return an iterable of matching results.
        """


class ITestRunTarget(ITestTarget, IHasTestRuns, IHasTestResults):
    """An entity on which a test run can be created.

    Examples include an ITestDistroSeries and ITestProjectSeries for now.
    """
    export_as_webservice_entry()

    @call_with(person=REQUEST_USER)
    @export_factory_operation(ITestRun, ["date_created"])
    def createTestRun(person, date_created=None):
        """Create a new test run on this target."""


class ITestRunSet(IDjangoLocation, ISearchSet):

    def create(create_params):
        """Create a test run and return it."""


class TestRunCreateParams(TestTargetParamsMixin):
    """Encapsulates create parameters for ITestRunSet.create()"""

    def __init__(self, person, date_created=None):
        self.person = person
        self.date_created = date_created


class TestRunSearchParams(SearchParams, TestTargetParamsMixin):
    """Encapsulates search parameters for ITestRunSet.search()"""

    def __init__(
        self, user, person=None, sequence=None, date_created=None,
        created_before=None, created_after=None, **kwargs):

        super(TestRunSearchParams, self).__init__(**kwargs)
        self.user = user
        self.person = person
        self.sequence = sequence
        self.date_created = date_created
        self.created_before = created_before
        self.created_after = created_after


ITestRun["target"].schema = ITestRunTarget
