# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "IHasTestStats",
    "IHasTestStatDistributions",
    "IHasTestStatPeople",
    "IHasTestStatProjects",
    "ITestStat",
    "ITestStatPerson",
    "ITestStatTarget",
    "TestStatIncrementParams",
    ]

from zope.interface import Interface
from zope.schema import (
    Datetime,
    Int,
    )

from lazr.restful.declarations import (
    call_with,
    export_read_operation,
    exported,
    operation_returns_collection_of,
    )

from lpresults.registry.interfaces.date import (
    IHasDateCreated,
    IHasDateModified,
    )

from lpresults.tracker.interfaces.testtarget import (
    IHasTestTarget,
    TestTargetParamsMixin,
    )


class ITestStat(IHasTestTarget, IHasDateCreated, IHasDateModified):
    """Test statistic."""

    counter = Int(
        title=u"Counter",
        required=True,
        readonly=True,
        description=u"The counter for the number increments that happened")


class IHasTestStats(Interface):
    """An entity which has general test statistics."""

    test_run_count = exported(
        Int(
            title=u"Test run count",
            required=False,
            readonly=True))

    last_test_run_date = exported(
        Datetime(
            title=u"Last test run date",
            required=False,
            readonly=True))


class IHasTestStatDistributions(Interface):
    """An entity which has test statistics on distributions."""

    @operation_returns_collection_of(Interface)  # ITestDistribution
    @call_with(limit=5)
    @export_read_operation()
    def getLatestTestDistributions(limit=None):
        """Get the latest distributions tested.

        :param limit: Optional maximum number of items to return.
        """

    @operation_returns_collection_of(Interface)  # ITestDistribution
    @call_with(limit=5)
    @export_read_operation()
    def getTopTestDistributions(limit=None):
        """Get the top distributions which have been tested the most.

        :param limit: Optional maximum number of items to return.
        """


class IHasTestStatPeople(Interface):
    """An entity which has test statistics on people."""

    @operation_returns_collection_of(Interface)  # ITestPerson
    @call_with(limit=5)
    @export_read_operation()
    def getLatestTesters(limit=None):
        """Get the latest people testing.

        :param limit: Optional maximum number of items to return.
        """

    @operation_returns_collection_of(Interface)  # ITestPerson
    @call_with(limit=5)
    @export_read_operation()
    def getTopTesters(limit=None):
        """Get the top people which have been testing the most.

        :param limit: Optional maximum number of items to return.
        """


class IHasTestStatProjects(Interface):
    """An entity which has test statistics on projects."""

    @operation_returns_collection_of(Interface)  # ITestProject
    @call_with(limit=5)
    @export_read_operation()
    def getLatestTestProjects(limit=None):
        """Get the latest projects tested.

        :param limit: Optional maximum number of items to return.
        """

    @operation_returns_collection_of(Interface)  # ITestProject
    @call_with(limit=5)
    @export_read_operation()
    def getTopTestProjects(limit=None):
        """Get the top projects which have been tested the most.

        :param limit: Optional maximum number of items to return.
        """


class ITestStatPerson(
    IHasTestStats, IHasTestStatDistributions, IHasTestStatProjects):
    """A person by whom test statistics can be incremented.

    Examples include just an ITestPerson for now.
    """


class ITestStatTarget(IHasTestStats, IHasTestStatPeople):
    """A target for which test statistics can be incremented.

    Examples include an ITestDistroSeries and ITestProjectSeries for now.
    """
    def incrementTestStat(person, date_modified=None):
        """Increment the test statistic counter on this target.

        :param person: An `IPerson` for whom to increment the statistics.
        :param date_modified: Optional `datetime` to update the statistics.
        """


class TestStatIncrementParams(TestTargetParamsMixin):
    """Encapsulates parameters for ITestStatTarget.incrementTestStat()"""

    def __init__(self, person, date_modified=None):
        self.person = person
        self.date_modified = date_modified


ITestStat["target"].schema = ITestStatTarget
