# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Command",
    ]

import os

from optparse import make_option
from subprocess import Popen

# Set database engine to the dummy backend to avoid depending on a connection
from django.conf import settings
settings.DATABASE_ENGINE = "django.db.backends.dummy"

from django.core.management.base import NoArgsCommand
from django.template.loader import render_to_string

from zope.component import (
    getUtility,
    provideAdapter,
    )
from zope.publisher.publish import publish

from lazr.restful.interfaces import IWebServiceConfiguration

from lpresults.tracker.services.api import APIApplication
from lpresults.tracker.services.webservice import (
    WebServiceRoot,
    WebServiceRootAbsoluteURL,
    )


DEFAULT_FORMAT = "wadl-%(version)s.xml"
DEFAULT_STYLESHEET = "wadl-to-refhtml.xsl"


class ApidocRootAbsoluteURL(WebServiceRootAbsoluteURL):

    def __str__(self):
        """Return a Django template variable tag.

        The variable tag will be mapped to the URL that contains the hostname
        by the apidoc view. This is necessary so that apidoc can be generated
        at build time and the hostname can be defined afterwards in the
        configuration.
        """
        return "{{ LAUNCHPAD_RESULTS_SERVICE_ROOT }}"


class Command(NoArgsCommand):
    option_list = NoArgsCommand.option_list + (
        make_option(
            "--format",
            default=DEFAULT_FORMAT,
            dest="format",
            help="Template for creating WADL files.",
            ),
        make_option(
            "--stylesheet",
            default=DEFAULT_STYLESHEET,
            dest="stylesheet",
            help="Stylesheet for creating HTML files.",
            ),
    )
    help = "Create the WADL file and HTML documentation for the webservice."

    requires_model_validation = False

    def handle(self, **options):
        # Do not use cached file version
        WebServiceRoot.cached_wadl = None

        # Configure API application with default values
        APIApplication.configure_server()

        provideAdapter(ApidocRootAbsoluteURL)

        config = getUtility(IWebServiceConfiguration)
        format = options.get("format", DEFAULT_FORMAT)
        stylesheet = options.get("stylesheet", DEFAULT_STYLESHEET)
        directory = os.path.dirname(format)

        # First, create index.html
        content = render_to_string("apidoc-index.html", {
            "service_description": config.service_description,
            "service_versions": [{
                "version": version,
                "description": config.version_descriptions.get(version),
                } for version in config.active_versions],
            })

        index_filename = os.path.join(directory, "index.html")
        index_file = open(index_filename, "w")
        try:
            index_file.write(content)
        finally:
            index_file.close()

        # Create WADL and HTML files for each active version
        for version in config.active_versions:
            path_info = config.service_root_uri_prefix + version
            request = config.createRequest("", environ={
                "HTTP_ACCEPT": "application/vd.sun.wadl+xml",
                "HTTP_HOST": "localhost",
                "PATH_INFO": path_info,
                })
            request = publish(request)
            content = request.response.consumeBody()

            # Persist the WADL to file
            wadl_filename = format % {"version": version}
            wadl_file = open(wadl_filename, "w")
            try:
                wadl_file.write(content)
            finally:
                wadl_file.close()

            # Process the WADL into HTML
            html_filename = os.path.join(directory, version + ".html")
            stdout = open(html_filename, "w")
            Popen(["xsltproc", stylesheet, wadl_filename], stdout=stdout)
            stdout.close()
