# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Command",
    ]

import os
import re
import sys

from optparse import make_option

# Set database engine to the dummy backend to avoid depending on a connection
from django.conf import settings
settings.DATABASE_ENGINE = "django.db.backends.dummy"

from django.core.management.base import NoArgsCommand
from django.core import urlresolvers

from django.template import Context
from django.template.loader import (
    get_template,
    get_template_from_string,
    )

from lpresults.tracker.services.spriteutils import SpriteUtil


DEFAULT_MACROS = "macros.html"
DEFAULT_BUILDDIR = os.path.join(settings.MEDIA_ROOT, "build")
DEFAULT_TEMPLATE = os.path.join(settings.MEDIA_ROOT, "style.css")


class Command(NoArgsCommand):
    option_list = NoArgsCommand.option_list + (
        make_option(
            "--macros",
            default=DEFAULT_MACROS,
            dest="macros",
            help="Macros from which to load JS dependencies.",
            ),
        make_option(
            "--builddir",
            default=DEFAULT_BUILDDIR,
            dest="builddir",
            help="The directory that should contain build files.",
            ),
        make_option(
            "--template",
            default=DEFAULT_TEMPLATE,
            dest="template",
            help="Template from which to load CSS sprites",
            ),
    )
    help = "Minify and create the JS and CSS build directory."

    def get_macros_string(self, macros, context=None):
        """Return the macros as a rendered string.

        :param macros: Template or filename to the macros.
        """
        try:
            template_file = open(macros)
            try:
                string = template_file.read()
            finally:
                template_file.close()
            template = get_template_from_string(string)
        except IOError:
            template = get_template(macros)

        if context is None:
            context = Context({
                "DEBUG": True,
                "MEDIA_URL": settings.MEDIA_URL,
                })

        return template.render(context)

    def get_macros_paths(self, macros):
        """Return all the absolute paths resolved from the macros."""

        urlconf = settings.ROOT_URLCONF
        resolver = urlresolvers.RegexURLResolver(r"^/", urlconf)

        macros_string = self.get_macros_string(macros)

        path_re = re.compile(r'(href|src)="(?P<path>[^"]+)"')
        for line in macros_string.split("\n"):
            match = path_re.search(line)
            if not match:
                continue

            callback, args, kwargs = resolver.resolve(match.group("path"))
            yield os.path.join(kwargs["document_root"], kwargs["path"])

    def filter_macros_paths(
        self, macros, include=None, exclude=None, modify=None):
        """Filter all the paths from the macros with the given parameters.

        :param include: Function to include paths.
        :param exclude: Function to exclude paths.
        :param modify: Function to modify paths.
        """
        paths = self.get_macros_paths(macros)
        if include:
            paths = [p for p in paths if include(p)]

        if exclude:
            paths = [p for p in paths if not exclude(p)]

        if modify:
            paths = [modify(p) for p in paths]

        for path in paths:
            if not os.path.isfile(path):
                sys.stderr.write("Found invalid path: %s\n" % path)

            else:
                yield path

    def create_sprites(self, builddir, template):
        """Create sprites for the icons defined in the given template.

        :param builddir: Path to the build directory.
        :param template: CSS template.
        """
        css_file = os.path.join(builddir,
            os.path.basename(template.replace(".in", "")))
        combined_image_file = os.path.join(builddir, "icon-sprites")
        positioning_file = os.path.join(builddir, "icon-sprites.positioning")

        sprite_util = SpriteUtil(
            template,
            "icon-sprites",
            url_prefix_substitutions={"/@@/": "../launchpad/images/"})

        # Create image
        sprite_util.combineImages(builddir)
        sprite_util.savePNG(combined_image_file)
        sprite_util.savePositioning(positioning_file)

        # Create css
        sprite_util.loadPositioning(positioning_file)
        # The build/icon-sprites file is relative to the css file
        # in the build/ directory.
        sprite_util.saveConvertedCSS(
            css_file, settings.MEDIA_URL + "build/icon-sprites")

    def combine_css(self, builddir, macros):
        """Combine all the CSS files from the paths in the given macros.

        :param builddir: Path to the build directory.
        :param macros: Macro definitions.
        """
        from lazr.js.combo import combine_files

        paths = self.filter_macros_paths(
            macros, lambda path: path.endswith(".css"))
        names = [path.replace(settings.MEDIA_ROOT, ".") for path in paths]
        results = combine_files(names, os.path.abspath(settings.MEDIA_ROOT))

        target = os.path.join(builddir, "lpresults.css")
        f = open(target, "w")
        try:
            for result in results:
                f.write(result)
        finally:
            f.close()

    def combine_js(self, builddir, macros):
        """Combine all the JS files from the paths in the given macros.

        :param builddir: Path to the build directory.
        :param macros: Macro definitions.
        """
        from lazr.js.build import main as jsbuild

        # Remove target to ensure it gets generated again
        target = os.path.join(builddir, "lpresults.js")
        if os.path.isfile(target):
            os.remove(target)

        # Minify lazr and yui dependencies without the lazr.testing module.
        sys.argv = [
            "jsbuild",
            "-b", os.path.join(builddir, "lazr"),
            "-c", os.path.join(builddir, "yui"),
            "-x", "testing/",
            ]
        jsbuild()

        # Minify lpresults dependencies and combine JS files.
        sys.argv = [
            "jsbuild",
            "-b", builddir,
            "-n", "lpresults",
            "-s", "static/js",
            ]
        paths = self.filter_macros_paths(macros,
            lambda p: p.endswith(".js"),
            lambda p: "static/js" in p,
            lambda p: p.replace(".js", "-min.js"))
        sys.argv.extend(paths)
        jsbuild()

    def handle(self, **options):
        macros = options.get("macros", DEFAULT_MACROS)
        builddir = options.get("builddir", DEFAULT_BUILDDIR)
        template = options.get("template", DEFAULT_TEMPLATE)

        # Create build directory if it doesn't exist already.
        if not os.path.exists(builddir):
            os.mkdir(builddir)

        # Combine js first to extract files from lazr and yui
        self.combine_js(builddir, macros)

        # Then, create sprites in preparation for the css files
        self.create_sprites(builddir, template)

        # Finally, combine the css files
        self.combine_css(builddir, macros)
