# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Command",
    ]

import os
import sys

from optparse import make_option

from django.core.management.base import (
    BaseCommand,
    CommandError,
    )

from lpresults.tracker.wsgi import WSGIHandler


class Command(BaseCommand):
    """Run the Django dev server with a custom handler.

    Unfortunately there is no way to provide your own handler to the Django
    runserver command, so this a copy of Django runnserver but using a
    custom handler instead of the default one.
    """
    option_list = BaseCommand.option_list + (
        make_option(
            "--noreload",
            action="store_false",
            dest="use_reloader",
            default=True,
            help="Tells Django to NOT use the auto-reloader.",
            ),
    )
    help = "Starts a lightweight Web server for development."
    args = "[optional port number, or ipaddr:port]"

    # Validation is called explicitly each time the server is reloaded.
    requires_model_validation = False

    def handle(self, addrport="", *args, **options):
        import django
        from django.core.servers.basehttp import (
            run,
            WSGIServerException,
            )

        if args:
            raise CommandError("Usage is runserver %s" % self.args)
        if not addrport:
            addr = ""
            port = "8000"
        else:
            try:
                addr, port = addrport.split(":")
            except ValueError:
                addr, port = "", addrport
        if not addr:
            addr = "localhost"

        if not port.isdigit():
            raise CommandError("%r is not a valid port number." % port)

        use_reloader = options.get("use_reloader", True)
        shutdown_message = options.get("shutdown_message", "")
        if sys.platform == "win32":
            quit_command = "CTRL-BREAK"
        else:
            quit_command = "CONTROL-C"

        def inner_run():
            print "\nDjango version %s, using settings %r" % (
                django.get_version(),
                os.environ["DJANGO_SETTINGS_MODULE"],
                )
            print "Development server is running at http://%s:%s/" % (
                addr,
                port,
                )
            print "Quit the server with %s." % quit_command

            try:
                handler = WSGIHandler()
                run(addr, int(port), handler)
            except WSGIServerException, e:
                # Use helpful error messages instead of ugly tracebacks.
                ERRORS = {
                    13: "You don't have permission to access that port.",
                    98: "That port is already in use.",
                    99: "That IP address can't be assigned-to.",
                }
                try:
                    error_text = ERRORS[e.args[0].args[0]]
                except (AttributeError, KeyError):
                    error_text = str(e)
                sys.stderr.write(
                    self.style.ERROR("Error: %s" % error_text) + "\n")
                # Need to use _exit because sys.exit doesn't work in a thread
                os._exit(1)
            except KeyboardInterrupt:
                if shutdown_message:
                    print shutdown_message
                sys.exit(0)

        if use_reloader:
            from django.utils import autoreload
            autoreload.main(inner_run)
        else:
            inner_run()
