# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestCase",
    "TestCaseTargetMixin",
    "test_case_set",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Reference,
    Unicode,
    )

from zope.component import getUtility
from zope.interface import implements

from lazr.restful.simple import TraverseWithGet
from lazr.restful.utils import get_current_browser_request

from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import string_to_datetime
from lpresults.registry.models.search import SearchSet

from lpresults.tracker.interfaces.testcase import (
    ITestCase,
    ITestCaseSet,
    ITestCaseTarget,
    TestCaseGetParams,
    TestCaseSearchParams,
    )
from lpresults.tracker.interfaces.teststore import IHasTestStore
from lpresults.tracker.models.testresult import (
    HasTestResultsMixin,
    TestResult,
    )
from lpresults.tracker.models.testtarget import (
    TestTargetHelper,
    TestTargetMixin,
    )


class TestCase(TestTargetHelper, HasTestResultsMixin, TraverseWithGet):

    implements(ITestCase)

    __storm_table__ = "testcase"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    distro_series_id = Int(name="distro_series")
    project_series_id = Int(name="project_series")

    @property
    def __parent__(self):
        return self.target

    @property
    def __path_parts__(self):
        return ("testcases", self.name,)

    def get(self, request, name):
        """See `ITraverseWithGet`."""
        try:
            # XXX: This should probably use a sequence instead.
            date_created = string_to_datetime(name)
        except ValueError:
            return None

        return self.searchTestResults(
            request.user,
            name=self.name,
            date_created=date_created).one()

    def updateTestTargetParams(self, target_params):
        """See `TestTargetMixin`."""
        target_params.setTarget(
            distro_series=self.distro_series,
            project_series=self.project_series)

TestResult.test_case = Reference(TestResult.test_case_id, TestCase.id)


class TestCaseSet(SearchSet, TraverseWithGet):

    implements(ITestCaseSet)

    __url_path__ = "testcases"

    @property
    def __parent__(self):
        request = get_current_browser_request()
        return request.target

    def get(self, request, name):
        """See `ITraverseWithGet`."""
        return request.target.searchTestCases(name=name).one()

    def getId(self, get_params):
        """See `ITestCaseSet`."""
        store = IHasTestStore(get_params.target).test_store
        get_args = (
            get_params.distro_series.id
                if get_params.distro_series else None,
            get_params.project_series.id
                if get_params.project_series else None,
            get_params.name,
            )
        return store.execute(
            "SELECT get_test_case(?, ?, ?)", get_args).get_one()[0]

    def getAll(self):
        """See `ITestCaseSet`."""
        request = get_current_browser_request()
        return request.target.searchTestCases()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestCaseSet, self).buildSearchExpressions(
            search_params)

        if search_params.distro_series:
            expressions.append(
                search_to_expression(
                    TestCase.distro_series_id,
                    search_params.distro_series,
                    modifier=lambda x: x.id))

        if search_params.project_series:
            expressions.append(
                search_to_expression(
                    TestCase.project_series_id,
                    search_params.project_series,
                    modifier=lambda x: x.id))

        if search_params.name is not None:
            expressions.append(
                search_to_expression(
                    TestCase.name,
                    search_params.name.lower()))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return TestCase

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return IHasTestStore(search_params.target).test_store


class TestCaseTargetMixin(TestTargetMixin):

    implements(ITestCaseTarget)

    def getTestCaseId(self, *args, **kwargs):
        """See `ITestCaseTarget`."""
        get_params = TestCaseGetParams(*args, **kwargs)
        self.updateTestTargetParams(get_params)
        return getUtility(ITestCaseSet).getId(get_params)

    def searchTestCases(self, *args, **kwargs):
        """See `IHasTestCases`."""
        search_params = TestCaseSearchParams(*args, **kwargs)
        self.updateTestTargetParams(search_params)
        return getUtility(ITestCaseSet).search(search_params)


test_case_set = TestCaseSet()
