# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestDistribution",
    "test_distribution_set",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import Max

from zope.cachedescriptors.property import Lazy
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple
from lpresults.registry.interfaces.distribution import IDistribution
from lpresults.registry.interfaces.distroseries import IDistroSeriesSet
from lpresults.registry.models.distribution import (
    Distribution,
    DistributionSet,
    )
from lpresults.registry.models.distroseries import DistroSeries
from lpresults.registry.models.series import SeriesTargetMixin

from lpresults.tracker.interfaces.testdistribution import (
    ITestDistribution,
    ITestDistributionSet,
    TestDistributionSearchParams,
    )
from lpresults.tracker.models.testcase import TestCaseTargetMixin
from lpresults.tracker.models.testrank import TestRankTarget
from lpresults.tracker.models.testrun import TestRunTargetMixin
from lpresults.tracker.models.teststat import (
    HasTestStatDistributionsMixin,
    TestStat,
    TestStatTarget,
    )
from lpresults.tracker.models.teststore import HasTestStore


class TestDistribution(
    TestRankTarget, TestStatTarget, HasTestStore, SeriesTargetMixin,
    TestCaseTargetMixin, TestRunTargetMixin):
    """See `ITestDistribution`."""
    implements(ITestDistribution)
    delegates(IDistribution)

    # XXX This needs to be stored for future sharding implementation
    test_store_key = 1

    @Lazy
    def current_series(self):
        """See `IDistribution`."""
        return IDistroSeriesSet(self).get(self.context.current_series_id)

    def incrementTestStat(self, *args, **kwargs):
        """See `ITestStatTarget`."""
        counter = super(TestDistribution, self).incrementTestStat(
            *args, **kwargs)

        # Update the cache from the development focus.
        self.test_run_count = self.current_series.test_run_count
        self.last_test_run_date = self.current_series.last_test_run_date

        return counter

    def updateTestTargetParams(self, target_params):
        """See `TestTargetTargetMixin`."""
        target_params.setTarget(distro_series=self.current_series)


class TestDistributionSet(DistributionSet, HasTestStatDistributionsMixin):
    """See `ITestDistributionSet`."""
    implements(ITestDistributionSet)

    search_params_factory = TestDistributionSearchParams

    def get(self, id):
        """See `IDistributionSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestDistributionSet, self).buildSearchExpressions(
            search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(Distribution.id, search_params.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestDistributionSet, self).buildSearchResult(
            search_params)

        return result.group_by(Distribution)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestDistributionSet, self).buildSearchTables(
            search_params)

        tables.append(
            LeftJoin(DistroSeries,
                DistroSeries.distribution_id == Distribution.id))

        tables.append(
            LeftJoin(TestStat,
                TestStat.distro_series_id == DistroSeries.id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestDistribution(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(
            TestDistributionSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


test_distribution_set = TestDistributionSet()
