# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestDistroSeries",
    "TestDistroSeriesSet",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import Max

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple
from lpresults.registry.interfaces.distroseries import IDistroSeries
from lpresults.registry.models.distroseries import (
    DistroSeries,
    DistroSeriesSet,
    )
from lpresults.registry.models.distroserieshelper import DistroSeriesTwin

from lpresults.tracker.interfaces.testdistribution import ITestDistributionSet
from lpresults.tracker.interfaces.testdistroseries import (
    ITestDistroSeries,
    ITestDistroSeriesSet,
    )
from lpresults.tracker.interfaces.testseries import TestSeriesSearchParams
from lpresults.tracker.interfaces.teststore import IHasTestStore
from lpresults.tracker.models.testcase import TestCaseTargetMixin
from lpresults.tracker.models.testrank import TestRankTarget
from lpresults.tracker.models.testrun import TestRunTargetMixin
from lpresults.tracker.models.teststat import (
    TestStat,
    TestStatTarget,
    )


class TestDistroSeries(
    TestRankTarget, TestStatTarget, TestCaseTargetMixin, TestRunTargetMixin):
    """See `ITestDistroSeries`."""
    implements(ITestDistroSeries)
    delegates(IDistroSeries)

    @Lazy
    def distribution(self):
        """See `IDistroSeries`."""
        return getUtility(ITestDistributionSet).get(
            self.context.distribution_id)

    @property
    def test_store(self):
        """See `IHasTestStore`."""
        return self.distribution.test_store

    def updateTestTargetParams(self, target_params):
        """See `TestTargetTargetMixin`."""
        target_params.setTarget(distro_series=self)


class TestDistroSeriesSet(DistroSeriesSet):
    """See `ITestDistroSeriesSet`."""
    implements(ITestDistroSeriesSet)

    search_params_factory = TestSeriesSearchParams

    def new(self, name):
        """See `ISeriesSet`."""
        series = DistroSeries(self.target.context, name)
        return self.get(series.id)

    def get(self, id):
        """See `ISeriesSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(
            TestDistroSeriesSet, self).buildSearchExpressions(search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(DistroSeries.id, search_params.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestDistroSeriesSet, self).buildSearchResult(
            search_params)

        return result.group_by(DistroSeries)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestDistroSeriesSet, self).buildSearchTables(
            search_params)

        tables.append(
            LeftJoin(TestStat,
                TestStat.distro_series_id == DistroSeries.id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestDistroSeries(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(
            TestDistroSeriesSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


def distro_series_to_has_store(series):
    return TestDistroSeries(series, None, None)


def distro_series_created(series, event):
    # Flush the store to ensure correct ordering of
    # SQL statements emitted by Storm.
    series_twin = DistroSeriesTwin(series)
    test_store = IHasTestStore(series).test_store
    test_store.add(series_twin)
    test_store.flush()
