# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestPillarSet",
    "test_pillar_set",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import (
    Max,
    Or,
    )

from zope.component import getUtility
from zope.interface import implements

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple
from lpresults.registry.models.pillar import (
    PillarName,
    PillarSet,
    )

from lpresults.tracker.interfaces.testdistribution import ITestDistributionSet
from lpresults.tracker.interfaces.testpillar import (
    ITestPillarSet,
    TestPillarSearchParams,
    )
from lpresults.tracker.interfaces.testproject import ITestProjectSet
from lpresults.tracker.models.teststat import TestStat


class TestPillarSet(PillarSet):
    """See `ITestPillarSet`."""
    implements(ITestPillarSet)

    search_params_factory = TestPillarSearchParams

    entry_type = ITestPillarSet

    def createDistribution(self, *args, **kwargs):
        """See `IPillarSet`."""
        return getUtility(ITestDistributionSet).create(*args, **kwargs)

    def createProject(self, *args, **kwargs):
        """See `IPillarSet`."""
        return getUtility(ITestProjectSet).create(*args, **kwargs)

    def searchPillars(self, text):
        """See ITestPillarSet."""
        # Only search names for now
        search_params = self.search_params_factory(name=text)
        return self.search(search_params)

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestPillarSet, self).buildSearchExpressions(
            search_params)

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        from lpresults.registry.models.distribution import Distribution
        from lpresults.registry.models.project import Project

        result = super(TestPillarSet, self).buildSearchResult(search_params)

        return result.group_by(PillarName, Distribution, Project)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        from lpresults.registry.models.distroseries import DistroSeries
        from lpresults.registry.models.projectseries import ProjectSeries

        tables = super(TestPillarSet, self).buildSearchTables(search_params)

        tables.append(
            LeftJoin(DistroSeries,
                DistroSeries.distribution_id == PillarName.distribution_id))

        tables.append(
            LeftJoin(ProjectSeries,
                ProjectSeries.project_id == PillarName.project_id))

        tables.append(
            LeftJoin(TestStat, Or(
                TestStat.distro_series_id == DistroSeries.id,
                TestStat.project_series_id == ProjectSeries.id)))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        from lpresults.tracker.models.testdistribution import TestDistribution
        from lpresults.tracker.models.testproject import TestProject

        if result[1] is not None:
            return TestDistribution(result[1], *result[3:])
        else:
            return TestProject(result[2], *result[3:])

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(TestPillarSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


test_pillar_set = TestPillarSet()
