# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestProject",
    "test_project_set",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import Max

from zope.cachedescriptors.property import Lazy
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple
from lpresults.registry.interfaces.project import IProject
from lpresults.registry.interfaces.projectseries import IProjectSeriesSet
from lpresults.registry.models.project import (
    Project,
    ProjectSet,
    )
from lpresults.registry.models.projectseries import ProjectSeries
from lpresults.registry.models.series import SeriesTargetMixin

from lpresults.tracker.interfaces.testproject import (
    ITestProject,
    ITestProjectSet,
    TestProjectSearchParams,
    )
from lpresults.tracker.models.testcase import TestCaseTargetMixin
from lpresults.tracker.models.testrank import TestRankTarget
from lpresults.tracker.models.testrun import TestRunTargetMixin
from lpresults.tracker.models.teststat import (
    HasTestStatProjectsMixin,
    TestStat,
    TestStatTarget,
    )
from lpresults.tracker.models.teststore import HasTestStore


class TestProject(
    TestRankTarget, TestStatTarget, HasTestStore, SeriesTargetMixin,
    TestCaseTargetMixin, TestRunTargetMixin):
    """See `ITestProject`."""
    implements(ITestProject)
    delegates(IProject)

    # XXX This needs to be stored for future sharding implementation
    test_store_key = 1

    @Lazy
    def development_focus(self):
        """See `IProject`."""
        return IProjectSeriesSet(self).get(
            self.context.development_focus_id)

    def incrementTestStat(self, *args, **kwargs):
        """See `ITestStatTarget`."""
        counter = super(TestProject, self).incrementTestStat(*args, **kwargs)

        # Update the cache from the development focus.
        self.test_run_count = self.development_focus.test_run_count
        self.last_test_run_date = self.development_focus.last_test_run_date

        return counter

    def updateTestTargetParams(self, target_params):
        """See `TestTargetTargetMixin`."""
        target_params.setTarget(project_series=self.development_focus)


class TestProjectSet(ProjectSet, HasTestStatProjectsMixin):
    """See `ITestProjectSet`."""
    implements(ITestProjectSet)

    search_params_factory = TestProjectSearchParams

    def get(self, id):
        """See `IProjectSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestProjectSet, self).buildSearchExpressions(
            search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(Project.id, search_params.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestProjectSet, self).buildSearchResult(search_params)

        return result.group_by(Project)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestProjectSet, self).buildSearchTables(search_params)

        tables.append(
            LeftJoin(ProjectSeries,
                ProjectSeries.project_id == Project.id))

        tables.append(
            LeftJoin(TestStat,
                TestStat.project_series_id == ProjectSeries.id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestProject(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(TestProjectSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


test_project_set = TestProjectSet()
