# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestProjectSeries",
    "TestProjectSeriesSet",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import Max

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple
from lpresults.registry.interfaces.projectseries import IProjectSeries
from lpresults.registry.models.projectseries import (
    ProjectSeries,
    ProjectSeriesSet,
    )
from lpresults.registry.models.projectserieshelper import ProjectSeriesTwin

from lpresults.tracker.interfaces.testproject import ITestProjectSet
from lpresults.tracker.interfaces.testprojectseries import (
    ITestProjectSeries,
    ITestProjectSeriesSet,
    )
from lpresults.tracker.interfaces.testseries import TestSeriesSearchParams
from lpresults.tracker.interfaces.teststore import IHasTestStore
from lpresults.tracker.models.testcase import TestCaseTargetMixin
from lpresults.tracker.models.testrank import TestRankTarget
from lpresults.tracker.models.testrun import TestRunTargetMixin
from lpresults.tracker.models.teststat import (
    TestStat,
    TestStatTarget,
    )


class TestProjectSeries(
    TestRankTarget, TestStatTarget, TestCaseTargetMixin, TestRunTargetMixin):
    """See `ITestProjectSeries`."""
    implements(ITestProjectSeries)
    delegates(IProjectSeries)

    @Lazy
    def project(self):
        """See `IProjectSeries`."""
        return getUtility(ITestProjectSet).get(self.context.project_id)

    @property
    def test_store(self):
        """See `IHasTestStore`."""
        return self.project.test_store

    def updateTestTargetParams(self, target_params):
        """See `TestTargetTargetMixin`."""
        target_params.setTarget(project_series=self)


class TestProjectSeriesSet(ProjectSeriesSet):
    """See `ITestProjectSeriesSet`."""
    implements(ITestProjectSeriesSet)

    search_params_factory = TestSeriesSearchParams

    def new(self, name):
        """See `ISeriesSet`."""
        series = ProjectSeries(self.target.context, name)
        return self.get(series.id)

    def get(self, id):
        """See `ISeriesSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(
            TestProjectSeriesSet, self).buildSearchExpressions(search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(ProjectSeries.id, search_params.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestProjectSeriesSet, self).buildSearchResult(
            search_params)

        return result.group_by(ProjectSeries)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestProjectSeriesSet, self).buildSearchTables(
            search_params)

        tables.append(
            LeftJoin(TestStat,
                TestStat.project_series_id == ProjectSeries.id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestProjectSeries(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(
            TestProjectSeriesSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


def project_series_to_has_store(series):
    return TestProjectSeries(series, None, None)


def project_series_created(series, event):
    # Flush the store to ensure correct ordering of
    # SQL statements emitted by Storm.
    series_twin = ProjectSeriesTwin(series)
    test_store = IHasTestStore(series).test_store
    test_store.add(series_twin)
    test_store.flush()
