# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestResult",
    "HasTestResultsMixin",
    "TestResultTargetMixin",
    "test_result_set",
    ]

import re

from datetime import datetime

from storm.locals import (
    And,
    Int,
    Join,
    Like,
    Unicode,
    )

from zope.component import getUtility
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.constants import (
    INFINITY,
    UTC_NOW,
    )
from lpresults.database.datetime import UTCDateTime
from lpresults.database.enum import DBEnum
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import datetime_to_string
from lpresults.registry.models.search import SearchSet

from lpresults.tracker.interfaces.testcase import (
    ITestCase,
    ITestCaseSet,
    )
from lpresults.tracker.interfaces.testresult import (
    IHasTestResults,
    ITestResult,
    ITestResultSet,
    ITestResultTarget,
    TestResultAddParams,
    TestResultRemoveParams,
    TestResultSearchParams,
    TestResultStatus,
    )
from lpresults.tracker.interfaces.teststore import IHasTestStore
from lpresults.tracker.models.testtarget import TestTargetMixin


class TestResult:

    implements(ITestResult)
    delegates(ITestCase, context="test_case")

    __storm_table__ = "testresult"
    __storm_primary__ = ("test_case_id", "date_created",)

    test_case_id = Int(name="test_case", allow_none=False)
    status = DBEnum(
        allow_none=False,
        enum=TestResultStatus,
        default=TestResultStatus.UNINITIATED)
    output = Unicode(allow_none=False, default=u"")
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    date_deleted = UTCDateTime(allow_none=False, default=INFINITY)

    # test_case reference defined after TestCase

    @property
    def __parent__(self):
        return self.test_case

    @property
    def __url_path__(self):
        return datetime_to_string(self.date_created)


class TestResultSet(SearchSet):

    implements(ITestResultSet)

    def add(self, add_params):
        """See `ITestResultSet`."""
        test_case_id = getUtility(ITestCaseSet).getId(add_params)

        # Sanitize parameters
        if isinstance(add_params.status, basestring):
            add_params.status = TestResultStatus.get_enum(add_params.status)
        elif add_params.status is None:
            add_params.status = TestResultStatus.UNINITIATED
        if add_params.output is None:
            add_params.output = u""
        if add_params.date_created is None:
            add_params.date_created = datetime.utcnow()

        # Add the test result
        store = self.getSearchStore(add_params)
        store.execute(
            "SELECT add_test_result(?, ?, ?, ?)", (
                test_case_id,
                add_params.status.value,
                add_params.output,
                add_params.date_created,
                ),
            noresult=True)

    def remove(self, remove_params):
        """See `ITestResultSet`."""
        search_params = TestResultSearchParams(
            person=remove_params.person,
            name=remove_params.name,
            date_created=remove_params.date_created)
        search_params.setTarget(
            distro_series=remove_params.distro_series,
            project_series=remove_params.project_series)
        test_result = self.search(search_params).one()

        # Remove the test result from the target store if it exists
        if test_result:
            test_store = IHasTestStore(remove_params.target).test_store
            test_store.remove(test_result)
            test_store.commit()

        return test_result

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        from lpresults.tracker.models.testcase import TestCase

        tables = super(TestResultSet, self).buildSearchTables(
            search_params)

        tables.append(
            Join(
                TestCase,
                TestCase.id == TestResult.test_case_id))

        return tables

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        from lpresults.tracker.models.testcase import TestCase

        expressions = super(TestResultSet, self).buildSearchExpressions(
            search_params)

        if search_params.distro_series:
            expressions.append(
                search_to_expression(
                    TestCase.distro_series_id,
                    search_params.distro_series,
                    modifier=lambda x: x.id))

        if search_params.project_series:
            expressions.append(
                search_to_expression(
                    TestCase.project_series_id,
                    search_params.project_series,
                    modifier=lambda x: x.id))

        if search_params.name is not None:
            expressions.append(
                search_to_expression(TestCase.name, search_params.name))

        if search_params.status is not None:
            expressions.append(
                search_to_expression(TestResult.status, search_params.status))

        if search_params.date_created:
            expressions.append(
                And(
                    TestResult.date_created <= search_params.date_created,
                    TestResult.date_deleted > search_params.date_created))

        return expressions

    def buildSearchText(self, search_text):
        """See SearchSet."""
        from lpresults.tracker.models.testcase import TestCase

        expressions = super(TestResultSet, self).buildSearchText(search_text)

        for part in re.split(r"\s+", search_text):
            expressions.append(Like(TestCase.name, "%s%%" % part))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        from lpresults.tracker.models.testcase import TestCase

        return (TestResult, TestCase,)

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return IHasTestStore(search_params.target).test_store


class HasTestResultsMixin(TestTargetMixin):

    implements(IHasTestResults)

    def searchTestResults(self, *args, **kwargs):
        """See `IHasTestResults`."""
        params = TestResultSearchParams(*args, **kwargs)
        self.updateTestTargetParams(params)
        return getUtility(ITestResultSet).search(params)


class TestResultTargetMixin(HasTestResultsMixin):

    implements(ITestResultTarget)

    def addTestResult(self, *args, **kwargs):
        """See `ITestResultTarget`."""
        params = TestResultAddParams(*args, **kwargs)
        self.updateTestTargetParams(params)
        getUtility(ITestResultSet).add(params)

        # Increment the count attribute
        attr = "%s_count" % params.status.name.lower()
        count = getattr(self, attr)
        setattr(self, attr, count + 1)

    def removeTestResult(self, *args, **kwargs):
        """See `ITestResultTarget`."""
        params = TestResultRemoveParams(*args, **kwargs)
        self.updateTestTargetParams(params)
        test_result = getUtility(ITestResultSet).remove(params)

        # Decrement the count attribute
        attr = "%s_count" % test_result.status.name.lower()
        count = getattr(self, attr)
        setattr(self, attr, count - 1)


test_result_set = TestResultSet()
