# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestRun",
    "TestRunTargetMixin",
    "test_run_set",
    ]

from storm.exceptions import IntegrityError
from storm.locals import (
    AutoReload,
    Int,
    )

from zope.component import getUtility
from zope.interface import implements

from lazr.restful.simple import TraverseWithGet
from lazr.restful.utils import get_current_browser_request

from lpresults.database.constants import UTC_NOW
from lpresults.database.datetime import UTCDateTime
from lpresults.database.search import (
    search_to_expression,
    SearchLessThan,
    SearchGreaterThan,
    )

from lpresults.registry.models.search import SearchSet

from lpresults.tracker.interfaces.testresult import (
    ITestResultSet,
    TestResultSearchParams,
    )
from lpresults.tracker.interfaces.testrun import (
    ITestRun,
    ITestRunSet,
    ITestRunTarget,
    TestRunCreateParams,
    TestRunDateCreatedExists,
    TestRunSearchParams,
    )
from lpresults.tracker.interfaces.teststore import IHasTestStore
from lpresults.tracker.models.attachment import AttachmentTargetMixin
from lpresults.tracker.models.testpersonhelper import TestPersonHelper
from lpresults.tracker.models.testresult import TestResultTargetMixin
from lpresults.tracker.models.testsystemunit import TestSystemUnitTargetMixin
from lpresults.tracker.models.testtarget import (
    TestTargetHelper,
    TestTargetMixin,
    )


class TestRun(
    TestPersonHelper, TestTargetHelper,
    TestResultTargetMixin, TestSystemUnitTargetMixin,
    AttachmentTargetMixin):

    implements(ITestRun)

    __storm_table__ = "testrun"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    sequence = Int(allow_none=False)
    person_id = Int(name="person")
    system_unit_id = Int(name="system_unit")
    distro_series_id = Int(name="distro_series")
    project_series_id = Int(name="project_series")
    pass_count = Int(allow_none=False, default=0)
    fail_count = Int(allow_none=False, default=0)
    uninitiated_count = Int(allow_none=False, default=0)
    unresolved_count = Int(allow_none=False, default=0)
    unsupported_count = Int(allow_none=False, default=0)
    untested_count = Int(allow_none=False, default=0)
    xfail_count = Int(allow_none=False, default=0)

    def __init__(
        self, person=None, date_created=None, distro_series=None,
        project_series=None):

        TestPersonHelper.__init__(self, person)
        TestTargetHelper.__init__(
            self, distro_series=distro_series, project_series=project_series)

        if date_created is not None:
            self.date_created = date_created

    @property
    def __parent__(self):
        return self.target

    @property
    def __path_parts__(self):
        return ("+testruns", self.name,)

    @property
    def name(self):
        return str(self.sequence)

    def setSystemUnit(self, *args, **kwargs):
        """See TestSystemUnitTargetMixin."""
        super(TestRun, self).setSystemUnit(*args, **kwargs)
        self.target.setTestRank(self.person, self.system_unit, 0)

    def updateTestTargetParams(self, target_params):
        """See TestResultTargetMixin."""
        target_params.setTarget(
            distro_series=self.distro_series,
            project_series=self.project_series)
        target_params.date_created = self.date_created


class TestRunSet(SearchSet, TraverseWithGet):

    implements(ITestRunSet)

    __url_path__ = "+testruns"

    @property
    def __parent__(self):
        request = get_current_browser_request()
        return request.target

    def get(self, request, name):
        """See `ITraverseWithGet`."""
        sequence = int(name)
        return request.target.searchTestRuns(
            request.user,
            sequence=sequence).one()

    def create(self, create_params):
        """See `ITestRunSet`."""
        # This makes sure the target is defined
        store = IHasTestStore(create_params.target).test_store
        test_run = TestRun(
            create_params.person,
            create_params.date_created,
            distro_series=create_params.distro_series,
            project_series=create_params.project_series)
        store.add(test_run)

        # Ensure that the test run is flushed to determine
        # if another one already exists with the same date
        try:
            store.flush()
        except IntegrityError:
            raise TestRunDateCreatedExists()

        test_run.target.incrementTestStat(
            create_params.person, create_params.date_created)
        return test_run

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestRunSet, self).buildSearchExpressions(
            search_params)

        if search_params.distro_series:
            expressions.append(
                search_to_expression(
                    TestRun.distro_series_id,
                    search_params.distro_series,
                    modifier=lambda x: x.id))

        if search_params.project_series:
            expressions.append(
                search_to_expression(
                    TestRun.project_series_id,
                    search_params.project_series,
                    modifier=lambda x: x.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestRun.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.sequence:
            expressions.append(
                search_to_expression(
                    TestRun.sequence,
                    search_params.sequence))

        if search_params.date_created:
            expressions.append(
                search_to_expression(
                    TestRun.date_created,
                    search_params.date_created))

        if search_params.created_before:
            expressions.append(
                search_to_expression(
                    TestRun.date_created,
                    SearchLessThan(search_params.created_before)))

        if search_params.created_after:
            expressions.append(
                search_to_expression(
                    TestRun.date_created,
                    SearchGreaterThan(search_params.created_after)))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return TestRun

    def getSearchStore(self, search_params):
        """See SearchSet."""
        return IHasTestStore(search_params.target).test_store


class TestRunTargetMixin(TestTargetMixin):

    implements(ITestRunTarget)

    def createTestRun(self, *args, **kwargs):
        """See `ITestRunTarget`."""
        create_params = TestRunCreateParams(*args, **kwargs)
        self.updateTestTargetParams(create_params)
        return getUtility(ITestRunSet).create(create_params)

    def getLatestTestRuns(self, person, limit=None):
        """See `IHasTestRuns`."""
        search_params = TestRunSearchParams(person, order_by="-date_created")
        self.updateTestTargetParams(search_params)

        test_runs = getUtility(ITestRunSet).search(search_params)
        return test_runs.config(limit=limit)

    def searchTestResults(self, *args, **kwargs):
        """See `IHasTestResults`."""
        search_params = TestResultSearchParams(*args, **kwargs)
        self.updateTestTargetParams(search_params)
        return getUtility(ITestResultSet).search(search_params)

    def searchTestRuns(self, *args, **kwargs):
        """See `IHasTestRuns`."""
        search_params = TestRunSearchParams(*args, **kwargs)
        self.updateTestTargetParams(search_params)
        return getUtility(ITestRunSet).search(search_params)


test_run_set = TestRunSet()
