# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import pytz

from datetime import (
    datetime,
    timedelta,
    )

from unittest import defaultTestLoader

from lpresults.tracker.interfaces.testresult import (
    ITestResultTarget,
    TestResultAddParams,
    TestResultRemoveParams,
    TestResultSearchParams,
    TestResultStatus,
    )
from lpresults.tracker.models.testresult import test_result_set
from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import FunctionalDatabaseLayer


UTC = pytz.timezone("UTC")


class TestTestResultSet(TestCaseWithFactory):

    layer = FunctionalDatabaseLayer

    def setUp(self):
        super(TestTestResultSet, self).setUp()

        self.person = self.factory.makePerson()
        self.project = self.factory.makeTestProject()

    def addTestResult(self, name, **kwargs):
        """Add a test result using an IProject as the target."""
        add_params = TestResultAddParams(self.person, name, **kwargs)
        self.project.updateTestTargetParams(add_params)
        return test_result_set.add(add_params)

    def removeTestResult(self, name, **kwargs):
        """Remove a test result using an IProject as the target."""
        remove_params = TestResultRemoveParams(self.person, name, **kwargs)
        self.project.updateTestTargetParams(remove_params)
        test_result_set.remove(remove_params)

    def searchTestResults(self, **kwargs):
        """Search for test results using an IProject as the target."""
        search_params = TestResultSearchParams(person=self.person, **kwargs)
        self.project.updateTestTargetParams(search_params)
        return test_result_set.search(search_params)

    def test_add(self):
        name = self.factory.getUniqueString(u"test-name")

        # There should not be any test results by the same name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 0)

        # Add a test result with the given name
        self.addTestResult(name)

        # There should now be one test result with the given name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 1)

    def test_addTwice(self):
        name = self.factory.getUniqueString(u"test-name")

        # Add a test results with the give name twice
        self.addTestResult(name)
        self.addTestResult(name)

        # There should be only one test result with the given name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 1)

    def test_addWithStatus(self):
        name = self.factory.getUniqueString(u"test-name")
        status = TestResultStatus.UNRESOLVED
        self.addTestResult(name, status=status)

        # There should be one test result with the given name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 1)

        # The test result status should be the same as the given status
        test_result = test_results[0]
        self.assertEquals(test_result.status, status)

    def test_changeStatus(self):
        name = self.factory.getUniqueString(u"test-name")
        status = TestResultStatus.PASS
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        self.addTestResult(name, status=status, date_created=date_created)

        # The test result status should be the same as the given status
        test_result = self.searchTestResults(
            name=name,
            date_created=date_created)[0]
        self.assertEquals(test_result.status, status)

        # Update the test result
        status = TestResultStatus.FAIL
        date_created += timedelta(days=1)
        self.addTestResult(name, status=status, date_created=date_created)

        # The test result status should be changed to the new status
        test_result = self.searchTestResults(
            name=name,
            date_created=date_created)[0]
        self.assertEquals(test_result.status, status)

    def test_remove(self):
        name = self.factory.getUniqueString(u"test-name")
        self.addTestResult(name)

        # Remove the test result with the given name
        self.removeTestResult(name)

        # There should no longer be any test results with the given name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 0)

    def test_removeTwice(self):
        name = self.factory.getUniqueString(u"test-name")
        self.addTestResult(name)

        # Remove the test result with the given name twice
        self.removeTestResult(name)

        # Nothing should happen
        self.removeTestResult(name)

    def test_search(self):
        # There should no test results
        test_results = self.searchTestResults()
        self.assertEquals(test_results.count(), 0)

        # Add an arbitrary test result
        name = self.factory.getUniqueString(u"test-name")
        self.addTestResult(name)

        # There should one test result
        test_results = self.searchTestResults()
        self.assertEquals(test_results.count(), 1)

    def test_searchByName(self):
        name = self.factory.getUniqueString(u"test-name")
        self.addTestResult(name)

        # There should be one test result with the given name
        test_results = self.searchTestResults(name=name)
        self.assertEquals(test_results.count(), 1)

        # There should be no test results with another name
        other_name = self.factory.getUniqueString(u"other-name")
        test_results = self.searchTestResults(name=other_name)
        self.assertEquals(test_results.count(), 0)

    def test_searchByDateCreated(self):
        name = self.factory.getUniqueString(u"test-name")
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        self.addTestResult(name, date_created=date_created)

        # There should be one test result with the given date created
        test_results = self.searchTestResults(date_created=date_created)
        self.assertEquals(test_results.count(), 1)

        # There should also be the same test result for dates
        # after the date created
        date_created_after = date_created + timedelta(days=1)
        test_results = self.searchTestResults(date_created=date_created_after)
        self.assertEquals(test_results.count(), 1)

        # But there should be not test results for dates
        # before the date created
        date_created_before = date_created - timedelta(days=1)
        test_results = self.searchTestResults(
            date_created=date_created_before)
        self.assertEquals(test_results.count(), 0)


class TestTestResultTarget(TestCaseWithFactory):

    layer = FunctionalDatabaseLayer

    def setUp(self):
        super(TestTestResultTarget, self).setUp()

        self.person = self.factory.makePerson()
        self.project = self.factory.makeTestProject()

    def createTestResultTarget(self, **kwargs):
        """Make an arbitrary test run target with no runs on it."""
        test_run = self.project.createTestRun(self.person, **kwargs)
        return ITestResultTarget(test_run)

    def test_addTestResult(self):
        target = self.createTestResultTarget()
        test_results = target.searchTestResults(self.person)
        self.assertEquals(test_results.count(), 0)

        name = self.factory.getUniqueString(u"test-name")
        target.addTestResult(self.person, name)
        test_results = target.searchTestResults(self.person)
        self.assertEquals(test_results.count(), 1)

        test_result = test_results.one()
        self.assertNotEquals(test_result, None)
        self.assertEquals(test_result.name, name)

    def test_removeTestResult(self):
        target = self.createTestResultTarget()
        name = self.factory.getUniqueString(u"test-name")
        target.addTestResult(self.person, name)
        test_results = target.searchTestResults(self.person)
        self.assertEquals(test_results.count(), 1)

        target.removeTestResult(self.person, name)
        test_results = target.searchTestResults(self.person)
        self.assertEquals(test_results.count(), 0)

    def test_searchTestResults(self):
        target = self.createTestResultTarget()
        name = self.factory.getUniqueString(u"test-name")
        target.addTestResult(self.person, name)
        test_results = target.searchTestResults(self.person)
        self.assertEquals(test_results.count(), 1)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
