# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import pytz

from datetime import (
    datetime,
    timedelta,
    )

from unittest import defaultTestLoader

from lpresults.tracker.interfaces.testrun import (
    ITestRunTarget,
    TestRunCreateParams,
    TestRunDateCreatedExists,
    TestRunSearchParams,
    )
from lpresults.tracker.models.testrun import test_run_set
from lpresults.tracker.models.tests.test_attachment import (
    TestAttachmentTargetMixin,
    )
from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import FunctionalDatabaseLayer


UTC = pytz.timezone("UTC")


class TestTestRunSet(TestCaseWithFactory):

    layer = FunctionalDatabaseLayer

    def setUp(self):
        super(TestTestRunSet, self).setUp()
        self.person = self.factory.makePerson()
        self.project = self.factory.makeTestProject()

    def createTestRun(self, **kwargs):
        create_params = TestRunCreateParams(self.person, **kwargs)
        self.project.updateTestTargetParams(create_params)
        return test_run_set.create(create_params)

    def searchTestRuns(self, **kwargs):
        search_params = TestRunSearchParams(self.person, **kwargs)
        self.project.updateTestTargetParams(search_params)
        return test_run_set.search(search_params)

    def test_create(self):
        test_run = self.createTestRun()
        self.assertNotEquals(test_run, None)
        self.assertEquals(test_run.person, self.person)
        self.assertEquals(
            test_run.project_series, self.project.development_focus)

    def test_createWithDateCreated(self):
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        test_run = self.createTestRun(date_created=date_created)
        self.assertEquals(test_run.date_created, date_created)

    def test_createWithSameDateCreated(self):
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        self.createTestRun(date_created=date_created)
        self.assertRaises(
            TestRunDateCreatedExists,
            self.createTestRun,
            date_created=date_created)

    def test_search(self):
        # There should no test runs
        test_runs = self.searchTestRuns()
        self.assertEquals(test_runs.count(), 0)

        # Create an arbitrary test run
        self.createTestRun()

        # There should one test run
        test_runs = self.searchTestRuns()
        self.assertEquals(test_runs.count(), 1)

    def test_searchCreatedBefore(self):
        # Create a test run on a given date
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        self.createTestRun(date_created=date_created)

        # There should no test runs before the given date
        test_runs = self.searchTestRuns(created_before=date_created)
        self.assertEquals(test_runs.count(), 0)

        # There should one test run before the day after
        date_after = date_created + timedelta(days=1)
        test_runs = self.searchTestRuns(created_before=date_after)
        self.assertEquals(test_runs.count(), 1)

    def test_searchCreatedAfter(self):
        # Create a test run on a given date
        date_created = datetime(2010, 1, 1, tzinfo=UTC)
        self.createTestRun(date_created=date_created)

        # There should no test runs after the given date
        test_runs = self.searchTestRuns(created_after=date_created)
        self.assertEquals(test_runs.count(), 0)

        # There should one test run after the day before
        date_before = date_created - timedelta(days=1)
        test_runs = self.searchTestRuns(created_after=date_before)
        self.assertEquals(test_runs.count(), 1)


class TestTestRunTarget(TestCaseWithFactory):

    layer = FunctionalDatabaseLayer

    def createTestRunTarget(self):
        """Make an arbitrary test run target with no runs on it."""
        project = self.factory.makeTestProject()
        return ITestRunTarget(project)

    def test_createTestRun(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()
        test_run = target.createTestRun(person)
        self.assertNotEquals(test_run, None)
        self.assertEquals(test_run.person, person)
        self.assertEquals(test_run.target, target.development_focus)

    def test_getLatestTestRuns(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()
        date_created = datetime(2010, 1, 1, tzinfo=UTC)

        # Create a first test run
        test_run_first = target.createTestRun(person,
            date_created=date_created)
        test_runs = target.getLatestTestRuns(person)
        self.assertEquals(test_runs.count(), 1)

        # Create a second test run
        date_created += timedelta(days=1)
        test_run_last = target.createTestRun(person,
            date_created=date_created)
        test_runs = target.getLatestTestRuns(person)
        self.assertEquals(test_runs.count(), 2)

        # The last test run should appear first
        self.assertEquals(test_runs[0], test_run_last)
        self.assertEquals(test_runs[1], test_run_first)

    def test_searchTestRuns(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()
        test_runs = target.searchTestRuns(person)
        self.assertEquals(test_runs.count(), 0)

        # Create an arbitrary test run
        target.createTestRun(person)

        # There should one test run
        test_runs = target.searchTestRuns(person)
        self.assertEquals(test_runs.count(), 1)

    def test_searchTestResults(self):
        target = self.createTestRunTarget()

        # There should no test results
        test_results = target.searchTestResults()
        self.assertEquals(test_results.count(), 0)

    def test_searchTestResultsOneRun(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()

        # Create an arbitrary test run
        target.createTestRun(person)

        # There should still be no test results
        test_results = target.searchTestResults()
        self.assertEquals(test_results.count(), 0)

    def test_searchTestResultsOneRunOneResult(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()

        # Create an arbitrary test run and add a test result
        test_run = target.createTestRun(person)
        name = self.factory.getUniqueString(u"test-name")
        test_run.addTestResult(person, name)

        # There should now be one test result
        test_results = target.searchTestResults()
        self.assertEquals(test_results.count(), 1)

    def test_searchTestResultsOneRunTwoResult(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()

        # Create an arbitrary test run and add two test results
        test_run = target.createTestRun(person)
        first_name = self.factory.getUniqueString(u"first-name")
        test_run.addTestResult(person, first_name)
        second_name = self.factory.getUniqueString(u"second-name")
        test_run.addTestResult(person, second_name)

        # There should now be two test results
        test_results = target.searchTestResults()
        self.assertEquals(test_results.count(), 2)

    def test_searchTestResultsTwoRunsOneResultEach(self):
        person = self.factory.makePerson()
        target = self.createTestRunTarget()
        date_created = datetime(2010, 1, 1, tzinfo=UTC)

        # Create one test run and add a test result
        first_test_run = target.createTestRun(
            person, date_created=date_created)
        first_name = self.factory.getUniqueString(u"first-name")
        first_test_run.addTestResult(person, first_name)

        # Create another test run and add a test result
        date_created += timedelta(days=1)
        second_test_run = target.createTestRun(
            person, date_created=date_created)
        second_name = self.factory.getUniqueString(u"second-name")
        second_test_run.addTestResult(person, second_name)

        # There should now be two test results
        test_results = target.searchTestResults()
        self.assertEquals(test_results.count(), 2)


class TestAttachmentTarget(TestCaseWithFactory, TestAttachmentTargetMixin):

    def setUp(self):
        super(TestAttachmentTarget, self).setUp()

        person = self.factory.makePerson()
        project = self.factory.makeTestProject()
        self.test_run = project.createTestRun(person)

    def getAttachmentTarget(self):
        return self.test_run


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
