# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestHasTestStatDistributionsMixin",
    "TestHasTestStatPeopleMixin",
    "TestHasTestStatProjectsMixin",
    "TestTestStatPersonMixin",
    "TestTestStatTargetMixin",
    ]

import pytz

from datetime import (
    datetime,
    timedelta,
    )

from unittest import defaultTestLoader

from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import FunctionalDatabaseLayer


UTC = pytz.timezone("UTC")


class TestTestStatSet(TestCaseWithFactory):

    layer = FunctionalDatabaseLayer

    def test_increment(self):
        test_person = self.factory.makeTestPerson()
        self.assertEquals(test_person.test_run_count, None)

        test_project = self.factory.makeTestProject()
        self.assertEquals(test_project.test_run_count, None)

        count = test_project.incrementTestStat(test_person)
        self.assertEquals(count, 1)
        self.assertEquals(test_person.test_run_count, count)
        self.assertEquals(test_project.test_run_count, count)


class TestHasTestStatPeopleMixin:

    layer = FunctionalDatabaseLayer

    def getContainer(self):
        raise NotImplementedError()

    def getTestTarget(self):
        raise NotImplementedError()

    def test_getLatestTesters(self):
        container = self.getContainer()
        target = self.getTestTarget()

        # Create older target
        tester1 = self.factory.makePerson()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        target.incrementTestStat(tester1, date_modified=date)

        testers = container.getLatestTesters()
        self.assertNotEquals(testers, None)
        self.assertEquals(testers.count(), 1)

        # Create newer target
        tester2 = self.factory.makePerson()
        date += timedelta(days=1)
        target.incrementTestStat(tester2, date_modified=date)

        testers = container.getLatestTesters()
        self.assertNotEquals(testers, None)
        self.assertEquals(testers.count(), 2)
        self.assertEquals(testers[0], tester2)
        self.assertEquals(testers[1], tester1)

    def test_getTopTesters(self):
        container = self.getContainer()
        target = self.getTestTarget()

        # Create lower tester
        tester1 = self.factory.makePerson()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        target.incrementTestStat(tester1, date_modified=date)

        testers = container.getTopTesters()
        self.assertNotEquals(testers, None)
        self.assertEquals(testers.count(), 1)

        # Create upper tester
        tester2 = self.factory.makePerson()
        date += timedelta(days=1)
        target.incrementTestStat(tester2, date_modified=date)
        date += timedelta(days=2)
        target.incrementTestStat(tester2, date_modified=date)

        testers = container.getTopTesters()
        self.assertNotEquals(testers, None)
        self.assertEquals(testers.count(), 2)
        self.assertEquals(testers[0], tester2)
        self.assertEquals(testers[1], tester1)


class TestHasTestStatDistributionsMixin:

    layer = FunctionalDatabaseLayer

    def getContainer(self):
        raise NotImplementedError()

    def getTestPerson(self):
        return self.factory.makeTestPerson()

    def test_getLatestTestDistributions(self):
        container = self.getContainer()
        person = self.getTestPerson()

        # Create older distribution
        distribution1 = self.factory.makeTestDistribution()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        distribution1.incrementTestStat(person, date_modified=date)

        distributions = container.getLatestTestDistributions()
        self.assertNotEquals(distributions, None)
        self.assertEquals(distributions.count(), 1)

        # Create newer distribution
        distribution2 = self.factory.makeTestDistribution()
        date += timedelta(days=1)
        distribution2.incrementTestStat(person, date_modified=date)

        distributions = container.getLatestTestDistributions()
        self.assertNotEquals(distributions, None)
        self.assertEquals(distributions.count(), 2)
        self.assertEquals(distributions[0], distribution2)
        self.assertEquals(distributions[1], distribution1)

    def test_getTopTestDistributions(self):
        container = self.getContainer()
        person = self.getTestPerson()

        # Create lower distribution
        distribution1 = self.factory.makeTestDistribution()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        distribution1.incrementTestStat(person, date_modified=date)

        distributions = container.getTopTestDistributions()
        self.assertNotEquals(distributions, None)
        self.assertEquals(distributions.count(), 1)

        # Create upper distribution
        distribution2 = self.factory.makeTestDistribution()
        date += timedelta(days=1)
        distribution2.incrementTestStat(person, date_modified=date)
        date += timedelta(days=2)
        distribution2.incrementTestStat(person, date_modified=date)

        distributions = container.getTopTestDistributions()
        self.assertNotEquals(distributions, None)
        self.assertEquals(distributions.count(), 2)
        self.assertEquals(distributions[0], distribution2)
        self.assertEquals(distributions[1], distribution1)


class TestHasTestStatProjectsMixin:

    layer = FunctionalDatabaseLayer

    def getContainer(self):
        raise NotImplementedError()

    def getTestPerson(self):
        return self.factory.makeTestPerson()

    def test_getLatestTestProjects(self):
        container = self.getContainer()
        person = self.getTestPerson()

        # Create older project
        project1 = self.factory.makeTestProject()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        project1.incrementTestStat(person, date_modified=date)

        projects = container.getLatestTestProjects()
        self.assertNotEquals(projects, None)
        self.assertEquals(projects.count(), 1)

        # Create newer project
        project2 = self.factory.makeTestProject()
        date += timedelta(days=1)
        project2.incrementTestStat(person, date_modified=date)

        projects = container.getLatestTestProjects()
        self.assertNotEquals(projects, None)
        self.assertEquals(projects.count(), 2)
        self.assertEquals(projects[0], project2)
        self.assertEquals(projects[1], project1)

    def test_getTopTestProjects(self):
        container = self.getContainer()
        person = self.getTestPerson()

        # Create lower project
        project1 = self.factory.makeTestProject()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        project1.incrementTestStat(person, date_modified=date)

        projects = container.getTopTestProjects()
        self.assertNotEquals(projects, None)
        self.assertEquals(projects.count(), 1)

        # Create upper project
        project2 = self.factory.makeTestProject()
        date += timedelta(days=1)
        project2.incrementTestStat(person, date_modified=date)
        date += timedelta(days=2)
        project2.incrementTestStat(person, date_modified=date)

        projects = container.getTopTestProjects()
        self.assertNotEquals(projects, None)
        self.assertEquals(projects.count(), 2)
        self.assertEquals(projects[0], project2)
        self.assertEquals(projects[1], project1)


class TestTestStatPersonMixin(TestHasTestStatProjectsMixin):

    def getContainer(self):
        return self.getTestPerson()

    def test_TestRunCount(self):
        container = self.getContainer()
        person = self.getTestPerson()
        self.assertEquals(container.test_run_count, None)

        project = self.factory.makeTestProject()
        project.incrementTestStat(person)
        self.assertEquals(container.test_run_count, 1)

    def test_lastTestRunDate(self):
        container = self.getContainer()
        person = self.getTestPerson()
        self.assertEquals(container.last_test_run_date, None)

        project = self.factory.makeTestProject()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        project.incrementTestStat(person, date_modified=date)
        self.assertEquals(container.last_test_run_date, date)


class TestTestStatTargetMixin(TestHasTestStatPeopleMixin):

    def getContainer(self):
        return self.getTestTarget()

    def test_TestRunCount(self):
        container = self.getContainer()
        target = self.getTestTarget()
        self.assertEquals(container.test_run_count, None)

        person = self.factory.makeTestPerson()
        target.incrementTestStat(person)
        self.assertEquals(container.test_run_count, 1)

    def test_lastTestRunDate(self):
        container = self.getContainer()
        target = self.getTestTarget()
        self.assertEquals(container.last_test_run_date, None)

        person = self.factory.makeTestPerson()
        date = datetime(2010, 1, 1, tzinfo=UTC)
        target.incrementTestStat(person, date_modified=date)
        self.assertEquals(container.last_test_run_date, date)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
