# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "HasTestStats",
    "HasTestStatPeopleMixin",
    "HasTestStatProjectsMixin",
    "TestStat",
    "TestStatPerson",
    "TestStatTarget",
    ]

from datetime import datetime

from storm.locals import (
    AutoReload,
    Int,
    )

from zope.component import getUtility
from zope.interface import implements

from lpresults.database.constants import UTC_NOW
from lpresults.database.datetime import UTCDateTime
from lpresults.database.search import SearchNotEqual

from lpresults.tracker.interfaces.testdistribution import (
    ITestDistributionSet,
    TestDistributionSearchParams,
    )
from lpresults.tracker.interfaces.testperson import (
    ITestPersonSet,
    TestPersonSearchParams,
    )
from lpresults.tracker.interfaces.testproject import (
    ITestProjectSet,
    TestProjectSearchParams,
    )
from lpresults.tracker.interfaces.teststat import (
    IHasTestStats,
    ITestStat,
    ITestStatPerson,
    ITestStatTarget,
    TestStatIncrementParams,
    )
from lpresults.tracker.models.store import get_tracker_store
from lpresults.tracker.models.testpersonhelper import TestPersonHelper
from lpresults.tracker.models.testtarget import (
    TestTargetHelper,
    TestTargetMixin,
    )


class TestStat(TestPersonHelper, TestTargetHelper):

    implements(ITestStat)

    __storm_table__ = "teststat"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    person_id = Int(name="person")
    distro_series_id = Int(name="distro_series")
    project_series_id = Int(name="project_series")
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    date_modified = UTCDateTime(allow_none=False, default=UTC_NOW)
    counter = Int(allow_none=False, default=0)


class HasTestStats:

    def __init__(self, context, test_run_count, last_test_run_date):
        self.context = context
        self.test_run_count = test_run_count
        self.last_test_run_date = last_test_run_date

    def __cmp__(self, other):
        if IHasTestStats.providedBy(other):
            other = other.context

        return cmp(self.context, other)


class HasTestStatDistributionsMixin:

    def getLatestTestDistributions(self, limit=None):
        """See `IHasTestStatDistributions`."""
        return self.getTestDistributions("-last_test_run_date", limit)

    def getTopTestDistributions(self, limit=None):
        """See `IHasTestStatDistributions`."""
        return self.getTestDistributions("-test_run_count", limit)

    def getTestDistributions(self, order_by=None, limit=None):
        search_params = TestDistributionSearchParams(
            date_modified=SearchNotEqual(None),
            order_by=order_by,
            )
        self.updateTestStatParams(search_params)

        test_distributions = getUtility(ITestDistributionSet).search(
            search_params)
        return test_distributions.config(limit=limit)

    def updateTestStatParams(self, params):
        # Do nothing by default
        pass


class HasTestStatPeopleMixin(TestTargetMixin):

    def getLatestTesters(self, limit=None):
        """See `IHasTestStatPeople`."""
        return self.getTesters("-last_test_run_date", limit)

    def getTopTesters(self, limit=None):
        """See `IHasTestStatPeople`."""
        return self.getTesters("-test_run_count", limit)

    def getTesters(self, order_by=None, limit=None):
        search_params = TestPersonSearchParams(
            date_modified=SearchNotEqual(None),
            order_by=order_by,
            )
        self.updateTestStatParams(search_params)

        testers = getUtility(ITestPersonSet).search(search_params)
        return testers.config(limit=limit)

    def updateTestStatParams(self, params):
        # Do nothing by default
        pass


class HasTestStatProjectsMixin:

    def getLatestTestProjects(self, limit=None):
        """See `IHasTestStatProjects`."""
        return self.getTestProjects("-last_test_run_date", limit)

    def getTopTestProjects(self, limit=None):
        """See `IHasTestStatProjects`."""
        return self.getTestProjects("-test_run_count", limit)

    def getTestProjects(self, order_by=None, limit=None):
        search_params = TestProjectSearchParams(
            date_modified=SearchNotEqual(None),
            order_by=order_by,
            )
        self.updateTestStatParams(search_params)

        test_projects = getUtility(ITestProjectSet).search(search_params)
        return test_projects.config(limit=limit)

    def updateTestStatParams(self, params):
        # Do nothing by default
        pass


class TestStatPerson(
    HasTestStats, HasTestStatDistributionsMixin, HasTestStatProjectsMixin):

    def updateTestStatParams(self, params):
        params.setPerson(self)


class TestStatTarget(HasTestStats, HasTestStatPeopleMixin):

    def incrementTestStat(self, *args, **kwargs):
        """See `ITestStatTarget`."""
        increment_params = TestStatIncrementParams(*args, **kwargs)
        self.updateTestStatParams(increment_params)

        # This makes sure the target is defined
        person = increment_params.person
        target = increment_params.target

        person_id = getattr(person, "id", None)
        distro_series_id = getattr(
            increment_params.distro_series, "id", None)
        project_series_id = getattr(
            increment_params.project_series, "id", None)
        date_modified = increment_params.date_modified or datetime.utcnow()

        counter = get_tracker_store().execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (person_id, distro_series_id, project_series_id, date_modified,),
            ).get_one()[0]

        # Since we've updated the database behind Storm's back,
        # some caches can be updated instead of invalidating all.
        if ITestStatPerson.providedBy(person):
            person.test_run_count = counter
            person.last_test_run_date = date_modified
        if ITestStatTarget.providedBy(target):
            target.test_run_count = counter
            target.last_test_run_date = date_modified

        return counter

    def updateTestStatParams(self, params):
        self.updateTestTargetParams(params)
