# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestSystem",
    "test_system_set",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Max,
    Sum,
    )

from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple

from lpresults.hardware.interfaces.system import ISystem
from lpresults.hardware.models.system import (
    System,
    SystemSet,
    )
from lpresults.hardware.models.systemunit import SystemUnit
from lpresults.hardware.models.unit import UnitTargetMixin

from lpresults.tracker.interfaces.testsystem import (
    ITestSystem,
    ITestSystemSet,
    TestSystemSearchParams,
    )
from lpresults.tracker.models.testrank import (
    HasTestRankSystemsMixin,
    TestRank,
    )
from lpresults.tracker.models.teststat import (
    TestStat,
    HasTestStats,
    )


class TestSystem(HasTestStats, UnitTargetMixin):
    """See `ITestSystem`."""
    implements(ITestSystem)
    delegates(ISystem)


class TestSystemSet(SystemSet, HasTestRankSystemsMixin):
    """See `ITestSystemSet`."""
    implements(ITestSystemSet)

    search_params_factory = TestSystemSearchParams

    entry_type = ITestSystemSet

    def get(self, id):
        """See `ISystemSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def searchSystems(self, text=None):
        """See ITestSystemSet."""
        # Only search names for now
        search_params = self.search_params_factory(search_text=text)
        return self.search(search_params)

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(TestSystemSet, self).buildSearchExpressions(
            search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(System.id, search_params.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        if search_params.distro_series:
            expressions.append(
                search_to_expression(
                    TestStat.distro_series_id,
                    search_params.distro_series,
                    modifier=lambda x: x.id))

        if search_params.project_series:
            expressions.append(
                search_to_expression(
                    TestStat.project_series_id,
                    search_params.project_series,
                    modifier=lambda x: x.id))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestSystemSet, self).buildSearchResult(search_params)

        return result.group_by(System)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestSystemSet, self).buildSearchTables(search_params)

        tables.append(
            LeftJoin(SystemUnit,
                SystemUnit.system_id == System.id))

        tables.append(
            LeftJoin(TestRank,
                TestRank.system_unit_id == SystemUnit.id))

        tables.append(
            LeftJoin(TestStat,
                TestStat.id == TestRank.test_stat_id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestSystem(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(TestSystemSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


test_system_set = TestSystemSet()
