# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "TestSystemUnit",
    "TestSystemUnitSet",
    ]

from storm.expr import (
    Alias,
    LeftJoin,
    Sum,
    )
from storm.locals import Max

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implements

from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTimeAlias
from lpresults.database.search import search_to_expression

from lpresults.registry.formatters import item_to_tuple

from lpresults.hardware.interfaces.systemunit import ISystemUnit
from lpresults.hardware.models.systemunit import (
    SystemUnit,
    SystemUnitSet,
    )
from lpresults.hardware.models.systemunithelper import (
    SystemUnitHelper,
    SystemUnitTwin,
    )

from lpresults.tracker.interfaces.testsystem import ITestSystemSet
from lpresults.tracker.interfaces.testsystemunit import (
    ITestSystemUnit,
    ITestSystemUnitSet,
    ITestSystemUnitTarget,
    )
from lpresults.tracker.interfaces.testunit import TestUnitSearchParams
from lpresults.tracker.models.testrank import TestRank
from lpresults.tracker.models.teststat import (
    HasTestStats,
    TestStat,
    )
from lpresults.tracker.models.teststore import get_test_stores


class TestSystemUnit(HasTestStats):
    """Users within the Django authentication system are represented by
    this model.
    """
    implements(ITestSystemUnit)
    delegates(ISystemUnit)

    @Lazy
    def system(self):
        """See `ISystemUnit`."""
        return getUtility(ITestSystemSet).get(self.context.system_id)


class TestSystemUnitSet(SystemUnitSet):
    """See `ITestSystemUnitSet`."""

    implements(ITestSystemUnitSet)

    search_params_factory = TestUnitSearchParams

    def new(self, identifiers):
        """See `IUnitSet`."""
        series = SystemUnit(self.target.context, identifiers)
        return self.get(series.id)

    def get(self, id):
        """See `IUnitSet`."""
        search_params = self.search_params_factory(id=id)
        return self.search(search_params).one()

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(
            TestSystemUnitSet, self).buildSearchExpressions(search_params)

        if search_params.id is not None:
            expressions.append(
                search_to_expression(SystemUnit.id, search_params.id))

        if search_params.person:
            expressions.append(
                search_to_expression(
                    TestStat.person_id,
                    search_params.person,
                    modifier=lambda x: x.id))

        if search_params.date_modified:
            expressions.append(
                search_to_expression(
                    TestStat.date_modified,
                    search_params.date_modified))

        return expressions

    def buildSearchResult(self, search_params):
        """See SearchSet."""
        result = super(TestSystemUnitSet, self).buildSearchResult(
            search_params)

        return result.group_by(SystemUnit)

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(TestSystemUnitSet, self).buildSearchTables(
            search_params)

        tables.append(
            LeftJoin(TestRank,
                TestRank.system_unit_id == SystemUnit.id))

        tables.append(
            LeftJoin(TestStat,
                TestStat.id == TestRank.test_stat_id))

        return tables

    def getSearchResultDecorator(self, result):
        """See SearchSet."""
        return TestSystemUnit(*result)

    def getSearchResultTables(self):
        """See SearchSet."""
        result_tables = super(TestSystemUnitSet, self).getSearchResultTables()

        return item_to_tuple(result_tables) + (
            Alias(Sum(TestStat.counter), name="test_run_count"),
            UTCDateTimeAlias(
                Max(TestStat.date_modified), name="last_test_run_date"),
            )


class TestSystemUnitTargetMixin(SystemUnitHelper):

    implements(ITestSystemUnitTarget)

    def setSystemUnit(self, make, model, identifiers):
        """See `ITestSystemUnitTarget`."""
        system = getUtility(ITestSystemSet).retrieve(make, model)
        self.system_unit = system.retrieveUnit(identifiers)


def system_unit_created(unit, event):
    # Flush each store to ensure correct ordering of
    # SQL statements emitted by storm.
    for name, test_store in get_test_stores():
        unit_twin = SystemUnitTwin(unit)
        test_store.add(unit_twin)
        test_store.flush()

    return unit
