# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

from datetime import datetime

from unittest import defaultTestLoader

from lpresults.database.store import get_store

from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import DatabaseLayer


class TestResultTestCase(TestCaseWithFactory):

    layer = DatabaseLayer

    default_status = 0
    default_output = u""
    default_date_created = datetime(2010, 1, 1)

    def setUp(self):
        super(TestResultTestCase, self).setUp()

        self.store = get_store("test-1")

        self.project_series = self.factory.getUniqueInteger()
        self.project = self.factory.getUniqueInteger()
        self.store.execute(
            "INSERT INTO projectseriestwin VALUES (?, ?)",
            (self.project_series, self.project),
            )

        name = self.factory.getUniqueString(u"test-name")
        self.test_case = self.store.execute(
            "SELECT get_test_case(?, ?, ?)",
            (None, self.project_series, name,),
            ).get_one()[0]

    def tearDown(self):
        super(TestResultTestCase, self).tearDown()

        self.store.execute(
            "DELETE FROM projectseriestwin WHERE id = ?",
            (self.project_series,),
            )

    def getWhereCondition(self, status=None, output=None, date_created=None):
        where = ["test_case = %d" % self.test_case]
        if status is not None:
            where.append("status = %d" % status)
        if output is not None:
            where.append("output = '%s'" % output)
        if date_created is not None:
            where.append("date_created = '%s'" % date_created)

        return " AND ".join(where)

    def insertTestResult(self, status=None, output=None, date_created=None):
        if status is None:
            status = self.default_status
        if output is None:
            output = self.default_output
        if date_created is None:
            date_created = self.default_date_created

        self.store.execute(
            "INSERT INTO testresult VALUES (?, ?, ?, ?)",
            (self.test_case, status, output, date_created,),
            )

    def removeTestResult(self, **kwargs):
        where_condition = self.getWhereCondition(**kwargs)
        self.store.execute(
            "DELETE FROM testresult WHERE %s"
            % where_condition)

    def countTestResults(self, **kwargs):
        where_condition = self.getWhereCondition(**kwargs)
        return self.store.execute(
            "SELECT COUNT(*) FROM testresult WHERE %s"
            % where_condition).get_one()[0]

    def searchTestResults(self, **kwargs):
        keys = [
            "test_case",
            "status",
            "output",
            "date_created",
            "date_deleted"]
        select_expression = ", ".join(keys)
        where_condition = self.getWhereCondition(**kwargs)
        values = self.store.execute(
            "SELECT %s FROM testresult WHERE %s"
            % (select_expression, where_condition)).get_all()

        return [dict(zip(keys, value)) for value in values]

    def assertCountTestResults(self, count, **kwargs):
        test_result_count = self.countTestResults(**kwargs)
        self.assertEquals(test_result_count, count)


class TestTestResultInsert(TestResultTestCase):

    def test_insertOneResult(self):
        self.assertCountTestResults(0)
        self.insertTestResult()
        self.assertCountTestResults(1)

        test_result = self.searchTestResults()[0]
        self.assertEquals(test_result["test_case"], self.test_case)
        self.assertEquals(test_result["status"], self.default_status)
        self.assertEquals(test_result["output"], self.default_output)
        self.assertEquals(
            test_result["date_created"],
            self.default_date_created)
        self.assertTrue(
            test_result["date_deleted"] > self.default_date_created)

    def test_insertTwoDifferentResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two rows after creating different test results
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.assertCountTestResults(2)

        # Both test results should exist
        first_result = self.searchTestResults(date_created=first_date)[0]
        self.assertNotEquals(first_result, None)
        second_result = self.searchTestResults(date_created=second_date)[0]
        self.assertNotEquals(second_result, None)

        # The date_deleted of the first test result
        # should be the same as the date_created of the second test result
        self.assertEquals(
            first_result["date_deleted"],
            second_result["date_created"])

    def test_insertTwoSimilarResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two rows after creating different test results
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=0, date_created=second_date)
        self.assertCountTestResults(1)

        # Only the first test result should exist
        test_results = self.searchTestResults(date_created=first_date)
        self.assertEquals(len(test_results), 1)
        test_result = test_results[0]
        self.assertEquals(test_result["status"], 0)
        self.assertEquals(test_result["date_created"], first_date)

        test_results = self.searchTestResults(date_created=second_date)
        self.assertEquals(len(test_results), 0)

    def test_insertTwoSimilarResultsReversed(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # Add test results in reversed order
        self.insertTestResult(status=0, date_created=second_date)
        self.insertTestResult(status=0, date_created=first_date)
        self.assertCountTestResults(1)

        # Only the first test result should exist
        test_results = self.searchTestResults(date_created=first_date)
        self.assertEquals(len(test_results), 1)
        test_result = test_results[0]
        self.assertEquals(test_result["status"], 0)
        self.assertEquals(test_result["date_created"], first_date)

        test_results = self.searchTestResults(date_created=second_date)
        self.assertEquals(len(test_results), 0)


class TestTestResultRemove(TestResultTestCase):

    def test_deleteOneResult(self):
        self.assertCountTestResults(0)
        self.insertTestResult()
        self.assertCountTestResults(1)
        self.removeTestResult()
        self.assertCountTestResults(0)

    def test_deleteFirstOfTwoResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two rows after creating different test results
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.assertCountTestResults(2)

        # There should only be one test result
        # after removing the first one
        self.removeTestResult(date_created=first_date)
        self.assertCountTestResults(1)

        # The date_created of the second test result should no longer exist
        removed_results = self.searchTestResults(date_created=second_date)
        self.assertEquals(removed_results, [])

        # Instead, the date_created of the second test result should now
        # be the same as the first test result
        remaining_results = self.searchTestResults(date_created=first_date)
        self.assertNotEquals(remaining_results, [])

        # There should no longer be any test results
        # after removing the last one
        self.removeTestResult(date_created=first_date)
        self.assertCountTestResults(0)

    def test_deleteSecondOfTwoResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two test results after creating different ones
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.assertCountTestResults(2)

        # The date_deleted of the first test result
        # should be the same as the date_created of the second test result
        first_result = self.searchTestResults(date_created=first_date)[0]
        second_result = self.searchTestResults(date_created=second_date)[0]
        self.assertEquals(
            first_result["date_deleted"],
            second_result["date_created"])

        # There should only be one test result
        # after removing the second one
        self.removeTestResult(date_created=second_date)
        self.assertCountTestResults(1)

        # The date_deleted of the first test result
        # should be adjusted to the same as the second test result
        remaining_result = self.searchTestResults(date_created=first_date)[0]
        self.assertEquals(
            remaining_result["date_deleted"],
            second_result["date_deleted"])

        # There should no longer be any test results
        # after removing the last one
        self.removeTestResult(date_created=first_date)
        self.assertCountTestResults(0)

    def test_deleteSecondOfThreeDifferentResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)
        third_date = datetime(2010, 1, 3)
        self.assertCountTestResults(0, date_created=third_date)

        # There should be three test results after creating different ones
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.insertTestResult(status=2, date_created=third_date)
        self.assertCountTestResults(3)

        # There should only be two test results
        # after removing the second one
        self.removeTestResult(date_created=second_date)
        self.assertCountTestResults(2)

        # The date_deleted of the first test result
        # should be the same as the date_created of the third test result
        first_result = self.searchTestResults(date_created=first_date)[0]
        self.assertEquals(first_result["date_deleted"], third_date)

        # The date_created of the third test result
        # should be the same as before
        third_result = self.searchTestResults(date_created=third_date)[0]
        self.assertEquals(third_result["date_created"], third_date)

    def test_deleteSecondOfThreeSimilarResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)
        third_date = datetime(2010, 1, 3)
        self.assertCountTestResults(0, date_created=third_date)

        # There should be three test results after creating different ones
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.insertTestResult(status=0, date_created=third_date)
        self.assertCountTestResults(3)

        third_result = self.searchTestResults(date_created=third_date)[0]

        # There should only be one test result
        # after removing the second one
        self.removeTestResult(date_created=second_date)
        self.assertCountTestResults(1)

        # The date_created of the remaining test result
        # should be the same as the date_created of the first test result
        remaining_result = self.searchTestResults(date_created=first_date)[0]
        self.assertNotEquals(remaining_result, None)
        self.assertEquals(remaining_result["date_created"], first_date)

        # The date_deleted of the remaining test result
        # should be the same as the date_deleted of the third test result
        self.assertEquals(
            remaining_result["date_deleted"],
            third_result["date_deleted"])


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
