# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__all__ = [
    "create",
    ]

from storm.zope.schema import ZSchema

from lpresults.hardware import schema as hardware
from lpresults.registry import schema as registry


def unique_seen(iterable):
    seen = set()
    seen_add = seen.add
    for element in iterable:
        if element not in seen:
            seen_add(element)
            yield element


def create():
    from lpresults.tracker.schema import tracker as patch_package

    # Add aliased store schemas
    return ZSchema(
        unique_seen(registry.CREATES + hardware.CREATES + CREATES),
        unique_seen(DROPS + hardware.DROPS + registry.DROPS),
        unique_seen(DELETES + hardware.DELETES + registry.DELETES),
        patch_package)


CREATES = [
"""
GRANT SELECT
    ON TABLE patch
    TO lpresults
""",
"""
CREATE TABLE teststat (
    id SERIAL NOT NULL PRIMARY KEY,
    person INTEGER,
    distro_series INTEGER,
    project_series INTEGER,
    date_created TIMESTAMP WITHOUT TIME ZONE NOT NULL
        DEFAULT (CURRENT_TIMESTAMP AT TIME ZONE 'UTC'),
    date_modified TIMESTAMP WITHOUT TIME ZONE NOT NULL,
    counter INTEGER NOT NULL DEFAULT 0,
    CONSTRAINT valid_dates CHECK (date_modified >= date_created),
    CONSTRAINT valid_target CHECK (
        (distro_series IS NOT NULL) OR (project_series IS NOT NULL)))
""",
"""
GRANT SELECT, INSERT, UPDATE, DELETE
    ON TABLE teststat
    TO lpresults
""",
"""
GRANT SELECT, UPDATE
    ON TABLE teststat_id_seq
    TO lpresults
""",
"""
CREATE UNIQUE INDEX teststat__person__distro_series__key
    ON teststat (person, distro_series)
    WHERE person IS NOT NULL AND distro_series IS NOT NULL
""",
"""
CREATE UNIQUE INDEX teststat__person__project_series__key
    ON teststat (person, project_series)
    WHERE person IS NOT NULL AND project_series IS NOT NULL
""",
"""
CREATE TABLE testrank (
    id SERIAL NOT NULL PRIMARY KEY,
    test_stat INTEGER NOT NULL,
    system_unit INTEGER NOT NULL,
    rank INTEGER NOT NULL)
""",
"""
GRANT SELECT, INSERT, UPDATE, DELETE
    ON TABLE testrank
    TO lpresults
""",
"""
GRANT SELECT, UPDATE
    ON TABLE testrank_id_seq
    TO lpresults
""",
"""
CREATE UNIQUE INDEX testrank__stat__system_unit__key
    ON testrank (test_stat, system_unit)
""",
"""
CREATE LANGUAGE plpgsql
""",
"""
CREATE FUNCTION teststat_insert_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    -- Set date_modified to date_created if not defined
    IF NEW.date_modified IS NULL THEN
        NEW.date_modified := NEW.date_created;
    END IF;

    RETURN NEW;
END
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION teststat_insert_before()
    TO lpresults
""",
"""
CREATE TRIGGER teststat_insert_before_trigger
    BEFORE INSERT ON teststat
    FOR EACH ROW
    EXECUTE PROCEDURE teststat_insert_before()
""",
"""
CREATE FUNCTION teststat_update_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    IF NEW.person <> OLD.person THEN
        RAISE EXCEPTION 'Cannot update person';
    END IF;

    -- Set teststat columns to match constraints
    IF NEW.date_modified <> OLD.date_modified THEN
        IF NEW.date_created > NEW.date_modified THEN
            NEW.date_created = NEW.date_modified;
        END IF;

        IF NEW.date_modified < OLD.date_modified THEN
            NEW.date_modified = OLD.date_modified;
        END IF;

        IF NEW.counter = OLD.counter THEN
            NEW.counter := OLD.counter + 1;
        END IF;
    END IF;

    RETURN NEW;
END;
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION teststat_update_before()
    TO lpresults
""",
"""
CREATE TRIGGER teststat_update_before_trigger
    BEFORE UPDATE ON teststat
    FOR EACH ROW
    EXECUTE PROCEDURE teststat_update_before()
""",
"""
CREATE FUNCTION increment_teststat(
    person_id INTEGER,
    distro_series_id INTEGER,
    project_series_id INTEGER,
    increment_date TIMESTAMP)
    RETURNS INTEGER
LANGUAGE plpgsql AS
$$
DECLARE
    teststat_counter INTEGER;
BEGIN
    LOOP
        BEGIN
            UPDATE teststat
                SET date_modified = increment_date
                WHERE person IS NOT DISTINCT FROM person_id
                AND distro_series IS NOT DISTINCT FROM distro_series_id
                AND project_series IS NOT DISTINCT FROM project_series_id
                RETURNING counter INTO teststat_counter;

            IF NOT FOUND THEN
                INSERT INTO teststat (
                    person,
                    distro_series,
                    project_series,
                    date_created,
                    counter)
                VALUES (
                    person_id,
                    distro_series_id,
                    project_series_id,
                    increment_date,
                    1)
                RETURNING counter INTO teststat_counter;
            END IF;

            RETURN teststat_counter;
        EXCEPTION
            WHEN unique_violation THEN
                NULL;  -- do nothing
        END;
    END LOOP;
END;
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION increment_teststat(INTEGER, INTEGER, INTEGER, TIMESTAMP)
    TO lpresults
""",
"""
CREATE FUNCTION set_testrank(
    person_id INTEGER,
    distro_series_id INTEGER,
    project_series_id INTEGER,
    system_unit_id INTEGER,
    set_rank INTEGER)
    RETURNS VOID
LANGUAGE plpgsql AS
$$
DECLARE
    teststat_id INTEGER;
BEGIN
    SELECT id INTO teststat_id FROM teststat
        WHERE person IS NOT DISTINCT FROM person_id
        AND distro_series IS NOT DISTINCT FROM distro_series_id
        AND project_series IS NOT DISTINCT FROM project_series_id;
    IF NOT FOUND THEN
        RAISE EXCEPTION 'teststat must exist already';
    END IF;

    LOOP
        BEGIN
            UPDATE testrank
                SET rank = set_rank
                WHERE test_stat = teststat_id
                AND system_unit = system_unit_id;

            IF NOT FOUND THEN
                INSERT INTO testrank (
                    test_stat,
                    system_unit,
                    rank)
                VALUES (
                    teststat_id,
                    system_unit_id,
                    set_rank);
            END IF;

            RETURN;
        EXCEPTION
            WHEN unique_violation THEN
                NULL;  -- do nothing
        END;
    END LOOP;
END;
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION set_testrank(INTEGER, INTEGER, INTEGER, INTEGER, INTEGER)
    TO lpresults
""",
"""
ALTER TABLE ONLY testrank
    ADD CONSTRAINT testrank__test_stat__fk
    FOREIGN KEY (test_stat)
    REFERENCES teststat(id) ON DELETE CASCADE
""",
"""
ALTER TABLE ONLY testrank
    ADD CONSTRAINT testrank__system_unit__fk
    FOREIGN KEY (system_unit)
    REFERENCES systemunit(id) ON DELETE CASCADE
""",
"""
ALTER TABLE ONLY teststat
    ADD CONSTRAINT teststat__person__fk
    FOREIGN KEY (person)
    REFERENCES person(id) ON DELETE CASCADE
""",
"""
ALTER TABLE ONLY teststat
    ADD CONSTRAINT teststat__distro_series__fk
    FOREIGN KEY (distro_series)
    REFERENCES distroseries(id) ON DELETE CASCADE
""",
"""
ALTER TABLE ONLY teststat
    ADD CONSTRAINT teststat__project_series__fk
    FOREIGN KEY (project_series)
    REFERENCES projectseries(id) ON DELETE CASCADE
""",
]

DROPS = [
"DROP FUNCTION set_testrank(INTEGER, INTEGER, INTEGER, INTEGER, INTEGER)",
"DROP FUNCTION increment_teststat(INTEGER, INTEGER, INTEGER, TIMESTAMP)",
"DROP FUNCTION teststat_update_before",
"DROP FUNCTION teststat_insert_before",
"DROP LANGUAGE plpgsql",
"DROP TABLE teststat",
]

DELETES = [
"DELETE FROM teststat",
]
