# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

from datetime import (
    datetime,
    timedelta,
    )

from unittest import defaultTestLoader

from lpresults.tracker.models.store import get_tracker_store
from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import DatabaseLayer


class TestTestStat(TestCaseWithFactory):

    layer = DatabaseLayer

    def setUp(self):
        super(TestTestStat, self).setUp()

        self.store = get_tracker_store()

        person_name = self.factory.getUniqueString(u"person-name")
        person_identity = self.factory.getUniqueString(u"person-identity")
        self.person = self.store.execute(
            "INSERT INTO person "
            "(name, identity) "
            "VALUES (?, ?) RETURNING id",
            (person_name, person_identity,),
            ).get_one()[0]

        project_name = self.factory.getUniqueString(u"project-name")
        project = self.store.execute(
            "INSERT INTO project "
            "(name) "
            "VALUES (?) RETURNING id",
            (project_name,),
            ).get_one()[0]

        project_series_name = self.factory.getUniqueString(
            u"project-series-name")
        self.project_series = self.store.execute(
            "INSERT INTO projectseries "
            "(project, name) "
            "VALUES (?, ?) RETURNING id",
            (project, project_series_name,),
            ).get_one()[0]

    def test_incrementTestStat(self):
        # Check that there is no teststat row to start with
        result = self.store.execute(
            "SELECT counter "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertEquals(result, None)

        # Check that a teststat row is created after calling
        # increment_teststat
        date_modified = datetime(2010, 1, 1)
        self.store.execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (self.person, None, self.project_series, date_modified,),
            ).get_one()[0]
        result = self.store.execute(
            "SELECT counter "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertNotEquals(result, None)

    def test_incrementTestStatCounter(self):
        # Check that the counter is returned
        date_modified = datetime(2010, 1, 1)
        counter = self.store.execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (self.person, None, self.project_series, date_modified,),
            ).get_one()[0]
        self.assertEquals(counter, 1)

        result = self.store.execute(
            "SELECT counter "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertEquals(result[0], counter)

        # Check that the counter is incremented
        date_modified += timedelta(days=1)
        counter = self.store.execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (self.person, None, self.project_series, date_modified,),
            ).get_one()[0]
        self.assertEquals(counter, 2)

        result = self.store.execute(
            "SELECT counter "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertEquals(result[0], counter)

    def test_incrementTestStatDateModified(self):
        # Check that the date_modified is set
        new_modified = datetime(2010, 1, 1)
        self.store.execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (self.person, None, self.project_series, new_modified,))
        result = self.store.execute(
            "SELECT date_modified "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertEquals(result[0], new_modified)

        # Check that the date_modified is set to the most recent date
        old_modified = new_modified - timedelta(days=1)
        self.store.execute(
            "SELECT increment_teststat(?, ?, ?, ?)",
            (self.person, None, self.project_series, old_modified,))
        result = self.store.execute(
            "SELECT date_modified "
            "FROM teststat WHERE person = ? AND project_series = ?",
            (self.person, self.project_series,),
            ).get_one()
        self.assertEquals(result[0], new_modified)

    def test_insert(self):
        teststat_id = self.store.execute(
            "INSERT INTO teststat "
            "(person, project_series) "
            "VALUES (?, ?) RETURNING id",
            (self.person, self.project_series,),
            ).get_one()[0]

        # Check that the date_modified is the same as the date_created
        date_created, date_modified = self.store.execute(
            "SELECT date_created, date_modified "
            "FROM teststat WHERE id = ?",
            (teststat_id,),
            ).get_one()[:]
        self.assertEquals(date_created, date_modified)

    def test_insertWithDateCreated(self):
        date_created = datetime(2010, 1, 1)
        teststat_id = self.store.execute(
            "INSERT INTO teststat "
            "(person, project_series, date_created) "
            "VALUES (?, ?, ?) RETURNING id",
            (self.person, self.project_series, date_created,),
            ).get_one()[0]

        # Check that the date_modified is the same as the date_created
        date_created, date_modified = self.store.execute(
            "SELECT date_created, date_modified "
            "FROM teststat WHERE id = ?",
            (teststat_id,),
            ).get_one()[:]
        self.assertEquals(date_created, date_modified)
        self.assertEquals(date_created, datetime(2010, 1, 1))

    def test_insertWithDateCreatedAndDateModified(self):
        date_created = datetime(2010, 1, 1)
        date_modified = date_created + timedelta(days=1)
        teststat_id = self.store.execute(
            "INSERT INTO teststat "
            "(person, project_series, date_created, date_modified) "
            "VALUES (?, ?, ?, ?) RETURNING id",
            (self.person, self.project_series, date_created, date_modified),
            ).get_one()[0]

        # Check that the date_modified is the same as the date_created
        date_created, date_modified = self.store.execute(
            "SELECT date_created, date_modified "
            "FROM teststat WHERE id = ?",
            (teststat_id,),
            ).get_one()[:]
        self.assertNotEquals(date_created, date_modified)
        self.assertEquals(date_modified - date_created, timedelta(days=1))


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
