# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Config",
    ]

import os
import re
import logging

from ConfigParser import ConfigParser


class IncludeDict(dict):

    def __init__(self, parser):
        self._parser = parser

    def __setitem__(self, key, value):
        if key == "includes":
            for path in value.split():
                path = self._parser._interpolate("DEFAULT", None, path, self)
                path = os.path.expanduser(path)
                if not os.path.exists(path):
                    raise Exception("No such configuration file: %s" % path)

                if os.path.isdir(path):
                    logging.info(
                        "Parsing config filenames from directory: %s",
                        path)

                    def walk_func(arg, directory, names):
                        for name in names:
                            path = os.path.join(directory, name)
                            if not os.path.isdir(path):
                                arg._parser.read(path)

                    os.path.walk(path, walk_func, self)

                else:
                    logging.info("Parsing config filename: %s", path)
                    self._parser.read(path)

        else:
            super(IncludeDict, self).__setitem__(key, value)


class ConfigDefaults:

    def __init__(self, parent, name, attributes={}):
        self.parent = parent
        self.name = name
        self.attributes = attributes

    def __getattr__(self, name):
        if name in self.attributes:
            return self.get(name)

        raise AttributeError(name)

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name):
        return self.attributes.get(name)

    def set(self, name, value):
        value = value.strip()
        self.parent._parser.set(self.name, name, value)
        self.attributes[name] = value


class Config:

    def __init__(self):
        self._parser = ConfigParser()
        self._parser._defaults = IncludeDict(self._parser)

        # Copy attributes from the parser to avoid one additional
        # function call on each access.
        for attr in ["has_section", "remove_section"]:
            setattr(self, attr, getattr(self._parser, attr))

    def read_configs(self, configs):
        for config in configs:
            match = re.match("(.*)/([^/]+)=(.*)", config)
            if not match:
                raise Exception("Invalid config string: %s" % config)

            (name, option, value) = match.groups()
            if not self._parser.has_section(name):
                self._parser.add_section(name)

            self._parser.set(name, option, value)

        return self

    def read_file(self, file, filename="<stream>"):
        logging.info("Reading configurations from: %s", filename)

        self._parser.readfp(file, filename)
        return self

    def read_filename(self, filename):
        if not os.path.exists(filename):
            raise Exception("No such configuration file: %s" % filename)

        file = open(filename, "r")
        return self.read_file(file, filename)

    def get_defaults(self):
        attributes = self._parser.defaults()
        return ConfigDefaults(self, "DEFAULT", attributes)

    def get_section(self, name):
        if self._parser.has_section(name):
            attributes = dict(self._parser.items(name))
            return ConfigDefaults(self, name, attributes)

        else:
            return None

    def get_section_names(self):
        return self._parser.sections()

    def add_section(self, name):
        self._parser.add_section(name)
        return self.get_section(name)
