# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

from lpresults.registry.formatters import string_to_type


class Settings:
    """A holder for settings which can be updated and configured."""

    def __init__(self, schema, module=None):
        """Initialize with the specified schema.

        The schema consists of a dictionary or module defining the
        keys supported by the schema. Only the keys in upper case will
        be retained.

        The optional module argument is to enable the returned instance
        to behave like a module. Only the underscore attributes will be
        retained.

        :param schema: Dictionary/module of keys and values.
        :param module: Optional module of type attributes.
        """
        if not hasattr(schema, "items"):
            schema = vars(schema)

        for name, value in schema.items():
            # Only retain upper case names from the schema
            if name == name.upper():
                super(Settings, self).__setattr__(name, value)

        if module is not None:
            for name, value in vars(module).items():
                # Only retain underscore names from the module
                if name.startswith("__") and name.endswith("__"):
                    super(Settings, self).__setattr__(name, value)

    def __setattr__(self, name, value):
        """Set the attributes based on the initialized schema.

        If the name does not exist in the schema, it will be silently
        ignored. If the type of the value cannot be converted to the
        type in the schema, a ValueError exception will be raised.

        :param name: Name of the attribute.
        :param value: New value for the given name.
        """
        # Only consider upper case attributes
        if name != name.upper():
            return

        if not hasattr(self, name):
            if "_" not in name:
                return

            # Try to match parts of the name recursively
            name, key = name.rsplit("_", 1)
            if isinstance(value, dict) and len(value) == 1:
                old_key, old_value = value.items()[0]
                key = "_".join([key, old_key])
                value = old_value

            return setattr(self, name, {key: value})

        old_value = getattr(self, name)
        if isinstance(value, dict) \
           and isinstance(old_value, dict):
            for key in value.iterkeys():

                # Try to match the keys irrespectively
                # of dashes and case sensitivity
                for old_key in old_value.iterkeys():
                    if key.replace("-", "_").upper() \
                       == old_key.replace("-", "_").upper():
                        old_type = type(old_value[old_key])
                        value[old_key] = old_type(value.pop(key))
                        break
                else:
                    return

            for key in old_value.iterkeys():
                if key not in value:
                    value[key] = old_value[key]

        elif isinstance(value, basestring):
            if isinstance(old_value, (list, tuple)):
                value = type(old_value)(value.split())

            else:
                value = string_to_type(value, type(old_value))

        else:
            value = type(old_value)(value)

        return super(Settings, self).__setattr__(name, value)

    def configure(self, config):
        """Configure the settings using a `Config` object.

        :param config: `Config` object.
        """
        attributes = config.get_defaults().attributes
        for name, value in attributes.iteritems():
            name = name.replace("-", "_").upper()
            if hasattr(self, name):
                setattr(self, name, value)

        for section_name in config.get_section_names():
            attributes = config.get_section(section_name).attributes
            section_name = section_name.replace("-", "_").upper()

            # If the section name matches, assumes
            # the attributes are a dictionary
            if hasattr(self, section_name):
                setattr(self, section_name, attributes)

            # Otherwise, iterate over each attribute
            else:
                for name, value in attributes.iteritems():
                    name = name.replace("-", "_").upper()
                    name = "_".join([section_name, name])
                    if hasattr(self, name):
                        setattr(self, name, value)

        return self

    def update(self, *args, **kwargs):
        """Update the settings with keys and values.

        :param args: Dictionary/iterable arguments.
        :param kwargs: Keyword arguments.
        """
        for arg in args:
            if hasattr(arg, "items"):
                arg = arg.iteritems()

            for key, value in arg:
                setattr(self, key, value)

        for key, value in kwargs.iteritems():
            setattr(self, key, value)

        return self
