# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

import sys

from cStringIO import StringIO

from unittest import (
    defaultTestLoader,
    TestCase,
    )

from lpresults.tracker.services.config import Config
from lpresults.tracker.services.settings import Settings


class TestSettings(TestCase):

    def test_schemaUpper(self):
        # Only upper case keys should be created
        settings = Settings({"FOO": "bar"})
        self.assertTrue(hasattr(settings, "FOO"))

    def test_schemaLower(self):
        # Lower case keys should be ignored
        settings = Settings({"foo": "bar"})
        self.assertFalse(hasattr(settings, "foo"))
        self.assertFalse(hasattr(settings, "FOO"))

    def test_module(self):
        module = sys.modules[__name__]
        settings = Settings({}, module=module)
        self.assertEquals(settings.__file__, __file__)
        self.assertEquals(settings.__name__, __name__)
        self.assertEquals(settings.__package__, __package__)

    def test_updateBool(self):
        settings = Settings({"FOO": True})
        self.assertEquals(settings.FOO, True)

        # Boolean should remain the same type
        settings = Settings({"FOO": False})
        self.assertEquals(settings.FOO, False)

        # True/false string should be converted to bool
        settings.update({"FOO": "true"})
        self.assertEquals(settings.FOO, True)

        # Yes/no string should be converted to bool
        settings.update({"FOO": "no"})
        self.assertEquals(settings.FOO, False)

    def test_updateInt(self):
        settings = Settings({"FOO": 0})
        self.assertEquals(settings.FOO, 0)

        # Integer should remain the same type
        settings.update({"FOO": 1})
        self.assertEquals(settings.FOO, 1)

        # String should be converted to int
        settings.update({"FOO": "2"})
        self.assertEquals(settings.FOO, 2)

        # Invalid integer should raise a value error
        self.assertRaises(
            ValueError,
            settings.update,
            {"FOO": "a"})

    def test_updateString(self):
        settings = Settings({"FOO": "0"})
        self.assertEquals(settings.FOO, "0")

        # String should remain the same type
        settings.update({"FOO": "1"})
        self.assertEquals(settings.FOO, "1")

        # Integer hsould be converted to string
        settings.update({"FOO": 2})
        self.assertEquals(settings.FOO, "2")

    def test_updateDict(self):
        settings = Settings({"FOO": {"bar": 0}})
        self.assertEquals(settings.FOO, {"bar": 0})

        settings.update({"FOO_BAR": 1})
        self.assertEquals(settings.FOO, {"bar": 1})

    def test_configureInt(self):
        settings = Settings({"FOO_BAR": 0})
        self.assertEquals(settings.FOO_BAR, 0)

        config = Config().read_file(StringIO("""
[foo]

bar = 1
"""))
        settings.configure(config)
        self.assertEquals(settings.FOO_BAR, 1)

    def test_configureDict(self):
        settings = Settings({"FOO": {"bar": 0}})
        self.assertEquals(settings.FOO, {"bar": 0})

        config = Config().read_file(StringIO("""
[foo]

bar = 1
"""))
        settings.configure(config)
        self.assertEquals(settings.FOO, {"bar": 1})


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
