# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "WebServiceConfiguration",
    "WebServiceRoot",
    "WebServiceRootAbsoluteURL",
    ]

from os import path
from decimal import Decimal

from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string

from zope.component import (
    adapts,
    getGlobalSiteManager,
    getUtility,
    )
from zope.interface import implements
from zope.security.management import queryInteraction
from zope.traversing.browser import absoluteURL

from lazr.restful.frameworks.django import DjangoWebServiceConfiguration
from lazr.restful.interfaces import IJSONPublishable
from lazr.restful.simple import (
    RootResource,
    RootResourceAbsoluteURL,
    )
from lazr.restful.wsgi import BaseWSGIWebServiceConfiguration

from lpresults.tracker.interfaces.webservice import (
    ITestDistribution,
    ITestDistributionSet,
    ITestPerson,
    ITestPersonSet,
    ITestPillarSet,
    ITestProject,
    ITestProjectSet,
    ITestSystemSet,
    )

from lpresults.tracker.services.request import get_host_from_request


class WebServiceConfiguration(
    DjangoWebServiceConfiguration, BaseWSGIWebServiceConfiguration):
    """
    Configuration information for this web service.

    This class is defined here so that Grok can pick it up, but whole
    configuration is actually stored in settings.py
    """
    # The active versions should contain the launchpadlib default.
    active_versions = ["1.0"]
    last_version_with_mutator_named_operations = None

    service_description = """The Launchpad Results web service allows
        automated clients to access most of the functionality available
        on the web site. For help getting started, see
        <a href="{{ LAUNCHPAD_HELP_ROOT }}API/">the help wiki.</a>"""

    version_descriptions = {
        "1.0": """This version of the web service reflects the default
        interface. It may abruptly change without warning."""}

    # The default batch size should reflect the web interface
    default_batch_size = 75

    def get_request_user(self):
        """See `IWebServiceConfiguration`."""
        interaction = queryInteraction()
        if interaction is None:
            return None

        users = [participation.user
            for participation in interaction.participations
            if participation.user is not None]

        if not users:
            return None
        elif len(users) > 1:
            raise ValueError('Too many users')
        else:
            return users[0]


class WebServiceRoot(RootResource):
    """Root resource for the web service"""

    cached_wadl = {}

    def get(self, request, name):
        if name in self.top_level_objects:
            return self.top_level_objects[name]

        # Allow traversal to ~foo for People
        if name.startswith("~"):
            return getUtility(ITestPersonSet).getByName(name[1:])

        return getUtility(ITestPillarSet).retrieveByName(name)

    def toWADL(self):
        """See `HTTPResource`.

        Look for a cached WADL file for the request version at the
        location used by the `apidoc` command. If the file is present,
        load the file and cache its contents rather than generating
        new WADL. Otherwise, generate new WADL and cache it.
        """
        if self.__class__.cached_wadl is None:
            # The cache has been disabled, generate the WADL.
            return super(WebServiceRoot, self).toWADL()

        version = self.request.version
        if  version not in self.__class__.cached_wadl:
            try:
                # It's not cached, try to render it.
                template_name = path.join("apidoc", "wadl-%s.xml" % version)
                service_root = absoluteURL(self, self.request)
                wadl = render_to_string(template_name, {
                    "LAUNCHPAD_RESULTS_SERVICE_ROOT": service_root,
                    })

            except TemplateDoesNotExist:
                # Template does not exist, generate it.
                wadl = super(WebServiceRoot, self).toWADL()

            self.__class__.cached_wadl[version] = wadl

        return self.__class__.cached_wadl[version]

    def _build_top_level_objects(self):
        """Create data structure of top level objects"""
        collections = {
            "distributions": (
                ITestDistribution,
                getUtility(ITestDistributionSet),
                ),
            "people": (
                ITestPerson,
                getUtility(ITestPersonSet),
                ),
            "projects": (
                ITestProject,
                getUtility(ITestProjectSet),
                ),
            }

        entry_links = {
            "pillars": getUtility(ITestPillarSet),
            "systems": getUtility(ITestSystemSet),
            }

        return collections, entry_links


class WebServiceRootAbsoluteURL(RootResourceAbsoluteURL):
    """
    This class is mostly defined so that grok will pick it up but also
    provides support for getting host information from the given request.
    """

    def __init__(self, context, request):
        super(WebServiceRootAbsoluteURL, self).__init__(context, request)

        if not self.hostname:
            self.hostname, self.port = get_host_from_request(request)


class DecimalJSON:
    """Adapter for Decimal objects for JSON serialization"""
    implements(IJSONPublishable)
    adapts(Decimal)

    def __init__(self, context):
        self.context = context

    def toDataForJSON(self):
        return str(self.context)


gsm = getGlobalSiteManager()
gsm.registerAdapter(DecimalJSON)
