# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "WSGIHandler",
    ]

import logging
import traceback

from zope.component import getUtility

from lazr.restful.interfaces import IWebServiceConfiguration

from lpresults.storage.zope.interfaces import IStorageConfiguration

from lpresults.tracker.services.api import APIApplication
from lpresults.tracker.services.download import DownloadApplication
from lpresults.tracker.services.root import RootApplication


class WSGIDispatch:
    """
    Dispatch request to various WSGI applications based on beginning of the
    PATH_INFO.
    """

    def __init__(self, mapping, default_application):
        """
        :param mapping: iterable with each item as (url, wsgi application)
        :param default_application: application to use if none of urls are
                                    matching
        """
        self.mapping = mapping
        self.default_application = default_application

    def __call__(self, environ, start_response):
        path_info = environ.get("PATH_INFO", "/")
        for prefix, application in self.mapping:
            if path_info.startswith(prefix):
                try:
                    return application(environ, start_response)
                except Exception:
                    logging.exception("An unexpected error occurred while "
                        "processing an application request.")
                    start_response("500 Internal Server Error",
                        [("Content-type", "text/plain")])
                    message = traceback.format_exc()
                    return [message]

        return self.default_application(environ, start_response)


class WSGIHandler:

    def __init__(self):
        """Configure and instantiate dispatch handlers."""
        mapping = []

        root_application = RootApplication
        root_application.configure_server(config_file="site.zcml")

        # Add web service mapping only if prefix is defined
        web_service_configuration = getUtility(IWebServiceConfiguration)
        web_service_prefix = web_service_configuration.service_root_uri_prefix
        if web_service_prefix:
            mapping.append(("/%s" % web_service_prefix, APIApplication,))

        # Add storage service mapping only if prefix is defined
        storage_configuration = getUtility(IStorageConfiguration)
        storage_prefix = storage_configuration.root_uri_prefix
        if storage_prefix:
            mapping.append(("/%s" % storage_prefix, DownloadApplication,))

        self.dispatch = WSGIDispatch(mapping, root_application)

    def __call__(self, environ, start_response):
        return self.dispatch(environ, start_response)
