# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Application",
    "ApplicationError",
    "ApplicationRunner",
    ]

import sys

from os import (
    environ,
    path,
    )
from optparse import (
    OptionGroup,
    OptionParser,
    )
from urllib2 import (
    HTTPError,
    URLError,
    urlopen,
    )
from urlparse import urlsplit

# Make launchpadlib happy
try:
    import lazr.restful
except ImportError:
    pass

from lazr.restfulclient import _browser
from lazr.restfulclient.errors import HTTPError as RestfulHTTPError

from launchpadlib.launchpad import Launchpad

from lpresults.xunit.functor import Functor
from lpresults.xunit.logger import get_logger
from lpresults.xunit.retry import retry_except


class ApplicationError(Exception):
    """Error raised when application fails."""


class Application:

    def __init__(self, launchpad_factory, logger):
        self.launchpad_factory = launchpad_factory
        self.logger = logger

    def getLaunchpad(self, *args, **kwargs):
        try:
            return self.launchpad_factory(*args, **kwargs)
        except AttributeError:
            # In stock httplib2, this exception is raised when failing
            # to connect to the remote host:
            #
            #   AttributeError: 'NoneType' object has no attribute 'makefile'
            #
            # See: http://code.google.com/p/httplib2/issues/detail?id=62
            raise ApplicationError(
                "Failed network connection to service root")

        except RestfulHTTPError, error:
            # When launchpadlib fails connecting to the service root,
            # this is an example of exception raised:
            #
            #   HTTPError: HTTP Error 504: Gateway Time-out
            raise ApplicationError(
                "Failed HTTP connection to service root: %s"
                % error.response.reason)

    @retry_except(ApplicationError)
    def getURL(self, url):
        try:
            return urlopen(url)
        except URLError, error:
            # 404 errors indicate a data inconsistency: more than one
            # attempt to open the URL is pointless.
            #
            # Note that URLError is a base class of HTTPError.
            if isinstance(error, HTTPError) and error.code == 404:
                raise

            # HTTPErrors with a 5xx error code ("server problem") are
            # a reason to retry the access again, as well as generic,
            # non-HTTP, URLErrors like "connection refused".
            if (isinstance(error, HTTPError)
                and 500 <= error.code <= 599
                or isinstance(error, URLError) and
                    not isinstance(error, HTTPError)):
                raise ApplicationError(str(error))

            else:
                raise

    def start(self):
        raise NotImplementedError()


class ApplicationRunner:

    # Runner defaults
    application_factory = Application
    usage = "Usage: %prog [OPTIONS]"

    # Launchpad defaults
    default_consumer_name = "Launchpad Results"
    default_service_root = "http://localhost:8000/+api/"
    default_version = "1.0"

    # Logging defaults
    default_log_level = "info"

    def __init__(self, application_factory=None, usage=None):
        if application_factory is not None:
            self.application_factory = application_factory
        if usage is not None:
            self.usage = usage

    def getParser(self, **kwargs):
        parser = OptionParser(usage=self.usage, **kwargs)

        group = OptionGroup(parser, "Launchpad options")
        group.add_option("--consumer-name",
            default=self.default_consumer_name,
            metavar="NAME",
            help="The consumer appropriate for this application,"
                " defaults to: %default")
        group.add_option("--credentials-file",
            metavar="FILE",
            help="If given, the credentials are stored in that file.")
        group.add_option("--launchpadlib-dir",
            metavar="DIR",
            help="The directory for cache and credentials.")
        group.add_option("--service-root",
            default=self.default_service_root,
            metavar="URL",
            help="The URL to the root of the web service,"
                " defaults to: %default")
        group.add_option("--version",
            default=self.default_version,
            help="The version of the web service,"
                " defaults to: %default")
        parser.add_option_group(group)

        group = OptionGroup(parser, "Logging options")
        group.add_option("-l", "--log",
            metavar="FILE",
            help="The file to write the log to.")
        group.add_option("--log-level",
            default=self.default_log_level,
            help="One of debug, info, warning, error or critical.")
        parser.add_option_group(group)

        return parser

    def parseArgs(self, args):
        parser = self.getParser()
        options, args = parser.parse_args(args)
        kwargs = options.__dict__

        # Define logger early
        log = kwargs.pop("log")
        if log:
            if log == "-":
                log_stream, log_file = sys.stdout, None
            else:
                log_stream, log_file = None, log
        else:
            log_stream, log_file = sys.stderr, None
        kwargs["logger"] = get_logger(
            kwargs.pop("log_level"), log_stream, log_file)

        # Define launchpadlib_dir and credentials_file if empty
        launchpadlib_dir = kwargs.pop("launchpadlib_dir")
        if launchpadlib_dir is None:
            home_dir = environ["HOME"]
            launchpadlib_dir = path.join(home_dir, ".lpresults")

        credentials_file = kwargs.pop("credentials_file")
        if credentials_file is None:
            host_name = urlsplit(options.service_root)[1]
            credentials_file = path.join(
                launchpadlib_dir, host_name, "credentials",
                options.consumer_name)

        kwargs["launchpad_factory"] = Functor(Launchpad.login_with,
            launchpadlib_dir=path.expanduser(launchpadlib_dir),
            credentials_file=path.expanduser(credentials_file),
            consumer_name=kwargs.pop("consumer_name"),
            service_root=kwargs.pop("service_root"),
            version=kwargs.pop("version"))

        return args, kwargs

    def createApplication(self, args):
        args, kwargs = self.parseArgs(args)

        return self.application_factory(*args, **kwargs)

    def preApplication(self):
        """
        This should set up any state necessary before loading and
        running the Application.
        """

    def postApplication(self):
        """
        This will be called after the application has finished.
        """

    def run(self, args=None):
        """Run the application."""
        self.preApplication()

        try:
            application = self.createApplication(args)
            application.start()

        except ApplicationError, error:
            sys.stderr.write("Error: " + str(error) + "\n")
            sys.exit(1)

        finally:
            self.postApplication()


# An application can sometimes crash with a 5xx error in
# Launchpad.Browser._request(). This monkey patch attempts
# to retry a few times before giving up.
_original_request = _browser.Browser._request


def retry_request(*args, **kwargs):
    try:
        return _original_request(*args, **kwargs)
    except RestfulHTTPError, error:
        # 404 errors indicate a data inconsistency: more than one
        # attempt to open the URL is pointless.
        if error.response.status == 404:
            raise

        raise ApplicationError(str(error))

_browser.Browser._request = retry_except(ApplicationError)(
    retry_request)
