#!/usr/bin/python

__metaclass__ = type

import shutil

from os import path
from urlparse import urlsplit

from multiprocessing import (
    Process,
    Queue,
    )

from lpresults.xunit.application import (
    Application,
    ApplicationRunner,
    )
from lpresults.xunit.functor import Functor


class MultiApplication(Application):

    # Application defaults
    concurrency = 10

    def __init__(self, concurrency=None, **kwargs):
        super(MultiApplication, self).__init__(**kwargs)

        if concurrency is not None:
            self.concurrency = concurrency

        self.input_queue = Queue()
        self.output_queue = Queue()

    def getLaunchpad(self, *args, **kwargs):
        """See Application."""
        # Cache is broken with multiple processes, so each child process
        # needs to be given a different cache directory but the same
        # credentials can be shared safely.
        #
        # See: https://bugs.launchpad.net/launchpadlib/+bug/459418
        index = kwargs.pop("index", None)
        if index is not None:
            launchpad_kwargs = self.launchpad_factory._kwargs
            parent_launchpadlib_dir = launchpad_kwargs["launchpadlib_dir"]
            child_launchpadlib_dir = path.join(
                parent_launchpadlib_dir, str(index))
            kwargs.setdefault("launchpadlib_dir", child_launchpadlib_dir)

            # Copy the parent cache directory to the child
            host_name = urlsplit(launchpad_kwargs["service_root"])[1]
            child_cache_dir = path.join(
                child_launchpadlib_dir, host_name, "cache")
            if not path.exists(child_cache_dir):
                parent_cache_dir = path.join(
                    parent_launchpadlib_dir, host_name, "cache")
                shutil.copytree(parent_cache_dir, child_cache_dir)

        return super(MultiApplication, self).getLaunchpad(*args, **kwargs)

    def processInput(self):
        pass

    def processOutput(self):
        pass

    def processMessages(self, index):
        raise NotImplementedError()

    def start(self):
        # Establish an early Launchpad connection to get the credentials
        self.getLaunchpad()

        if not self.concurrency:
            self.processInput()
            self.processMessages(None)
            self.processOutput()

        else:
            message_process_functor = Functor(
                Process, target=self.processMessages)

            processes = []
            processes.append(Process(target=self.processInput))
            processes.extend([message_process_functor(args=(index,))
                for index in xrange(self.concurrency)])
            processes.append(Process(target=self.processOutput))

            for process in processes:
                process.start()

            try:
                for process in processes:
                    process.join()
            except KeyboardInterrupt:
                for process in processes:
                    process.terminate()
                    process.join()


class MultiApplicationRunner(ApplicationRunner):

    # Runner defaults
    application_factory = MultiApplication

    # Multi defaults
    default_concurrency = application_factory.concurrency

    def getParser(self, **kwargs):
        """See ApplicationRunner."""
        parser = super(MultiApplicationRunner, self).getParser(**kwargs)

        parser.add_option("-c", "--concurrency",
            default=self.default_concurrency,
            metavar="NUM",
            type="int",
            help="Number of processes to run at a time,"
                " defaults to %default")

        return parser
