# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "DmidecodeParser",
    ]

from re import compile
from string import (
    hexdigits,
    uppercase,
    )

from lpresults.xunit.parsers.dmi import (
    Dmi,
    DmiDevice,
    )


HANDLE_RE = compile(
    r"^Handle (?P<handle>0x[%s]{4}), "
    "DMI type (?P<type>\d+), "
    "(?P<size>\d+) bytes$"
    % hexdigits)
KEY_VALUE_RE = compile("^\t(?P<key>[%s].+):( (?P<value>.+))?$"
    % uppercase)


class DmidecodeParser:

    _attribute_map = {
        "Manufacturer": "vendor",
        "Product Name": "name",
        "Serial Number": "serial",
        "Type": "type",
        "Vendor": "vendor",
        "Version": "version",
        }

    def __init__(self, stream):
        self.stream = stream

    def run(self, result):
        output = self.stream.read()
        for record in output.split("\n\n"):
            # Skip empty records
            if not record:
                continue

            # Skip header record
            lines = record.split("\n")
            line = lines.pop(0)
            if line.startswith("#"):
                continue

            # Skip records with an unsupported handle
            match = HANDLE_RE.match(line)
            if not match:
                continue

            # Skip records that are inactive
            line = lines.pop(0)
            if line == "Inactive":
                continue

            # Skip disabled entries and end-of-table marker
            type_index = int(match.group("type"))
            if type_index >= len(Dmi.type_names):
                continue

            category = Dmi.type_names[type_index]
            category = category.upper().split(" ")[-1]
            if category not in ("BOARD", "BIOS", "CHASSIS", "SYSTEM"):
                continue

            # Parse attributes
            attributes = {}
            for line in lines:
                match = KEY_VALUE_RE.match(line)
                if match:
                    key = match.group("key")
                    value = match.group("value")
                    if key in self._attribute_map:
                        key = "%s_%s" % (
                            category.lower(), self._attribute_map[key])
                        attributes[key] = value

            device = DmiDevice(attributes, category)
            result.addDevice(device)

        return result
