/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The LPAction module provides an abstraction layer to run Launchpad
 * client requests in a deferred context.
 *
 * @module lpresults.lp
 * @submodule action
 */
YUI.add('lpresults.lp.action', function(Y) {

    var namespace = Y.namespace('lpresults.lp');

    /**
     * The LPAction class provides the ability to make a Launchpad request
     * and get a Deferred on which additional callbacks can be defined.
     *
     * @class LPAction
     * @extends Base
     * @constructor
     */
    var LPAction = function(config) {
        LPAction.superclass.constructor.apply(this, arguments);
    };

    LPAction.NAME = 'lpaction';
    LPAction.ATTRS = {
        /**
         * The LP client to use. If not defined, one will be created
         * during initialization.
         *
         * @attribute client
         */
        client: {
            valueFn: function() {
                return new Y.lp.client.Launchpad();
            }
        },

        /**
         * The uri to use for the LP.get request.
         *
         * @attribute uri
         * @type String
         */
        uri: {
            value: null
        },

        /**
         * The LP API operation name (if applicable).
         *
         * @attribute operationName
         * @type String
         */
        operationName: {
            value: null
        }
    };

    Y.extend(LPAction, Y.lpresults.command.Action, {
        /**
         * The run method which calls the LP API based on the configuration
         * passed to the constructor. If the uri is not defined, the API is
         * never called.
         *
         * @method run
         * @public
         */
        run: function(parameters) {
            var deferred = new Y.lpresults.command.Deferred(),
                operationName = this.get('operationName'),
                uri = this.get('uri');

            if (uri) {
                var config = {
                    parameters: parameters,
                    on: {
                        failure: Y.bind(this.runFailure, this, deferred),
                        success: Y.bind(this.runSuccess, this, deferred)
                        }};

                // Call the appropriate method on the client.
                if (operationName !== null) {
                    this.get('client').named_get(uri, operationName, config);
                }
                else {
                    this.get('client').get(uri, config);
                }
            }

            return deferred;
        },

        /**
         * Override to handle a successful request.
         *
         * @method runSuccess
         * @protected
         */
        runSuccess: function(deferred, result) {
            var entries = result.entries,
                attrs, i, value;

            // Optimization to set attribute names based on links
            for (i = 0; i < entries.length; i++) {
                attrs = entries[i].getAttrs();
                for (var key in attrs) {
                    if (attrs.hasOwnProperty(key)) {
                        if (key.substr(-5) === '_link') {
                            value = attrs[key];
                            if (value !== null) {
                                value = value.substr(value.lastIndexOf('/') + 1);
                                // Handle ~ in person names
                                if (value[0] === '%') {
                                    value = value.substr(3);
                                }
                            }
                            key = key.substr(0, key.length - 5);
                            entries[i].addAttr(key, {value: value});
                        }
                    }
                }
            }

            // Need to get the total_size as a separate request if necessary
            if (result.hasOwnProperty('total_size_link')) {
                this.get('client').get(result.total_size_link, {
                    on: {
                        success: Y.bind(
                            this.totalSizeSuccess, this, deferred, result),
                        failure: Y.bind(this.runFailure, this, deferred)
                    }});
            }
            else {
                LPAction.superclass.runSuccess.apply(this, [deferred, result]);
            }
        },

        /**
         * Handle a successful request to show the total size.
         *
         * @method totalSizeSuccess
         * @protected
         */
        totalSizeSuccess: function(deferred, result, total_size) {
            result.total_size = total_size;
            deferred.callback(result);
        }
    });

    /*
     * Ensure that LPAction is available within the lp namespace.
     */
    namespace.LPAction = LPAction;

}, '0.1', {'requires': ['lp.client', 'lpresults.command.action']});
