/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The Batching module provides a class for paginating results in a DOM tree.
 *
 * @module lpresults.ui
 * @submodule batching
 */
YUI.add('lpresults.ui.batching', function(Y) {

    var namespace = Y.namespace('lpresults.ui');

    var BATCHING_INACTIVE = 'inactive',
        INVALID_BATCH_NUMBER = 'Invalid Batch Number',
        PREVIOUS_NODE = 'previousNode',
        NEXT_NODE = 'nextNode',
        FIRST_NODE = 'firstNode',
        LAST_NODE = 'lastNode',
        FIRST_ITEM_NODE = 'firstItemNode',
        LAST_ITEM_NODE = 'lastItemNode',
        TOTAL_ENTRIES_NODE = 'totalEntriesNode',
        NEXT_TEXT = 'nextText',
        PREVIOUS_TEXT = 'previousText',
        FIRST_TEXT = 'firstText',
        LAST_TEXT = 'lastText',
        CHILDREN = 'children',
        CLICK = 'click',
        HOST = 'host',
        TEXT = 'text',
        CONTENT_SIZE = 'contentSize',
        CONTROL_BOX = 'controlBox',
        CURRENT_BATCH = 'currentBatch',
        TOTAL_BATCHES = 'totalBatches',
        TOTAL_ENTRIES = 'totalEntries';

    /**
     * The Batching class extends ContentTable with functionality to batch
     * the results returned from an action.
     *
     * @class Batching
     * @extends ContentTable
     * @constructor
     */
    var Batching = function() {
        Batching.superclass.constructor.apply(this, arguments);
    };

    Batching.NAME = 'batching';
    Batching.ATTRS = {
        /**
         * Parameter name to indicate the number of entries to return.
         *
         * @attribute sizeParameter
         * @type String
         */
        sizeParameter: {
            value: 'size'
        },

        /**
         * Parameter name to indicate start position in the total entries set.
         *
         * @attribute startParameter
         * @type String
         */
        startParameter: {
            value: 'start'
        },

        /**
         * Should the widget attempt to render the first batch. In a progressive
         * enhancement scenario, we can set this to false and render the first batch
         * directly in html, letting the widget render additional batches.
         *
         * @attribute renderFirstBatch
         * @type boolean
         */
        renderFirstBatch: {
            value: true
        },

        /**
         * The total number of entries across all batches.
         *
         * @attribute totalEntries
         * @type Int
         */
        totalEntries: {
            value: 0
        },

        /**
         * The current batch.  Default is 1, the first batch.
         *
         * @attribute currentBatch
         * @type Int
         */
        currentBatch: {
            value: 1
        },

        /**
         * The total number of batches.
         *
         * @attribute totalBatches
         * @type Int
         */
        totalBatches: {
            getter: function() {
                var totalEntries = this.get(TOTAL_ENTRIES);
                if (totalEntries) {
                    return Math.ceil(totalEntries / this.get(CONTENT_SIZE));
                }
                return 0;
            }
        },

        /**
         * Optionally, the control to use as the 'Next' button.
         *
         * @attribute nextControl
         * @type Node
         */
        nextControl: {
            value: null
        },

        /**
         * Optionally, the control to use as the 'Previous' button.
         *
         * @attribute previousControl
         * @type Node
         */
        previousControl: {
            value: null
        }
    };

    Y.extend(Batching, Y.lpresults.ui.ContentTable, {
        /**
         * Override to set the total entries for the batching controls.
         *
         * @method actionSuccess
         * @protected
         */
        actionSuccess: function(result) {
            this.set(TOTAL_ENTRIES, result.total_size);

            Batching.superclass.actionSuccess.apply(this, arguments);
        },

        /**
         * Bind event handlers to hook the widget into the DOM.
         *
         * @method bindUI
         * @protected
         */
        bindUI: function() {
            var nextControl = this.get('nextControl');
            if (nextControl) {
                nextControl.on(CLICK, function(evt) {
                    this.nextBatch();
                }, this);
            }

            var previousControl = this.get('previousControl');
            if (previousControl) {
                previousControl.on(CLICK, function(evt) {
                    this.previousBatch();
                }, this);
            }
        },

        /**
         * Advance to the next batch.  If this is the last batch, this method has no
         * effect.
         *
         * @method nextBatch
         * @public
         */
        nextBatch: function() {
            var batch = this.get(CURRENT_BATCH);
            var totalBatches = this.get(TOTAL_BATCHES);
            if (batch + 1 <= totalBatches) {
                this._switchBatch(batch + 1);
            }
        },

        /**
         * Advance to the previous batch.  If this is the first batch, this method
         * has no effect.
         *
         * @method previousBatch
         * @public
         */
        previousBatch: function() {
            var batch = this.get(CURRENT_BATCH);
            if (batch - 1 >= 1) {
                this._switchBatch(batch - 1);
            }
        },

        /**
         * Method to switch the current batch. Upon changing the current batch
         * the UI is automatically re-rendered.
         *
         * @method jumpToBatch
         * @public
         */
        jumpToBatch: function(batchNumber) {
            var totalBatches = this.get(TOTAL_BATCHES);
            batchNumber = parseInt(batchNumber, 10);
            if (isNaN(batchNumber)) {
                throw new Error(INVALID_BATCH_NUMBER);
            }
            if (batchNumber > totalBatches || batchNumber < 1) {
                throw new Error(INVALID_BATCH_NUMBER);
            }
            this._switchBatch(batchNumber);
        },

        _getParameters: function() {
            var parameters = {};

            var startParameter = this.get('startParameter');
            var start = (this.get(CURRENT_BATCH) - 1) * this.get(CONTENT_SIZE);
            parameters[startParameter] = start;

            var sizeParameter = this.get('sizeParameter');
            var size = this.get(CONTENT_SIZE);
            parameters[sizeParameter] = size;

            return parameters;
        },

        _switchBatch: function(batch) {
            this.set(CURRENT_BATCH, batch);

            var parameters = this._getParameters();
            this.runAction.apply(this, [parameters]);
        }
    });

    /*
     * Ensure that Batching is available within the ui namespace.
     */
    namespace.Batching = Batching;

    /**
     * Plugin renders batching and location status elements when the 'render'
     * event on the host widget fires.
     *
     * @class BatchingPlugin
     * @extends Plugin
     * @constructor
     */
    var BatchingPlugin = function() {
        BatchingPlugin.superclass.constructor.apply(this, arguments);
    };

    BatchingPlugin.NAME = 'batching_plugin';
    BatchingPlugin.NS = 'batching';
    BatchingPlugin.ATTRS = {
        /**
         * The element within which to render batching controls.
         *
         * @attribute controlBox
         * @type Node
         */
        controlBox: {
            value: null
        },

        /**
         * The text to display as the next-batch control, such as 'Older'.  Default
         * is 'Next'.
         *
         * @attribute nextText
         * @type String
         */
        nextText: {
            value: 'Next'
        },

        /**
         * The text to display as the previous-batch control, such as 'Newer'.
         * Default is 'Previous'.
         *
         * @attribute previousText
         * @type String
         */
        previousText: {
            value: 'Previous'
        },

        /**
         * The text to display as the first-batch control, such as 'Newest'.
         * Default is 'First'.
         *
         * @attribute firstText
         * @type String
         */
        firstText: {
            value: 'First'
        },

        /**
         * The text to display as the last-batch control, such as 'Oldest'.
         * Default is 'Last'.
         *
         * @attribute lastText
         * @type String
         */
        lastText: {
            value: 'Last'
        },

        /**
         * An HTML element that represents the 'Next' control.
         *
         * @attribute nextNode
         * @type Node
         */
        nextNode: {
            value: null
        },

        /**
         * An HTML element that represents the 'Previous' control.
         *
         * @attribute previousNode
         * @type Node
         */
        previousNode: {
            value: null
        },

        /**
         * An HTML element that represents the 'First' control.
         *
         * @attribute firstNode
         * @type Node
         */
        firstNode: {
            value: null
        },

        /**
         * An HTML element that represents the 'Last' control.
         *
         * @attribute lastNode
         * @type Node
         */
        lastNode: {
            value: null
        },

        /**
         * An HTML element with the sequence number of the first record on the
         * current batch as text content.
         *
         * @attribute firstItemNode
         * @type Node
         */
        firstItemNode: {
            value: null
        },

        /**
         * An HTML element with the sequence number of the last record on the
         * current batch as text content.
         *
         * @attribute lastItemNode
         * @type Node
         */
        lastItemNode: {
            value: null
        },

        /**
         * An HTML element with the total number of records available in the
         * underlying data set as text content.
         *
         * @attribute totalEntriesNode
         * @type Node
         */
        totalEntriesNode: {
            value: null
        },

        /**
         * Disable the last element navigation, if it can't be used for the batch.
         *
         * @attribute disableLast
         * @type Boolean
         */
        disableLast: {
            value: false
        }
    };

    Y.extend(BatchingPlugin, Y.Plugin.Base, {
        /**
         * Constructor logic.
         *
         * @method initializer
         * @protected
         */
        initializer: function(config) {
            if (config.nextText) {
                this.set(NEXT_TEXT, config.nextText);
            }
            if (config.previousText) {
                this.set(PREVIOUS_TEXT, config.previousText);
            }

            var host = this.get(HOST);
            host.on('before-content-change',
                Y.bind(this._renderMessage, this));
            host.after('after-content-change',
                Y.bind(this._renderControls, this));
        },

        /**
         * Render a message in the control box.
         *
         * @method _renderMessage
         * @private
         */
        _renderMessage: function() {
            var controls = new Y.Node.create(['',
                '<div style="vertical-align: top; text-align: right;">',
                '  <img src="/@@/spinner.gif" />',
                '</div>',
                ''].join(''));
            var controlBox = this.get(CONTROL_BOX);
            controlBox.get(CHILDREN).remove();
            controlBox.appendChild(controls);
        },

        /**
         * Render controls and bind event handlers.
         *
         * @method _renderControls
         * @private
         */
        _renderControls: function() {
            this._renderUI();
            this._bindUI();
            this._syncUI();
        },

        /**
         * Render the control box.
         *
         * @method _renderUI
         * @private
         */
        _renderUI: function() {
            var next = this.get(NEXT_TEXT);
            var previous = this.get(PREVIOUS_TEXT);
            var first = this.get(FIRST_TEXT);
            var last = this.get(LAST_TEXT);
            var controls = new Y.Node.create(['',
                '<table style="width: 100%;" class="upper-batch-nav">',
                '  <tbody>',
                '    <tr>',
                '      <td style="white-space: nowrap" class="batch-navigation-index">',
                '        <strong class="first-item"></strong>',
                '        &rarr; ',
                '        <strong class="last-item"></strong>',
                '        of <span class="total-entries"></span> results',
                '      </td>',
                '      <td style="text-align: right; white-space: nowrap" class="batch-navigation-links">',
                '        <a href="" class="first">' + first + '</a>',
                '        &#149;',
                '        <a href="" class="previous">' + previous + '</a>',
                '        &#149;',
                '        <a href="" class="next"><strong>' + next + '</strong></a>',
                '        &#149;',
                '        <a href="" class="last">' + last + '</a>',
                '      </td>',
                '    </tr>',
                '  </tbody>',
                '</table>',
                ''].join(''));
            var controlBox = this.get(CONTROL_BOX);
            controlBox.get(CHILDREN).remove();
            controlBox.appendChild(controls);

            this.set(NEXT_NODE, controlBox.one('.next'));
            this.set(PREVIOUS_NODE, controlBox.one('.previous'));
            this.set(FIRST_NODE, controlBox.one('.first'));
            this.set(LAST_NODE, controlBox.one('.last'));
            this.set(FIRST_ITEM_NODE, controlBox.one('.first-item'));
            this.set(LAST_ITEM_NODE, controlBox.one('.last-item'));
            this.set(TOTAL_ENTRIES_NODE, controlBox.one('.total-entries'));
        },

        /**
         * Bind the location status in the control box to event handlers.
         *
         * @method bindUI
         * @private
         */
        _bindUI: function() {
            this.get(NEXT_NODE).on(CLICK, Y.bind(this._nextBatch, this));
            this.get(PREVIOUS_NODE).on(CLICK, Y.bind(this._previousBatch, this));
            this.get(FIRST_NODE).on(CLICK, Y.bind(this._firstBatch, this));
            this.get(LAST_NODE).on(CLICK, Y.bind(this._lastBatch, this));
        },

        /**
         * Switch to the next batch and update the location status controls.
         *
         * @method nextBatch
         * @private
         */
        _nextBatch: function(evt) {
            evt.preventDefault();
            if (!evt.target.hasClass(BATCHING_INACTIVE)) {
                this.get(HOST).nextBatch();
            }
        },

        /**
         * Switch to the previous batch and update location status controls.
         *
         * @method previousBatch
         * @private
         */
        _previousBatch: function(evt) {
            evt.preventDefault();
            if (!evt.target.hasClass(BATCHING_INACTIVE)) {
                this.get(HOST).previousBatch();
            }
        },

        /**
         * Switch to the first batch and update location status controls.
         *
         * @method firstBatch
         * @private
         */
        _firstBatch: function(evt) {
            evt.preventDefault();
            if (!evt.target.hasClass(BATCHING_INACTIVE)) {
                this.get(HOST).jumpToBatch(1);
            }
        },

        /**
         * Switch to the last batch and update location status controls.
         *
         * @method lastBatch
         * @private
         */
        _lastBatch: function(evt) {
            evt.preventDefault();
            var lastBatch = this.get(HOST).get(TOTAL_BATCHES);
            if (!evt.target.hasClass(BATCHING_INACTIVE)) {
                this.get(HOST).jumpToBatch(lastBatch);
            }
        },

        /**
         * Synchronize this plugin's state with the host widget's state.
         *
         * @method _syncUI
         * @private
         */
        _syncUI: function() {
            var host = this.get(HOST);
            var totalEntriesCount = host.get(TOTAL_ENTRIES);
            var totalBatches = host.get(TOTAL_BATCHES);
            var contentSize = host.get(CONTENT_SIZE);
            var currentBatch = host.get(CURRENT_BATCH);
            var firstItemCount = (currentBatch - 1) * contentSize + 1;
            var lastItemCount = Math.min(
                firstItemCount + contentSize - 1, totalEntriesCount);
            var onlyOneBatch = totalEntriesCount <= contentSize;
            var onLastBatch = totalEntriesCount == lastItemCount;
            var disableNext = onlyOneBatch || onLastBatch;
            var disablePrevious = firstItemCount <= 1;
            var disableFirst = currentBatch <= 2;
            var disableLast = this.get('disableLast') || ! (
                totalBatches > 2 && currentBatch < totalBatches - 1);
            this._updateUI(
                firstItemCount, lastItemCount, totalEntriesCount,
                disableNext, disablePrevious, disableFirst, disableLast);
        },

        /**
         * Update the location status in the control box to reflect the
         * specified values.
         *
         * @method _updateUI
         * @private
         */
        _updateUI: function(
            firstItemCount, lastItemCount, totalEntriesCount,
            disableNext, disablePrevious, disableFirst, disableLast) {

            this.get(FIRST_ITEM_NODE).set(TEXT, firstItemCount);
            this.get(LAST_ITEM_NODE).set(TEXT, lastItemCount);
            this.get(TOTAL_ENTRIES_NODE).set(TEXT, totalEntriesCount);
            if (disablePrevious) {
                this.get(PREVIOUS_NODE).addClass(BATCHING_INACTIVE);
            } else {
                this.get(PREVIOUS_NODE).removeClass(BATCHING_INACTIVE);
            }
            if (disableNext) {
                this.get(NEXT_NODE).addClass(BATCHING_INACTIVE);
            } else {
                this.get(NEXT_NODE).removeClass(BATCHING_INACTIVE);
            }
            if (disableFirst) {
                this.get(FIRST_NODE).addClass(BATCHING_INACTIVE);
            } else {
                this.get(FIRST_NODE).removeClass(BATCHING_INACTIVE);
            }
            if (disableLast) {
                this.get(LAST_NODE).addClass(BATCHING_INACTIVE);
            } else {
                this.get(LAST_NODE).removeClass(BATCHING_INACTIVE);
            }
        }
    });

    /*
     * Ensure that BatchingPlugin is available within the ui namespace.
     */
    namespace.BatchingPlugin = BatchingPlugin;

}, '0.1', {'requires': ['node', 'plugin', 'lpresults.ui.content']});
