/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The Content module provides a class for rendering a content box.
 *
 * @module lpresults.ui
 * @submodule content
 */
YUI.add('lpresults.ui.content', function(Y) {

    var namespace = Y.namespace('lpresults.ui');

    var DEFAULT_CONTENT_SIZE = 10,
        EVT_BEFORE_CONTENT_CHANGE = 'before-content-change',
        EVT_BEFORE_CONTENT_RENDER = 'before-content-render',
        EVT_AFTER_CONTENT_CHANGE = 'after-content-change',
        ENTRIES = 'entries',
        CONTENT_BOX = 'contentBox',
        ELEMENT_TEMPLATE = 'elementTemplate';

    /**
     * The Content class provides the ability to plugin functionality into
     * a DOM tree so that it can update itself when results are returned
     * from an action.
     *
     * @class Content
     * @extends Widget
     * @constructor
     */
    var Content = function() {
        Content.superclass.constructor.apply(this, arguments);
    };

    Content.NAME = 'content';
    Content.ATTRS = {
        /**
          * The action on which the Content will add a callback to update
          * the DOM tree with the results from a successful run. If this
          * is undefined, the tree will never be updated.
          *
          * @attribute action
          * @type Action
          */
        action: {
            value: null
        },

        /**
         * Parameter name to indicate the number of entries to return.
         *
         * @attribute sizeParameter
         * @type String
         */
        sizeParameter: {
            value: "size"
        },

        /**
         * An array of objects representing rows for the current batch.
         *
         * @attribute entries
         * @type Array
         */
        entries: {
            value: null
        },

        /**
         * The element to use as a template, when creating new items by
         * the default updater. Custom updaters may or may not use this.
         *
         * @attribute templateItem
         * @type Node
         */
        elementTemplate: {
            value: null
        },

        /**
         * An optional custom function used to update an element with data,
         * this function is called with the YUI node, the data entry and
         * the key.
         *
         * @attribute elementUpdater
         * @type Function
         */
        elementUpdater: {
            value: null
        },

        /**
         * The size of the content where entries will be populated.
         *
         * @attribute contentSize
         * @type NUM
         * @default DEFAULT_CONTENT_SIZE
         */
        contentSize: {
            value: DEFAULT_CONTENT_SIZE
        }
    };

    Y.extend(Content, Y.Widget, {
        /**
         * Initialization Logic. Publishes custom events.
         *
         * @method initializer
         * @protected
         */
        initializer: function() {
            /**
             * Notification event before the widget requests data for
             * a content change.
             *
             * @event content:before-content-change
             * @param {EventFacade} e The Event Facade
             */
            this.publish(EVT_BEFORE_CONTENT_CHANGE);

            /**
             * Notification event before the widget renders the items
             * for new content data.
             *
             * @event content:before-content-render
             * @param {EventFacade} e The Event Facade
             */
            this.publish(EVT_BEFORE_CONTENT_RENDER);

            /**
             * Notification event after the content has been rendered
             * with new data.
             *
             * @event content:after-content-change
             * @param {EventFacade} e The Event Facade
             */
            this.publish(EVT_AFTER_CONTENT_CHANGE);
        },

        /**
         * Return the container node which contains the elements to update.
         *
         * @method getContainer
         * @protected
         */
        getContainer: function() {
            return this.get(CONTENT_BOX);
        },

        /**
         * Return all the elements in the container.
         *
         * @method getElements
         * @protected
         */
        getElements: function() {
            var container = this.getContainer();
            return container.all();
        },

        /**
         * Run the content action with the given parameters. Return a Deferred
         * firing with the result of the action, or failing with an error
         * if the action failed.
         *
         * @method runAction
         * @protected
         */
        runAction: function() {
            this.fire(EVT_BEFORE_CONTENT_CHANGE);

            var action = this.get('action');
            if (action) {
                var deferred = action.run.apply(action, arguments);
                deferred.addCallback(Y.bind(this.actionSuccess, this));
                return deferred;
            }
        },

        /**
         * Callback when the action succeeds.
         *
         * @method actionSuccess
         * @protected
         */
        actionSuccess: function(result) {
            this.set(ENTRIES, result.entries);

            // Fire event after we have data, before modifying items.
            this.fire(EVT_BEFORE_CONTENT_RENDER);

            // Ensure proper number of items in the list.
            this._renderContainer();

            // Update items in the list.
            this._updateContainer();

            // Fire event after render.
            this.fire(EVT_AFTER_CONTENT_CHANGE);
        },

        /**
         * Render the widget.
         *
         * @method renderUI
         * @protected
         */
        renderUI: function() {
            var elementTemplate = this.getElements().item(0);
            this.set(ELEMENT_TEMPLATE, elementTemplate);
            elementTemplate.remove();

            this.runAction.apply(this);
        },

        /**
         * Render the container to match the number of elements in the
         * result entries.
         *
         * @method _renderContainer
         * @private
         */
        _renderContainer: function() {
            var i,
                entries = this.get(ENTRIES),
                contentSize = this.get('contentSize'),
                container = this.getContainer(),
                elements = this.getElements(),
                existingElementCount = elements.size(),
                neededElementCount = (entries.length > contentSize) ?
                    contentSize :
                    entries.length;

            // If the current batch has too many elements remove the excess
            if (existingElementCount > neededElementCount) {
                for (i = neededElementCount; i < existingElementCount; i++) {
                    elements.item(i).remove();
                }
                return;
            }

            // Create additional elements as needed.
            var elementTemplate = this.get(ELEMENT_TEMPLATE);
            for (i = existingElementCount; i < neededElementCount; ++i) {
                var element = elementTemplate.cloneNode(true);
                container.appendChild(element);
            }
        },

        /**
         * Update the container to match the values from the result entries.
         *
         * @method update
         * @private
         */
        _updateContainer: function() {
            var element, entry,
                entries = this.get(ENTRIES),
                elements = this.getElements(),
                elementCount = elements.size();

            updater = this.get('elementUpdater');
            if (updater === null) {
                updater = this._updateElement;
            }
            for (var i = 0; i < elementCount; i++) {
                element = elements.item(i);
                entry = entries[i];
                updater.apply(this, [element, entry, i]);
            }
        },

        /**
         * Update a single element by looking for children matching the id
         * of each key for the given entry.
         *
         * @method _updateElement
         * @private
         */
        _updateElement: function(element, entry, index) {
            var attrs = entry.getAttrs();
            for (var key in attrs) {
                if (attrs.hasOwnProperty(key)) {
                    // XXX this might be a bit excessive O(element*col),
                    // basically does a query per cell.
                    var node = element.one('.' + key);
                    if (node) {
                        this._updateNode(node, entry, key);
                    }
                }
            }
        },

        /**
         * Update a single node by looking for links, dates, etc.
         *
         * @method _updateNode
         * @private
         */
        _updateNode: function(node, entry, key) {
            var value = entry.get(key);
            if (key.match('_link$')) {
                value = value.replace(/\/\+api\/[^\/]*\//, '/');
                node.setAttribute('href', value);
            }
            else {
                if (typeof value === "string") {
                    var date = Y.DataType.Date.parse(value);
                    if (date !== null) {
                        var formatter = new Y.lpresults.ui.DateFormatter({
                            date: date
                            });
                        value = formatter.displayDate();
                    }
                }
                node.set('text', value);
            }
        }
    });

    /*
     * Ensure that Content is available within the ui namespace.
     */
    namespace.Content = Content;

    /**
     * The ContentList class extends Content for updating a list.
     *
     * @class ContentList
     * @extends Content
     * @constructor
     */
    var ContentList = function() {
        ContentList.superclass.constructor.apply(this, arguments);
    };

    ContentList.NAME = 'content_list';

    Y.extend(ContentList, Content, {
        /**
         * Return the list node which contains the items to update.
         *
         * @method getContainer
         * @protected
         */
        getContainer: function() {
            return this.get(CONTENT_BOX).one('ul');
        },

        /**
         * Return all the items in the list.
         *
         * @method getElements
         * @protected
         */
        getElements: function() {
            var container = this.getContainer();
            return container.all('li');
        }
    });

    /*
     * Ensure that ContentList is available within the ui namespace.
     */
    namespace.ContentList = ContentList;

    /**
     * The ContentTable class extends Content for updating a table.
     *
     * @class ContentTable
     * @extends Content
     * @constructor
     */
    var ContentTable = function() {
        ContentTable.superclass.constructor.apply(this, arguments);
    };

    ContentTable.NAME = 'content_table';

    Y.extend(ContentTable, Content, {
        /**
         * Return the table body which contains the rows to update.
         *
         * @method getContainer
         * @protected
         */
        getContainer: function() {
            return this.get(CONTENT_BOX).one('tbody');
        },

        /**
         * Return all the rows in the table.
         *
         * @method getElements
         * @protected
         */
        getElements: function() {
            var container = this.getContainer();
            return container.all('tr');
        }
    });

    /*
     * Ensure that ContentTable is available within the ui namespace.
     */
    namespace.ContentTable = ContentTable;

}, '0.1', {'requires': [
    'widget', 'event', 'event-custom', 'datatype-date',
    'lpresults.ui.formatter']});
