/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The DateFormatter module provides a class for formatting dates to a string.
 *
 * @module lpresults.ui
 * @submodule formatter
 */
YUI.add('lpresults.ui.formatter', function(Y) {

    var namespace = Y.namespace('lpresults.ui');

    var DAY = 24*60*60,
        HOUR = 60*60,
        MINUTE = 60;

    /**
     * Adapter from Date objects to a formatted string.
     *
     * @class DateFormatter
     * @extends Base
     * @constructor
     */
    var DateFormatter = function() {
        DateFormatter.superclass.constructor.apply(this, arguments);
    };

    DateFormatter.NAME = 'date_formatter';
    DateFormatter.ATTRS = {
        date: {
            value: null
        }
    };

    Y.extend(DateFormatter, Y.Base, {

        /**
         * Return the current date to determine the relative time. This
         * method is mostly intended to be overridden in tests.
         *
         * @method _now
         * @private
         */
        _now: function() {
            return new Date();
        },

        /**
         * Return the time portion of the date in a consistent format.
         *
         * @method time
         * @public
         */
        time: function() {
            return Y.DataType.Date.format(this.get('date'), {
                format: '%T'
                });
        },

        /**
         * Return the date portion of the date in a consistent format.
         *
         * @method date
         * @public
         */
        date: function() {
            return Y.DataType.Date.format(this.get('date'), {
                format: '%Y-%m-%d'
                });
        },

        /**
         * Return both the date and time portions of the date in a
         * consistent format.
         *
         * @method datetime
         * @public
         */
        datetime: function() {
            return this.date() + ' ' + this.time();
        },

        /**
         * Display the date in a relative format. If the date is within
         * a day of now, approximate the date.
         *
         * @method displaydate
         * @public
         */
        displayDate: function() {
            var delta = Math.abs(
                (this._now().getTime() - this.get('date').getTime()) / 1000);
            if (delta > DAY) {
                // far in the past or future, display the date
                return 'on ' + this.date();
            }

            return this.approximateDate();
        },

        /**
         * Approximate the date relative to now in a friendly format.
         *
         * @method approximateDate
         * @public
         */
        approximateDate: function() {
            var delta = Math.floor(
                (this._now().getTime() - this.get('date').getTime()) / 1000);
            if (Math.abs(delta) > DAY) {
                // far in the past or future, display the date
                return this.date();
            }

            var future = delta < 0;
            delta = Math.abs(delta);
            var days = Math.floor(delta / DAY);
            var hours = Math.floor(delta / HOUR);
            var minutes = Math.floor((delta - (HOUR * hours)) / MINUTE);
            var seconds = delta % MINUTE;

            var amount = 0;
            var result = future ? 'in ' : '';
            if (days !== 0) {
                amount = days;
                unit = 'day';
            }
            else if (hours !== 0) {
                amount = hours;
                unit = 'hour';
            }
            else if (minutes !== 0) {
                amount = minutes;
                unit = 'minute';
            }
            else if (seconds !== 0) {
                amount = seconds;
                unit = 'second';
            }

            if (amount !== 1) {
                unit += 's';
            }

            result += amount + ' ' + unit;
            if (!future) {
                result += ' ago';
            }

            return result;
        }
    });

    /*
     * Ensure that DateFormatter is available within the ui namespace.
     */
    namespace.DateFormatter = DateFormatter;

}, '0.1', {'requires': ['base', 'intl', 'datatype-date', 'datatype-date-format']});
