/* Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
 * GNU Affero General Public License version 3 (see the file LICENSE).
 *
 * The Listing module provides a class for rendering a list of results
 * in a DOM tree.
 *
 * @module lpresults.ui
 * @submodule listing
 */
YUI.add('lpresults.ui.listing', function(Y) {

    var namespace = Y.namespace('lpresults.ui');

    var ALL_CLASS = 'allClass',
        ALL_NODE = 'allNode',
        ALL_TEXT = 'allText',
        CHILDREN = 'children',
        CLICK = 'click',
        HOST = 'host',
        CONTROL_BOX = 'controlBox',
        LISTING_INACTIVE = 'inactive';

    /**
     * The Listing class extends ContentList with functionality to list
     * only part of the results returned from an action.
     *
     * @class Listing
     * @extends ContentList
     * @constructor
     */
    var Listing = function(config) {
        Listing.superclass.constructor.apply(this, arguments);
    };

    Listing.NAME = 'listing';
    Listing.ATTRS = {
        /**
         * An optional control to use as the 'All' button.
         *
         * @attribute allControl
         * @type Node
         */
        allControl: {
            value: null
        },

        /**
         * An optional link where to redirect the 'All' button.
         *
         * @attribute allLink
         * @type String
         */
        allLink: {
            value: null
        },

        /**
         * The DOM document object.
         */
        document: {
            value: Y.config.doc
        }
    };

    Y.extend(Listing, Y.lpresults.ui.ContentList, {
        /**
         * Bind event handlers to hook the widget into the DOM.
         *
         * @method bindUI
         * @protected
         */
        bindUI: function() {
            var allControl = this.get('allControl');
            if (allControl) {
                allControl.on(CLICK, function(evt) {
                    this.allList();
                }, this);
            }
        },

        /**
         * Redirect to the all link.
         *
         * @method allList
         * @public
         */
        allList: function() {
            var allLink = this.get('allLink');
            if (allLink) {
                var document = this.get('document');
                document.location = allLink;
            }
        }
    });

    /*
     * Ensure that Listing is available within the ui namespace.
     */
    namespace.Listing = Listing;

    /**
     * Plugin renders listing and control elements when the 'render'
     * event on the host widget fires.
     *
     * @class ListingPlugin
     * @extends Plugin
     * @constructor
     */
    var ListingPlugin = function() {
        ListingPlugin.superclass.constructor.apply(this, arguments);
    };

    ListingPlugin.NAME = 'listing_plugin';
    ListingPlugin.NS = 'listing';
    ListingPlugin.ATTRS = {
        /**
         * The class to display in the all-link control. Default is
         * 'batching-all'.
         *
         * @attribute allClass
         * @type String
         */
        allClass: {
            value: 'batching-all'
        },

        /**
         * The text to display as the all-link control. Default is 'All'.
         *
         * @attribute allText
         * @type String
         */
        allText: {
            value: 'All'
        },

        /**
         * An HTML element that represents the 'All' control.
         *
         * @attribute nextNode
         * @type Node
         */
        allNode: {
            value: null
        },

        /**
         * The element within which to render listing controls.
         *
         * @attribute controlBox
         * @type Node
         */
        controlBox: {
            value: null
        }
    };

    Y.extend(ListingPlugin, Y.Plugin.Base, {
        /**
         * Constructor logic.
         *
         * @method initializer
         * @protected
         */
        initializer: function(config) {
            if (config.allText) {
                this.set(ALL_TEXT, config.allText);
            }

            var host = this.get(HOST);
            host.on('before-content-change',
                Y.bind(this._renderMessage, this));
            host.after('after-content-change',
                Y.bind(this._renderControls, this));
        },

        /**
         * Render a message in the control box.
         *
         * @method _renderMessage
         * @private
         */
        _renderMessage: function() {
            var controls = new Y.Node.create(['',
                '<span style="vertical-align: top;">',
                '  <img src="/@@/spinner.gif" />',
                '</span>',
                ''].join(''));
            var controlBox = this.get(CONTROL_BOX);
            controlBox.get(CHILDREN).remove();
            controlBox.appendChild(controls);
        },

        /**
         * Render controls and bind event handlers.
         *
         * @method _renderControls
         * @private
         */
        _renderControls: function() {
            this._renderUI();
            this._bindUI();
            this._syncUI();
        },

        /**
         * Render the control box.
         *
         * @method _renderUI
         * @private
         */
        _renderUI: function() {
            var all = this.get(ALL_TEXT);
            var all_class = this.get(ALL_CLASS);
            var controls = new Y.Node.create(['',
                '<a href="" class="' + all_class + '">' + all + '</a>',
                ''].join(''));

            var controlBox = this.get(CONTROL_BOX);
            controlBox.get(CHILDREN).remove();
            controlBox.appendChild(controls);

            this.set(ALL_NODE, controlBox.one('.' + all_class));
        },

        /**
         * Bind the control box to event handlers.
         *
         * @method bindUI
         * @private
         */
        _bindUI: function() {
            this.get(ALL_NODE).on(CLICK, Y.bind(this._allList, this));
        },

        /**
         * Switch to the all list.
         *
         * @method _allList
         * @private
         */
        _allList: function(evt) {
            evt.preventDefault();
            if (!evt.target.hasClass(LISTING_INACTIVE)) {
                this.get(HOST).allList();
            }
        },

        /**
         * Synchronize this plugin's state with the host widget's state.
         *
         * @method _syncUI
         * @private
         */
        _syncUI: function() {
            var host = this.get(HOST);
            var entries = host.get('entries');
            var disableAll = !entries.length;
            this._updateUI(disableAll);
        },

        /**
         * Update the control box to reflect the specified values.
         *
         * @method _updateUI
         * @private
         */
        _updateUI: function(disableAll) {
            if (disableAll) {
                this.get(ALL_NODE).addClass(LISTING_INACTIVE);
            } else {
                this.get(ALL_NODE).removeClass(LISTING_INACTIVE);
            }
        }
    });

    /*
     * Ensure that ListingPlugin is available within the ui namespace.
     */
    namespace.ListingPlugin = ListingPlugin;

}, '0.1', {'requires': ['node', 'plugin', 'lpresults.ui.content']});
